
/************************************************************************
	File Name : cfiflash.c

	Function:	This is a driver for CFI compliant Flash Memory.

************************************************************************/


#include <stdio.h>
#include <stdlib.h>		  
#include <string.h>

#include "config.h"
#include "boarddef.h"
#include "boarddrv.h"

#ifdef ISP_CFI_EXTERNAL_FLASH



static cfi_result_t CfiFlashDataToggle( void );
static int CfiFlashBlockFailedErase( unsigned char Block );

#define min(a,b) ( a < b ? a : b)
#define max(a,b) ( a > b ? a : b)

#define MANUFACTURER_ST (0x0020) /* Manufacturer code */

/* This to support 8 and 16 BUS  
   MASK for 8 bit  bus is 0xFF 
   MASK for 16 bit bus is 0xFFFF
*/
#define BUS_SIZE_MASK 0xFFFF


#define EXPECTED_DEVICE (0x2249 & BUS_SIZE_MASK ) /* Device code for the M29W160DB */


/* Block organisation for Bottom Boot Block devices */
static const unsigned long BlockOffset[] =
  {
    0x00000L, /* Start offset of block 0 */
    0x02000L, /* Start offset of block 1 */
    0x03000L, /* Start offset of block 2 */
    0x04000L, /* Start offset of block 3 */
    0x08000L, /* Start offset of block 4 */
    0x10000L, /* Start offset of block 5 */
    0x18000L, /* Start offset of block 6 */
    0x20000L, /* Start offset of block 7 */
    0x28000L, /* Start offset of block 8 */
    0x30000L, /* Start offset of block 9 */
    0x38000L, /* Start offset of block 10 */
    0x40000L, /* Start offset of block 11 */
    0x48000L, /* Start offset of block 12 */
    0x50000L, /* Start offset of block 13 */
    0x58000L, /* Start offset of block 14 */
    0x60000L, /* Start offset of block 15 */
    0x68000L, /* Start offset of block 16 */
    0x70000L, /* Start offset of block 17 */
    0x78000L, /* Start offset of block 18 */
    0x80000L, /* Start offset of block 19 */
    0x88000L, /* Start offset of block 20 */
    0x90000L, /* Start offset of block 21 */
    0x98000L, /* Start offset of block 22 */
    0xA0000L, /* Start offset of block 23 */
    0xA8000L, /* Start offset of block 24 */
    0xB0000L, /* Start offset of block 25 */
    0xB8000L, /* Start offset of block 26 */
    0xC0000L, /* Start offset of block 27 */
    0xC8000L, /* Start offset of block 28 */
    0xD0000L, /* Start offset of block 29 */
    0xD8000L, /* Start offset of block 30 */
    0xE0000L, /* Start offset of block 31 */
    0xE8000L, /* Start offset of block 32 */
    0xF0000L, /* Start offset of block 33 */
    0xF8000L /* Start offset of block 34 */
  };


#define NUM_BLOCKS (sizeof(BlockOffset)/sizeof(BlockOffset[0]))

#define FLASH_SIZE (0x100000L) /* 1MB Words*/


#ifdef __IAR_SYSTEMS_ICC__
#define BASE_ADDR  ((  volatile unsigned int * ) 0x00800000)
#endif

#ifdef __NSC_CRCC__
#define BASE_ADDR  ((  FAR volatile unsigned int * const) 0x00800000)
#endif

/* Any address offset within the Flash Memory will do */
#define ANY_ADDR (0x0000L)

#define CfiFlashWrite( Offset, Value ) ( BASE_ADDR[Offset] = (unsigned int) Value )
#define CfiFlashRead( Offset ) ( BASE_ADDR[Offset] )

/* Pending a Microsecond timer facility */
#if 0


/*******************************************************************************
	Function: void CfiFlashPause( unsigned int uMicroSeconds )

	Arguments: uMicroSeconds is the length of the pause in microseconds
	
	Returns: none
	
	Description: This routine returns after uMicroSeconds have elapsed. It is used
				 in several parts of the code to generate a pause required for correct
				 operation of the flash part.
				 The routine here works by counting. 
				 The user may already have a more suitable routine for timing which can be used.
	Pseudo Code:
				Step 1: Compute count size for pause.
				Step 2: Count to the required size.
        
        come back here as timer_wait is not the right call for this granularity
*******************************************************************************/
static void CfiFlashPause( unsigned int uMicroSeconds )
{
}
#endif

/*******************************************************************************
	Function: CfiFlashGetNumBlocks

	Arguments: none
	
	Returns: the numbers of blocks 
	
	Description:  returns the value of global variable NUM_BLOCKS
  Pseudo Code:
			
*******************************************************************************/

int CfiFlashGetNumBlocks(void) 
{
  return NUM_BLOCKS;
} 



/*******************************************************************************
	Function: void FlashReadReset( void )
	Arguments: none
	Return Value: none
	Description: This function places the flash in the Read mode described
				 in the Data Sheet. In this mode the flash can be read as normal memory.
				 All of the other functions leave the flash in the Read mode so this is
				 not strictly necessary. It is provided for completeness.

	Note:		 A wait of 10us is required if the command is called during a program or
				 erase instruction. This is included here to guarantee correct operation.
				 The functions in this library call this function if they suspect 
				 an error during programming or erasing so that the 10us pause 
				 is included. Otherwise they use the single instruction 
				 technique for increased speed.
	Pseudo Code:
				Step 1: write command sequence (see Commands Table of the Data Sheet)
				Step 2: wait 10us
*******************************************************************************/
void CfiFlashReadReset( void )
{

  /* Step 1: write command sequence */
  CfiFlashWrite( 0x0555L, 0x00AA );	/* 1st Cycle */
  CfiFlashWrite( 0x02AAL, 0x0055 );	/* 2nd Cycle */
  CfiFlashWrite( ANY_ADDR, 0x00F0 ); /* 3rd Cycle: write 0x00F0 to ANY address */
  /* Step 2: wait 10us */	
  //CfiFlashPause( 10 );
}
/*******************************************************************************
	Function: int CfiFlashAutoSelect( int iFunc )
	Arguments: iFunc should be set either to the Read Signature values or to the
			   block number. The header file defines the values for reading the Signature.

	Note:	   the first block is Block 0
	Return Value: When iFunc is >= 0 the function returns 
  CFI_FLASH_BLOCK_PROTECTED   if the block is protected (0x01)  
  CFI_FLASH_BLOCK_UNPROTECTED if it is unprotected (0x00) 
				  See the Auto Select command in the Data Sheet for further information.
	When iFunc is CFI_FLASH_READ_MANUFACTURER  the function returns 
	the manufacturers code. The Manufacturer code for ST is 0020h.
	When iFunc is CFI_FLASH_READ_DEVICE_CODE  the function returns 
	the Device Code. 
	The device codes for the parts are:
		M29F160BT 22CCh
		M29F160BB 224Bh
		M29W160BT 22C4h
		M29W160BB 2249h
		M29W160DT 22C4h
		M29W160DB 2249h
	When iFunc is invalid the function returns FLASH_BLOCK_INVALID (-5)

	Description: This function can be used to read the electronic signature of the
				 device, the manufacturer code or the protection level of a block.
	Pseudo Code:
		Step 1: Send the Auto Select command to the device
		Step 2: Read the required function from the device.
		Step 3: Return the device to Read mode.
*******************************************************************************/
int CfiFlashAutoSelect(func_t func)
{
  int retVal = 0;
  int protection = 0;
  
  CfiFlashWrite( 0x0555L, 0x00AA ); /* 1st Cycle */
  CfiFlashWrite( 0x02AAL, 0x0055 ); /* 2nd Cycle */
  CfiFlashWrite( 0x0555L, 0x0090 ); /* 3rd Cycle */
	 
  if ( func == CFI_FLASH_READ_MANUFACTURER) {
    /* A0 = 0 , A1 = 0 */
    retVal = (int) CfiFlashRead( 0x0000L); 
  }
  else { 
    if (func == CFI_FLASH_READ_DEVICE_CODE) {
      /* A0 = 1 ; A1 = 0; */
      retVal = (int) CfiFlashRead( 0x0001L);
    }
    else { 
      if ( ((int)func >= 0) && ((int)func < NUM_BLOCKS) ) {
        protection = CfiFlashRead(BlockOffset[(int)func] + 0x0002L);
        retVal = protection ? CFI_FLASH_BLOCK_PROTECTED : CFI_FLASH_BLOCK_UNPROTECTED;
      }
      else {
        retVal = (int)CFI_FLASH_BLOCK_INVALID;
      }
    }
  }
  CfiFlashReadReset();
  return retVal;
} /* end CfiFlashAutoSelect */

/*******************************************************************************
	Function: int CfiFlashPageErase( unsigned char NumBlocks,unsigned char Block[] )
	Arguments: NumBlocks holds the number of blocks in the array Block
			   Block is an array containing the blocks to be erased.
	Return Value: The function returns the following conditions:
					CFI_FLASH_SUCCESS 
					CFI_FLASH_TOO_MANY_BLOCKS 
					CFI_FLASH_MPU_TOO_SLOW 
					CFI_FLASH_WRONG_TYPE 
					CFI_FLASH_ERASE_FAIL 
				Number of the first protected or invalid block
				The user's array, Block[] is used to report errors on the specified
				blocks. If a time-out occurs because the MPU is too slow then the blocks
				in Blocks which are not erased are overwritten with FLASH_BLOCK_NOT_ERASED
				(FFh) and the function returns FLASH_MPU_TOO_SLOW.
				If an error occurs during the erasing of the blocks the function returns
				FLASH_ERASE_FAIL.
				If both errors occur then the function will set the Block array to
				FLASH_BLOCK_NOT_ERASED for the unerased blocks. It will return
				FLASH_ERASE_FAIL even though the FLASH_MPU_TOO_SLOW has also occurred.
	
	  Description: This function erases up to NumBlocks in the flash. The blocks
					can be listed in any order. 
					The function does not return until the blocks are erased. 
					If any blocks are protected or invalid none of the blocks are erased.
					During the Erase Cycle the Data Toggle Flow Chart of the Data Sheet is
					followed. The polling bit, DQ7, is not used.
	Pseudo Code:
			Step 1: Check for correct flash type
			Step 2: Check for protected or invalid blocks
			Step 3: Write Block Erase command
			Step 4: Check for time-out blocks
			Step 5: Wait for the timer bit to be set.
			Step 6: Follow Data Toggle Flow Chart until Program/Erase Controller has completed
			Step 7: Return to Read mode (if an error occurred)
*******************************************************************************/
int CfiFlashPageErase( unsigned char NumBlocks, unsigned char Block[] )
{
  unsigned char CurrentBlock; /* Range Variable to track current block */
  int iRetVal = CFI_FLASH_SUCCESS; /* Holds return value: optimistic initially! */
  unsigned int FirstRead, SecondRead; /* used to check toggle bit DQ2 */

  /* Step 1: Check for correct flash type */
  if( !(CfiFlashAutoSelect( CFI_FLASH_READ_MANUFACTURER ) == MANUFACTURER_ST)		|| 
      !(CfiFlashAutoSelect( CFI_FLASH_READ_DEVICE_CODE ) == EXPECTED_DEVICE ) )
    return CFI_FLASH_WRONG_TYPE;

  /* Step 2: Check for protected or invalid blocks. */
  if( NumBlocks > NUM_BLOCKS ) /* Check specified blocks <= NUM_BLOCKS */
    return CFI_FLASH_TOO_MANY_BLOCKS;
  for( CurrentBlock = 0; CurrentBlock < NumBlocks; CurrentBlock++ ) {
    /* Use FlashAutoSelect to find protected or invalid blocks */
    if( CfiFlashAutoSelect((func_t)Block[CurrentBlock]) != CFI_FLASH_BLOCK_UNPROTECTED )
      return (int)Block[CurrentBlock]; /* Return protected/invalid blocks */
  }

  /* Step 3: Write Block Erase command */
  CfiFlashWrite( 0x0555L, 0x00AA );
  CfiFlashWrite( 0x02AAL, 0x0055 );
  CfiFlashWrite( 0x0555L, 0x0080 );
  CfiFlashWrite( 0x0555L, 0x00AA );
  CfiFlashWrite( 0x02AAL, 0x0055 );

  /* DSI!: Time critical section. Additional blocks must be added every 50us */
  __disable_interrupt();
  for( CurrentBlock = 0; CurrentBlock < NumBlocks; CurrentBlock++ ) {
    CfiFlashWrite( BlockOffset[Block[CurrentBlock]], 0x0030 );
    /* Check for Erase Timeout Period (is bit DQ3 set?) */
    if( (CfiFlashRead( BlockOffset[Block[0]] ) & 0x0008) == 0x0008 )
      break; /* Cannot set any more blocks due to timeout */
  }
  __enable_interrupt();
  /* ENI! */

  /* Step 4: Check for time-out blocks */
  /* if timeout occurred then check if current block is erasing or not */
  /* Use DQ2 of status register, toggle implies block is erasing */
  if( CurrentBlock < NumBlocks ) {
    FirstRead = CfiFlashRead( BlockOffset[Block[CurrentBlock]] ) & 0x0004;
    SecondRead = CfiFlashRead( BlockOffset[Block[CurrentBlock]] ) & 0x0004;
    if( FirstRead != SecondRead ) {
      CurrentBlock++; /* Point to the next block */
    }
    if( CurrentBlock < NumBlocks ) {
      /* Indicate that some blocks have been timed out of the erase list */
      iRetVal = CFI_FLASH_MPU_TOO_SLOW;
    }
    /* Now specify all other blocks as not being erased */
    while( CurrentBlock < NumBlocks ) {
      Block[CurrentBlock++] = CFI_FLASH_BLOCK_NOT_ERASED;
    }
  }

  /* Step 5: Wait for the Erase Timer Bit (DQ3) to be set */
  /* TimeOut!: If, for some reason, the hardware fails then this	
     loop may not exit. Use a timer function to implement a timeout
     from the loop. */
  while( 1 ) {
    if( ( CfiFlashRead( BlockOffset[Block[0]] ) & 0x0008 ) == 0x0008 )
      break; /* Break when device starts the erase cycle */
  }

  /* Step 6: Follow Data Toggle Flow Chart until Program/Erase Controller completes */
  if( CfiFlashDataToggle() != CFI_FLASH_SUCCESS ) {
    iRetVal = CFI_FLASH_ERASE_FAIL;
	
    /* Step 7: Return to Read mode (if an error occurred) */
    CfiFlashReadReset();
  }
  return iRetVal;
}  /* end CfiFlashPageErase */


/*******************************************************************************
	Function: int CfiFlashEraseAll( unsigned char *ucResults )

	Arguments: ucResults is a pointer to an array where the results will bestored. 
			   If ucResults == NULL then no results are stored. Otherwise the
			   results are written to the array if an error occurs. The array is left
			   unchanged if the function returns FLASH_SUCCESS.
			   The errors written to the array are:
					FLASH_BLOCK_ERASED  if the block erased correctly
					FLASH_BLOCK_ERASE_FAILURE  if the block failed to erased

  Return Value: On success the function returns FLASH_SUCCESS (-1)
				  If a block is protected then the function returns 
				  the number of the block and no blocks are erased.
			      If the erase algorithms fails then the function returns FLASH_ERASE_FAIL (-2)
				  If the wrong type of flash is detected then FLASH_WRONG_TYPE (-8) is returned.

  Description: The function can be used to erase the whole flash chip so long as
				no blocks are protected. If any blocks are protected then nothing is erased.

  Pseudo Code:
		Step 1: Check for correct flash type
		Step 2: Check that all blocks are unprotected
		Step 3: Send Chip Erase Command
		Step 4: Follow Data Toggle Flow Chart until Program/Erase Controller has
				completed.
		Step 5: Check for blocks erased correctly
		Step 6: Return to Read mode (if an error occurred)
*******************************************************************************/
int CfiFlashEraseAll( unsigned char *ucResults )
{
  unsigned char CurrentBlock; /* Used to track the current block in a range */
  int iRetVal = CFI_FLASH_SUCCESS; /* Holds return value: optimistic initially! */

  /* Step 1: Check for correct flash type */
  if( !(CfiFlashAutoSelect( CFI_FLASH_READ_MANUFACTURER ) == MANUFACTURER_ST)
      || !(CfiFlashAutoSelect( CFI_FLASH_READ_DEVICE_CODE ) == EXPECTED_DEVICE ) )
    return CFI_FLASH_WRONG_TYPE;
	
  /* Step 2: Check that all blocks are unprotected */
  for( CurrentBlock = 0; CurrentBlock < NUM_BLOCKS; CurrentBlock++ ) {
      if( CfiFlashAutoSelect( (func_t)CurrentBlock ) != CFI_FLASH_BLOCK_UNPROTECTED )
	return (int)CurrentBlock; /* Return the first protected block */
    }

  /* Step 3: Send Chip Erase Command */
  CfiFlashWrite( 0x0555L, 0x00AA );
  CfiFlashWrite( 0x02AAL, 0x0055 );
  CfiFlashWrite( 0x0555L, 0x0080 );
  CfiFlashWrite( 0x0555L, 0x00AA );
  CfiFlashWrite( 0x02AAL, 0x0055 );
  CfiFlashWrite( 0x0555L, 0x0010 );

  /* Step 4: Follow Data Toggle Flow Chart until Program/Erase Controller has completed */
  if( CfiFlashDataToggle() != CFI_FLASH_SUCCESS ) {
      iRetVal = CFI_FLASH_ERASE_FAIL;
    }

  /* Step 5: Check for blocks erased correctly */
  if( iRetVal != CFI_FLASH_SUCCESS && ucResults != NULL ) {
      for( CurrentBlock = 0; CurrentBlock < NUM_BLOCKS; CurrentBlock++ ) {
	  if( CfiFlashBlockFailedErase( CurrentBlock ) == CFI_FLASH_BLOCK_FAILED_ERASE )
	    ucResults[CurrentBlock] = CFI_FLASH_BLOCK_ERASE_FAILURE;
	  else
	    ucResults[CurrentBlock] = CFI_FLASH_BLOCK_ERASED;
	}
    }

  /* Step 6: Return to Read mode (if an error occurred) */
  if( iRetVal != CFI_FLASH_SUCCESS )
    CfiFlashReadReset();
	
  return iRetVal;
} /* end CfiFlashEraseAll */

/*******************************************************************************
	Function: int CfiFlashProgram( unsigned long ulOff, size_t NumWords,void *Array )
			  
	Arguments: ulOff is the word offset into the flash to be programmed
			   NumWords holds the number of words in the array.
				Array is a pointer to the array to be programmed.

	Return Value: The function returns the following conditions:
					CFI_FLASH_SUCCESS 
					CFI_FLASH_PROGRAM_FAIL 
					CFI_FLASH_OFFSET_OUT_OF_RANGE 
					CFI_FLASH_WRONG_TYPE 
				Number of the first protected or invalid block
				On success the function returns FLASH_SUCCESS .
				The function returns FLASH_PROGRAM_FAIL  if a programming failure occurs.
				If the address range to be programmed exceeds the address range of the Flash
				Device the function returns FLASH_OFFSET_OUT_OF_RANGE  and nothing is
				programmed.
				If the wrong type of flash is detected then FLASH_WRONG_TYPE  is
				returned and nothing is programmed.
				If part of the address range to be programmed falls within a protected block,
				the function returns the number of the first protected block encountered and
				nothing is programmed.

	Description: This function is used to program an array into the flash. It does
		     not erase the flash first and may fail if the block(s) are not erased first.

	Pseudo Code:
		Step 1: Check for correct flash type
		Step 2: Check the offset range is valid
		Step 3: Check that the block(s) to be programmed are not protected
		Step 4: While there is more to be programmed
		Step 5: Check for changes from 0 to 1
		Step 6: Program the next word
		Step 7: Follow Data Toggle Flow Chart until Program/Erase Controller has completed
		Step 8: Return to Read Mode (if an error occurred)
		Step 9: Update pointers
*******************************************************************************/
int CfiFlashProgram( unsigned long ulOff, size_t NumWords, void *Array )
{
  unsigned int *uArrayPointer; /* Use an unsigned int to access the array */
  unsigned long ulLastOff; /* Holds the last offset to be programmed */
  unsigned char CurrentBlock; /* Range Variable to track current block */

  /* Step 1: Check for correct flash type */
  if( !(CfiFlashAutoSelect( CFI_FLASH_READ_MANUFACTURER ) == MANUFACTURER_ST)
      || !(CfiFlashAutoSelect( CFI_FLASH_READ_DEVICE_CODE ) == EXPECTED_DEVICE ) )
    return CFI_FLASH_WRONG_TYPE;

  /* Step 2: Check the offset and range are valid */
  ulLastOff = ulOff+NumWords-1;
  if( ulLastOff >= FLASH_SIZE )
    return CFI_FLASH_OFFSET_OUT_OF_RANGE;
	
  /* Step 3: Check that the block(s) to be programmed are not protected */
  for( CurrentBlock = 0; CurrentBlock < NUM_BLOCKS; CurrentBlock++ ) {
    /* If the address range to be programmed ends before this block */
    if( BlockOffset[CurrentBlock] > ulLastOff )
      break; /* then we are done */
    /* Else if the address range starts beyond this block */
    else if( (CurrentBlock < (NUM_BLOCKS-1)) && (ulOff >= BlockOffset[CurrentBlock+1]) )
      continue; /* then skip this block */
    /* Otherwise if this block is not unprotected */
    else if( CfiFlashAutoSelect((func_t)CurrentBlock) != CFI_FLASH_BLOCK_UNPROTECTED )
      return (int)CurrentBlock; /* Return first protected block */
  }

  /* Step 4: While there is more to be programmed */
  uArrayPointer = (unsigned int *)Array;
  while( ulOff <= ulLastOff ) {
    /* Step 5: Check for changes from 0 to 1 */
    if( ~CfiFlashRead( ulOff ) & * uArrayPointer )
      /* Indicate failure as it is not possible to change a 0 to a 1
	 using a Program command. This must be done using an Erase command */
      return CFI_FLASH_PROGRAM_FAIL;

    /* Step 6: Program the next word */
    CfiFlashWrite( 0x0555L, 0x00AA ); /* 1st cycle */
    CfiFlashWrite( 0x02AAL, 0x0055 ); /* 2nd cycle */
    CfiFlashWrite( 0x0555L, 0x00A0 ); /* Program command */
    CfiFlashWrite( ulOff, *uArrayPointer ); /* Program value */

    /* Step 7: Follow Data Toggle Flow Chart until Program/Erase Controller has completed */
    /* See Data Toggle Flow Chart of the Data Sheet */
    if( CfiFlashDataToggle() == CFI_FLASH_TOGGLE_FAIL ) {
      /* Step 8: Return to Read Mode (if an error occurred) */
      CfiFlashReadReset();
      return CFI_FLASH_PROGRAM_FAIL;
    }

    /* Step 9: Update pointers */
    ulOff++;
    uArrayPointer++;
  }
  return CFI_FLASH_SUCCESS;
} /* end CfiFlashProgram */




/*******************************************************************************
	Function: static int CfiFlashDataToggle( void )
	
	Arguments: none

	Return Value: The function returns FLASH_SUCCESS if the Program/Erase Controller
				  is successful or FLASH_TOGGLE_FAIL if there is a problem.

	Description: The function is used to monitor the Program/Erase Controller during
					erase or program operations. It returns when the Program/Erase Controller has
					completed. In the Data Sheets, the Data Toggle Flow Chart shows the operation
					of the function.
	Pseudo Code:
		Step 1: Read DQ6 (into a word)
		Step 2: Read DQ5 and DQ6 (into another word)
		Step 3: If DQ6 did not toggle between the two reads then return FLASH_SUCCESS
		Step 4: Else if DQ5 is zero then operation is not yet complete, goto 1
		Step 5: Else (DQ5 != 0), read DQ6 again
		Step 6: If DQ6 did not toggle between the last two reads then return
				FLASH_SUCCESS
		Step 7: Else return FLASH_TOGGLE_FAIL
*******************************************************************************/
static cfi_result_t CfiFlashDataToggle( void )
{
  unsigned int u1, u2; /* hold values read from any address offset within the Flash Memory */
  /* TimeOut!: If, for some reason, the hardware fails then this
     loop may not exit. Use a timer function to implement a timeout
     from the loop. 
  */
  while( 1 ) {
    /* Step 1: Read DQ6 (into a word) */
    u1 = CfiFlashRead( ANY_ADDR ); /* Read DQ6 from the Flash (any address) */

    /* Step 2: Read DQ5 and DQ6 (into another word) */
    u2 = CfiFlashRead( ANY_ADDR ); /* Read DQ5 and DQ6 from the Flash (any address) */

    /* Step 3: If DQ6 did not toggle between the two reads then return FLASH_SUCCESS */
    if( (u1&0x0040) == (u2&0x0040) ) /* DQ6 == NO Toggle */
      return CFI_FLASH_SUCCESS;

    /* Step 4: Else if DQ5 is zero then operation is not yet complete */
    if( (u2&0x0020) == 0x0000 )
      continue;

    /* Step 5: Else (DQ5 == 1), read DQ6 again */
    u1 = CfiFlashRead( ANY_ADDR ); /* Read DQ6 from the Flash (any address) */

    /* Step 6: If DQ6 did not toggle between the last two reads then return FLASH_SUCCESS */
    if( (u2&0x0040) == (u1&0x0040) ) /* DQ6 == NO Toggle */
      return CFI_FLASH_SUCCESS;

    /* Step 7: Else return FLASH_TOGGLE_FAIL */
    else /* DQ6 == Toggle here means fail */
      return CFI_FLASH_TOGGLE_FAIL;
  } /* end of while loop */
  
} /* end CfiFlashDataToggle */

/*******************************************************************************
	Function: int CfiFlashBlockFailedErase( unsigned char Block )
	
	Arguments: Block specifies the block to be checked
	
	Return Value:	CFI_FLASH_SUCCESS  if the block erased successfully
					CFI_FLASH_BLOCK_FAILED_ERASE  if the block failed to erase

	Description: This function can only be called after an erase operation which
				 has failed the FlashDataPoll() function. 
				 It must be called before the reset is made.
				 The function reads bit 2 of the Status Register to determine if the block	
				 has erased successfully or not. Successfully erased blocks should have DQ2
				 set to 1 following the erase. Failed blocks will have DQ2 toggle.
	Pseudo Code:
		Step 1: Read DQ2 in the block twice
		Step 2: If they are both the same then return FLASH_SUCCESS
		Step 3: Else return FLASH_BLOCK_FAILED_ERASE
*******************************************************************************/
static int CfiFlashBlockFailedErase( unsigned char Block )
{
  unsigned int FirstRead, SecondRead; /* Two variables used for clarity,
					 Optimiser will probably not use any */
	
  /* Step 1: Read block twice */
  FirstRead = CfiFlashRead( BlockOffset[Block] ) & 0x0004;
  SecondRead = CfiFlashRead( BlockOffset[Block] ) & 0x0004;

  /* Step 2: If they are the same return FLASH_SUCCESS */
  if( FirstRead == SecondRead )
    return CFI_FLASH_SUCCESS;

  /* Step 3: Else return FLASH_BLOCK_FAILED_ERASE */
  return CFI_FLASH_BLOCK_FAILED_ERASE;
} /* end CfiFlashBlockFailedErase */

/*******************************************************************************
	Function: void CfiFlashQueryGet(uint baseAddr, uint16 *ptr, uint size)
	
	Arguments: baseAddr  - base address of the Flash Part
		   ptr	     - Pointer to buffer to receive flash data
		   size      - Number of bytes to be read
	
	Return Value: none
	Description:  C I Information relative to the flash part are contained
                      in various sections ( see data sheet for details)
                      This routine reads a single section once the part is in query mode
                      section address buffer pointer and size are provided to 
                      read the specified data
	Pseudo Code:
                      read wordwise the size wide area of memory starting at baseAddr
		
*******************************************************************************/

void CfiFlashQueryGet(uint baseAddr, uint16 *ptr, uint size)
{
  int i = 0;
  uint32 addr ;
  for ( i = 0 ; i < size >> 1 ; i++) {
    addr = baseAddr + i;
    *ptr = CfiFlashRead(addr);
    ptr++;
  }
} /* end CfiFlashQueryGet */

/*******************************************************************************
	Function: int CfiFlashReadQuery(query_t query, uint16 *resultBuf)
	
	Arguments: query_t - a type of query corresponding to the info area ( ref data sheet)
                   resultBuf - a buffer to contain the section data 
	Return Value:	A return code indicating results 
	Description:  Read the specified Flash Information 
	Pseudo Code:
                      Step1 : Put the part in query mode 
                      Step2 : Read the specified section 
		
*******************************************************************************/

int CfiFlashReadQuery(query_t query, uint16 *resultBuf)
{
  int retVal = 0;
  uint16 *resPtr = resultBuf;

  CfiFlashWrite( 0x0055L, 0x0098 );
  
  switch (query) {
  case  CFI_QUERY_ID_STRING:
    CfiFlashQueryGet(CFI_QUERY_STRING_ADDR, resPtr, sizeof(CfiQueryIdString_t));
    break;
  case CFI_QUERY_SYS_INTERFACE_INFO:
    CfiFlashQueryGet( CFI_QUERY_SYS_INTERFACE_INFO_ADDR, resPtr, sizeof(CfiQuerySysInterfaceInfo_t));
    break;
  case CFI_QUERY_DEVICE_GEOMETRY:
    CfiFlashQueryGet( CFI_QUERY_DEVICE_GEOMETRY_ADDR, resPtr, sizeof(CfiQueryDeviceGeometryDefinition_t));
    break;
  case CFI_QUERY_PRIM_ALG_EXTENDED:
    CfiFlashQueryGet( CFI_QUERY_PRIM_ALG_EXTENDED_ADDR, resPtr, sizeof(CfiQueryPrimAlgExtended_t));
    break;
  case CFI_QUERY_SECURITY_CODE:
    CfiFlashQueryGet( CFI_QUERY_SECURITY_CODE_AREA_ADDR, resPtr, sizeof(CfiQuerySecurityCodeArea_t));
    break;
  }
  CfiFlashReadReset();
  return retVal;  
} /* end CfiFlashReadQuery */


/*******************************************************************************
	Function: int CfiFlashBlock(uint8 *address, uint32 numBlock,  uint8 *result)
	
	Arguments:  address - memory address where the array of block start
				numBlock - Number of blocks in the resulting array
				result  - a pointer of the array

	
	Return Value:	0 indicate succes
					-1 indicate a failure for address specified below the base
	Description: 
				using the block offsets build the array 
	Pseudo Code:
		
*******************************************************************************/
/* build a block array from address and size */
int CfiFlashBlock(uint8 *address, uint32 numBlock,  uint8 *result)
{
  int retVal = 0;
  int CurrentBlock;
  uint8 startBlock;
  int i;
  unsigned long offset;

#ifdef __IAR_SYSTEMS_ICC__
  uint8 *base = ( uint8 *) BASE_ADDR;
#endif

#ifdef __NSC_CRCC__
  FAR uint8 *base = ( FAR uint8 *) BASE_ADDR;
#endif  


  if ( address >= base) { 
    offset = address - base;
    offset >>= 1;
    for( CurrentBlock = 0; CurrentBlock <  NUM_BLOCKS; CurrentBlock++ ) {
      if ( offset <  BlockOffset[CurrentBlock] ) break;
    }
    startBlock = CurrentBlock -1;
    for ( i = 0 ; i < min(numBlock, NUM_BLOCKS) ; i++) {
      *result++ = startBlock++;
    }
  } else {
    retVal = -1;
  }
  return retVal;
} /* end CfiFlashBlock */

#endif


