/******************************************************************************
*  File:	codec.c
*  
*  This is the board level codec driver.
******************************************************************************/

#include "config.h"
#include "boarddef.h"
#include "drivers.h"
#include "boarddrv.h"


#define MAX_VOLUME_LEVEL 7
#define MID_VOLUME_LEVEL 4

#if (CODEC == MC145483)
/* This is the audio driver configuration for the Motorola MC145483 codec */
static const audio_config_t AudioConfig = {
	FALSE,	// Sync mode
	13,		// Mot codec sample size = 13 bits
	TRUE,	// Long frame sync
	13,     // Frame sync pulse length,
	FALSE,	// No inverted frame sync
	1,		// 1 slots per frame
	47,		// bit clock prescalar is 47 to generate a bit clock of 255319 Hz
			// out of 12MHz
	32,		// 60 bits per frame, generates a 7978 Hz frame clock.
	0x1,	// use slot 0 only for tx
	0x1,	// use slot 0 only for rx
	NOCVSDCONV
			// Use CVSD <--> linear PCM conversion
};
#endif

#if (CODEC == MSM7732)
/* This is the audio driver configuration for the OKI MSM7732 codec */
static const audio_config_t AudioConfig = {
	FALSE,	// Sync mode
	14,		// OKI codec sample size = 14 bits
	TRUE,	// Long frame sync
	14,     // Frame sync pulse length (leave it low for at least 1 clock),
	FALSE,	// No inverted frame sync
	1,		// 1 slots per frame
	15,	// bit clock prescalar is 100 to generate a bit clock of 120 KHz
			// out of 12MHz.
	100,		// 15 bits per frame, generates a 8 KHz frame clock.
	0x1,	// use slot 0 only for tx
	0x1,	// use slot 0 only for rx
	NOCVSDCONV
			// Use CVSD <--> linear PCM conversion
};
#endif

#if (CODEC == MSM7716)
/* This is the audio driver configuration for the OKI MSM7716 codec */
static const audio_config_t AudioConfig = {
	FALSE,	// Sync mode
	14,		// OKI codec sample size = 14 bits
	TRUE,	// Long frame sync
	14,     // Frame sync pulse length (leave it low for at least 1 clock),
	FALSE,	// No inverted frame sync
	1,		// 1 slots per frame
	100,	    // bit clock prescalar is 100 to generate a bit clock of 200 KHz
			// out of 12MHz.
	15,		// generates a 8 KHz frame clock.
	0x1,	// use slot 0 only for tx
	0x1,	// use slot 0 only for rx
	NOCVSDCONV
			// Use CVSD <--> linear PCM conversion 
};
#endif

static int volume_level = MID_VOLUME_LEVEL;

void codec_ctrl( unsigned char addr, unsigned char val );


#if (CODEC != NO_CODEC)


/******************************************************************************  
    Function: codec_init

	Description
	-----------
    This function initialized the codec driver.
  
	Arguments
	---------
	xmt_dl:
	A pointer to data link (circular buffer) control structure to be used for
	voice transmission (audio output).
	
	rcv_dl:
	A pointer to data link (circular buffer) control structure to be used for
	voice reception (audio input).

	rx_callback:
	A pointer to a callback function to be called whenever input voice
	samples are written to the receive data link circular buffer.

	conv:
	Indicates if and what type of conversion should be done by the (on chip) 
	conversion block on both transmit and recive paths. This optional conversion
	may complement the codec functionality when audio encoding formats other than
	codec's have to be used by the application.  Values are: 
	  * NOCVSDCONV (no CVSD conversion)
	  * CVSD2ULAW (CVSD <--> ulaw)
	  * CVSD2ALAW (CVSD <--> alaw)
	  * CVSD2LINEARPCM (CVSD <--> linear PCM). 
 *******************************************************************************/
void codec_init(dl_t * xmt_dl, dl_t *rcv_dl, void (*rx_callback_p)(unsigned int),
				audio_conv_t conv)
{
	audio_config_t audio_config;
	
#if (CODEC == MSM7732)
	/* Init codec to sensible values */
	codec_ctrl( 0, 0x41 ); // Enable output amp, linear mode in CR0
/*	codec_ctrl( 2, 0x00 ); // Eg for max volume in CR2 */
#endif

#if (CODEC == MSM7716)
	/* Init codec to sensible values */
	codec_ctrl( 0, 0xa0 ); // Max vol, MAIN input enabled, AUX0 output disabled (defaults)
#endif

	/* Copy default audio configuration structure to a local one. */
	audio_config = AudioConfig;

	/* Modify conversion value*/
	audio_config.audio_conv = conv;

	/* Modify conversion parameter */
	audio_init((audio_config_t *)&audio_config, xmt_dl, rcv_dl,
			rx_callback_p);


	/* Set volume to currently selected level. */	
//	codec_set_volume(volume_level);
}

/******************************************************************************  
    Function: codec_set_volume

	Description
	-----------
    This function sets the volume level on the codec output.
  
	Arguments
	---------
	volume:
	A number in the range of 0-7 where 0 means low volume and 7 means high
	volume.
*******************************************************************************/
void codec_set_volume(unsigned int volume)
{

#if (CODEC == MC145483)

	volume_level = volume & 0x7;
	/* The value we actually use as codec control bits is the complement of
	   volume level: 0 is maximal volume and 7 is minimal volume.
	*/

	audio_set_control_bits(3, MAX_VOLUME_LEVEL - volume_level);
#endif

#if (CODEC == MSM7732)
	volume_level = 2;//volume;
	codec_ctrl( 2, volume_level | (volume_level << 4)); // Set in and out the same
#endif

#if (CODEC == MSM7716)
	volume_level = volume;
	codec_ctrl( 0, MAX_VOLUME_LEVEL - volume_level); // TODO retain MSB settings
#endif

}

/******************************************************************************  
    Function: codec_adjust_volume

	Description
	-----------
    This function adjusts the volume level on the codec output.
  
	Arguments
	---------
	delta:
	A number to be added to the current volume level. This number can be
	positive (to increase volume) or negative (to decrease volume). In
	any case the volume level will remain in the range of 0-7.
*******************************************************************************/
void codec_adjust_volume(int delta)
{
	volume_level += delta;
	if (volume_level < 0)
		volume_level = 0;
	if (volume_level > MAX_VOLUME_LEVEL)
		volume_level = MAX_VOLUME_LEVEL;

#if (CODEC == MC145483)

	/* The value we actually use as codec control bits is the complement of
	   volume level: 0 is maximal volume and 7 is minimal volume.
	*/
	audio_set_control_bits(3, MAX_VOLUME_LEVEL - volume_level);
#endif

#if (CODEC == MSM7732)
	codec_ctrl( 2, volume_level | (volume_level << 4)); // Set in and out the same
#endif

#if (CODEC == MSM7716)
	codec_ctrl( 0, volume_level ); // TODO retain MSB settings
#endif
	

}


/******************************************************************************  
    Function: codec_ctrl

	Description
	-----------
    Send a control byte out to a codecs register
  
	Arguments
	---------
	Address of register and value to be set
 *******************************************************************************/
void codec_ctrl( unsigned char addr, unsigned char val )
{
#if (CODEC == MSM7732)
	unsigned char msmdata[2];
	pindesc_t pdesc; 

	pdesc.port = PORTF;
	pdesc.clkmsk = 0x01;	// PF0 -> CLK
	pdesc.datamsk = 0x04;	// PF2 -> DIO
	pdesc.enmsk = 0x08;		// PF3 -> DEN

	msmdata[0] = 0x00 | (addr << 4) | (val >> 4); // W/R=0, AAA=addr[2:0], B[7:4]
	msmdata[1] = (val << 4); // B[3:0]
	sibb_write( &pdesc, msmdata, 12 );
#endif

#if (CODEC == MSM7716)
	unsigned char msmdata[1];
	pindesc_t pdesc; 

	pdesc.port = PORTF;
	pdesc.clkmsk = 0x01;	// PF0 -> CLK
	pdesc.datamsk = 0x04;	// PF2 -> DIO
	pdesc.enmsk = 0x08;		// PF3 -> DEN

	msmdata[0] = 0xa0 | val; // Just one register in this one

	sibb_write( &pdesc, msmdata, 8 );
#endif
}

/******************************************************************************  
    Function: codec_start

	Description
	-----------
    This function starts codec.
  
	Arguments
	---------
	None.
 *******************************************************************************/
void codec_start(void)
{
 	audio_start();
}


 /******************************************************************************  
    Function: codec_stop

	Description
	-----------
    This function stops the codec.
  
	Arguments
	---------
	None.
 *******************************************************************************/
 void codec_stop(void)
 {
	audio_stop();
 }

#else

/* No codec in system - put stub functions for the no-codec case. */

void codec_init(dl_t * xmt_dl, dl_t *rcv_dl, void (*rx_callback_p)(unsigned int),
				audio_conv_t conv)
{
}

void codec_set_volume(unsigned int volume)
{
}

void codec_adjust_volume(int delta)
{
}

void codec_ctrl( unsigned char addr, unsigned char val )
{
}

void codec_start(void)
{
}

void codec_stop(void)
{
}

#endif

 


