/************************************************************************
*	File Name:	eeprom.c
*
*	Function:	This is a generic EEPROM driver. It provides read/write
*				services for various EEPROM parts. 
*
************************************************************************/ 

#include "config.h"
#include "boarddef.h"
#include "drivers.h"

#if (EEPROM != NO_EEPROM)

#define EEPROM_ACB_ADDR 0x50

#if (EEPROM == ATMEL_24C16)  
#  define EEPROM_MAX_WRITE_SIZE 16
#  define WAIT_TIME_AFTER_WRITE	2
#endif

#if (EEPROM == MICROCHIP_24LC65)  
#  define EEPROM_MAX_WRITE_SIZE 64
#  define WAIT_TIME_AFTER_WRITE	4
#endif

static void eep_page_write(unsigned short, unsigned char *, unsigned int);


/* Allow room for maximal EEPROM single data transfer + 2 bytes for an address */
unsigned char TransmitBuf[EEPROM_MAX_WRITE_SIZE+2];

/******************************************************************************  
    Function: eep_init

	Description
	-----------
    This function initializes the the communication with the EEPROM over the
	ACCESS.bus.

	Arguments
	---------
	None.
*******************************************************************************/
void eep_init(void)
{
	/* Acess Bus enable. Set Communication Rate to 100K. */
	acbm_init(0, 120, NULL);
}

/******************************************************************************  
    Function: meep_page_write

	Description
	-----------
    This function writes up to 16 bytes of data to the EEPROM all within
	one EEPROM page (it does not handle page boundary crossing).

	Arguments
	---------
	address:
	The address (offset) inside the EEPROM to which data should be written.

	buf:
	A pointer to a data buffer containing the data to be written to the EEPROM.

	len:
	The number of bytes to be written (must be <= 16).
*******************************************************************************/
static void eep_page_write(unsigned short address, unsigned char *buf, unsigned int len)
{
	unsigned int i, tx_buf_offset = 0;
	unsigned char slave_address;

#if (EEPROM == ATMEL_24C16)  
	/* 3 most significant bits of EEPROM address assigned to bits 0-2 of the slave
	   address.
	*/
	slave_address = EEPROM_ACB_ADDR | ((address>>8) & 0x7);
#endif

#if (EEPROM == MICROCHIP_24LC65)
	slave_address = EEPROM_ACB_ADDR;

	/* Insert most significant address bit to transmit buffer. */
	TransmitBuf[tx_buf_offset++] = (unsigned char) (address >> 8);
#endif

	/* Insert least significant address bits to transmit buffer. */
	TransmitBuf[tx_buf_offset++] = (unsigned char) address;

	/* Copy data to transmit buffer. */
	for (i = 0; i < len; i++)
		TransmitBuf[tx_buf_offset++] = buf[i];

	/* Execute the write transaction */
	acbm_tx(0, slave_address, TransmitBuf, tx_buf_offset, TRUE);

	/* Wait a period of time to let the EEPROM settle down after write. */
	timer_wait(WAIT_TIME_AFTER_WRITE); 
}

/******************************************************************************  
    Function: eep_buf_read

	Description
	-----------
    This function reads any numbers of data bytes the EEPROM.

	Arguments
	---------
	address:
	The address (offset) inside the EEPROM from which data should be read.

	buf:
	A pointer to a data buffer to which the EEPROM contents should be copied.

	len:
	The number of bytes to be read.
*******************************************************************************/
void eep_buf_read(unsigned short address, unsigned char *buf, unsigned int len)
{
	unsigned int tx_buf_offset = 0;
	unsigned char slave_address;

#if (EEPROM == ATMEL_24C16)  
	/* 3 most significant bits of EEPROM address assigned to bits 0-2 of the slave
	   address.
	*/
	slave_address = EEPROM_ACB_ADDR | ((address>>8) & 0x7);
#endif

#if (EEPROM == MICROCHIP_24LC65)
	slave_address = EEPROM_ACB_ADDR;

	/* Insert most significant address bit to transmit buffer. */
	TransmitBuf[tx_buf_offset++] = (unsigned char) (address >> 8);
#endif

	/* Insert least significant address bits to transmit buffer. */
	TransmitBuf[tx_buf_offset++] = (unsigned char) address;

	/* Execute a dummy write operation just to set EEPROM address counter 
	   This means there should be NO stop condition at the end*/
	acbm_tx(0, slave_address, TransmitBuf, tx_buf_offset, FALSE);

	/* Execute the write transaction */
	acbm_rx(0, slave_address, buf, len, TRUE);
}

/******************************************************************************  
    Function: eep_buf_write

	Description
	-----------
    This function writes any number of data bytes to the EEPROM.

	Arguments
	---------
	address:
	The address (offset) inside the EEPROM to which data should be written.

	buf:
	A pointer to a data buffer containing the data to be written to the EEPROM.

	len:
	The number of bytes to be written.
*******************************************************************************/
void eep_buf_write(unsigned short address, unsigned char *buf, unsigned int len)
{
	unsigned int size, max_write;

	while (len > 0) {
		max_write = EEPROM_MAX_WRITE_SIZE - (address & (EEPROM_MAX_WRITE_SIZE-1));
		if (len >= max_write)
			size = max_write;
		else
			size = len;
		eep_page_write(address, buf, size);
		len -= size;
		address += size;
		buf+= size;
	}	
}

/******************************************************************************  
    Function: eep_total_size

	Description
	-----------
    This function returns the total size of the EEPROM device.

	Arguments
	---------
	None.

	Returns
	-------
	The EEPROM size in bytes.
*******************************************************************************/
unsigned short eep_total_size(void)
{
#if (EEPROM == ATMEL_24C16)
	return 2048;
#endif
#if (EEPROM == MICROCHIP_24LC65)
	return 16384;
#endif
}

#else

/* Some boards don't have an EEPROM. Include empty function for those. */
void eep_init(void)
{
}

void eep_buf_read(unsigned short address, unsigned char *buf, unsigned int len)
{
}

void eep_buf_write(unsigned short address, unsigned char *buf, unsigned int len)
{
}

unsigned short eep_total_size(void)
{
	return 0;
}

#endif
