/************************************************************************
	File Name : extuart.c

	Function:	This is a driver for an external UART.

************************************************************************/

#include "config.h"
#include "boarddef.h"
#include "boarddrv.h"


#if (EXTUART == NS16552)

volatile uart_t * const uart = (uart_t *) UART_ADDR;

static const unsigned char bitrate_reg_table[] = 
{
	DIVISOR_9600,
	DIVISOR_38400,
	DIVISOR_115200,
	DIVISOR_921600
};

/******************************************************************************  
    Function: extuart_init

	Description
	-----------
    This function initializes the external UART hardware and the its driver.
    After returning from this function the UART is ready to transmit and receive
	data.

	Arguments
	---------
	bitrate:
	The required communication bit rate. Check the bitrate_t type for supported
	bit rates.
*******************************************************************************/

void extuart_init(bitrate_t bitrate)
{
	unsigned char divisor;

	/* Set the clock divisor according to the required bit rate */
	divisor = bitrate_reg_table[bitrate];
	LCR = 0x80;
	DLL = divisor & 0xff;
	DLM = divisor >> 8;

	LCR = 0x3; /* 8 bits no parity 1 stop bit */

	FCR = 0x1;
	FCR = 0x7; /* Enable FIFO */

	//IER = 0x1; /* Enable interrupts on receive*/

	MCR = 0xa; /* Enable interrupts ??? and force RTS */
}


/******************************************************************************  
    Function: extuart_tx

	Description
	-----------
    This function transmits a specified amount (or less) of bytes
	starting from a specified memory location.

	Arguments
	---------
	data:
	A pointer to the start address of the data to be transmitted.

	size:
	The number of bytes to be transmitted.
	
	Return value
	------------
	The number of bytes that were actually transmitted. Could be less than
	the requested number.
*******************************************************************************/
unsigned int extuart_tx(unsigned char *data, unsigned int size)
{
	unsigned int count = size;

	while(count--) {
		while(!(LSR & 0x20)); /* wait until tx buffer is empty */
		THR = *data++;
	}
	return size;
}

/******************************************************************************  
    Function: extuart_rx

	Description
	-----------
    This function reads a specified amount (or less) of received bytes
	into a specified memory location.

	Arguments
	---------
	data:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be read.
	
	Return value
	------------
	The number of bytes that were actually read. Could be less than
	the requested number.
*******************************************************************************/
unsigned int extuart_rx(unsigned char *outbuf, unsigned int size)
{
  unsigned int rx_count = 0;

  while ((rx_count < size) && (LSR & 0x1)) {
    /* Data ready in receive buffer*/
	  *outbuf++ = RHR;
	  rx_count++; 
  }
  return rx_count;
}

/******************************************************************************  
    Function: extuart_rx_count

	Description
	-----------
    This function returns the number of bytes that can be currently read from
	the 16552 driver

	Arguments
	---------
	None.
	
	Return value
	------------
	The number of bytes that can be currently read from the USART driver.

	!--Note: with the current implementation this function can only return
	0 (in case no bytes are available) or 1 (in case any larger than zero
	number of bytes are available).
*******************************************************************************/
unsigned int extuart_rx_count(void)
{
  return((LSR & 0x1));	
}

#endif

#if (EXTUART == NO_EXTUART)

/* Put stub functions for the no-UART case. */


void extuart_init(bitrate_t bitrate)
{
}

unsigned int extuart_tx(unsigned char *data, unsigned int size)
{
	return 0;
}

unsigned int extuart_rx(unsigned char *outbuf, unsigned int size)
{
	return 0;
}

unsigned int extuart_rx_count(void)
{
	return 0;
}

#endif
