/******************************************************************************
*  File:	led.c
*  
* This is a generic LED driver. Provides several service functions for the
* binary LEDS and well as the seven segment LED.
******************************************************************************/

#include "config.h"
#include "boarddef.h"
#include "drivers.h"

void led_set_all(unsigned short mask);


#if (LED_TYPE != NO_LED)

#if (LED_TYPE == LED_MEM_MAPPED_WO)
static unsigned short led_shadow;
#endif

/******************************************************************************  
    Function: led_init

	Description
	-----------
    This function initializes the led hardware and clears all LEDs.

	Arguments
	---------
	none.
*******************************************************************************/

void led_init(void)
{

#if (LED_TYPE == LED_GPIO)
	/* Configure LED GPIO pins as outputs. */
	gpio_set_dir(LED_PORT, LED_MASK, GPIO_OUTPUT);
#endif

	/* Clear LEDs */
	led_set_all(0);
}

/******************************************************************************  
    Function: led_set_all

	Description
	-----------
    This function sets all LEDs in one shot.

	Arguments
	---------
	mask:
	A bit mask representing the binary values of all LEDs:
	A set bit indicates LED on. A cleared bit indicates LED off.
*******************************************************************************/
void led_set_all(unsigned short mask)
{
	unsigned short tmp;

	mask <<= LED_REG_OFFSET;
	if (!LED_POLARITY)
		mask ^= 0xffff;
	mask &= LED_MASK;
	
#if (LED_TYPE == LED_GPIO)
	gpio_write(LED_PORT, LED_MASK, mask);
#endif 

#if (LED_TYPE == LED_MEM_MAPPED_RW)
	tmp = LED_REG & (~(unsigned short)LED_MASK);
	LED_REG = tmp | mask;
#endif

#if (LED_TYPE == LED_MEM_MAPPED_WO)
	tmp = led_shadow & (~(unsigned short)LED_MASK);
	led_shadow = tmp | mask; 
	LED_REG = led_shadow;
#endif
}

/******************************************************************************  
    Function: led_set

	Description
	-----------
    This function sets turns on one LED.

	Arguments
	---------
	position:
	Specifies which LED should be turned on.
*******************************************************************************/
void led_set(unsigned int position)
{
	unsigned short tmp, mask, write_val;

	mask = 1 << (position + LED_REG_OFFSET);
	write_val =  (LED_POLARITY ? mask : 0);
	mask &= LED_MASK;

#if (LED_TYPE == LED_GPIO)
	gpio_write(LED_PORT, mask, write_val);
#endif

#if (LED_TYPE == LED_MEM_MAPPED_RW)
	tmp = LED_REG & (~mask);
	LED_REG = tmp | write_val;
#endif

#if (LED_TYPE == LED_MEM_MAPPED_WO)
	tmp = led_shadow & (~mask);
	led_shadow = tmp | write_val; 
	LED_REG = led_shadow;
#endif
}

/******************************************************************************  
    Function: led_clear

	Description
	-----------
    This function sets turns off one LED.

	Arguments
	---------
	position:
	Specifies which LED should be turned off.
*******************************************************************************/
void led_clear(unsigned int position)
{
	unsigned short tmp, mask, write_val;

	mask = 1 << (position + LED_REG_OFFSET);
	write_val =  (LED_POLARITY ? 0 : mask);
	mask &= LED_MASK;

#if (LED_TYPE == LED_GPIO)
	gpio_write(LED_PORT, mask, write_val);
#endif

#if (LED_TYPE == LED_MEM_MAPPED_RW)
	tmp = LED_REG & (~mask);
	LED_REG = tmp | write_val;
#endif

#if (LED_TYPE == LED_MEM_MAPPED_WO)
	tmp = led_shadow & (~mask);
	led_shadow = tmp | write_val; 
	LED_REG = led_shadow;
#endif
}

/******************************************************************************  
    Function: led_toggle

	Description
	-----------
    This function sets toggles one LED.

	Arguments
	---------
	position:
	Specifies which LED should be toggled.
*******************************************************************************/
void led_toggle(unsigned int position)
{
	unsigned short mask;

	mask = 1 << (position + LED_REG_OFFSET);
	mask &= LED_MASK;

#if (LED_TYPE == LED_GPIO)
	gpio_write(LED_PORT, mask, (gpio_read(LED_PORT, mask)) ^ mask);
#endif

#if (LED_TYPE == LED_MEM_MAPPED_RW)
	LED_REG ^= mask;
#endif

#if (LED_TYPE == LED_MEM_MAPPED_WO)
	led_shadow ^= mask; 
	LED_REG = led_shadow;
#endif
}


#else /* NO LEDs */

void led_init(void) {}
void led_set_all(unsigned short mask) {}
void led_set(unsigned int position) {}
void led_clear(unsigned int position) {}
void led_toggle(unsigned int position) {}

#endif

#if (LED7SEG_TYPE != NO_LED7SEG)

#ifndef LCD7SEG_DIRECT
/* This table maps hexadecimal digits to their 7-segment LED representation.
   Need to complete the A-F in the table.
*/

const unsigned char led_digits[] = {
	0x40, 0x79, 0x24, 0x30, 0x19, 0x12, 0x03, 0x78,
	0x00, 0x18, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};
#endif

#if (LED7SEG_TYPE == LED7SEG_MEM_MAPPED_WO)
static unsigned short led7seg_shadow;
#endif


/******************************************************************************  
    Function: led_7seg_set

	Description
	-----------
    This function displays a hexadecimal digit on the seven segment LED.

	Arguments
	---------
	digit:
	An 4-bit intger in the range 0-15 to be displayed in hex format on the seven
	segment LED.
*******************************************************************************/
void led_7seg_set(unsigned char digit)
{
	unsigned short tmp, val;

#ifdef LED7SEG_DIRECT
	val = (unsigned short)digit;
#else
	val = (unsigned short)led_digits[digit];
#endif
	val <<= LED7SEG_REG_OFFSET;
	if (!LED7SEG_POLARITY)
		val ^= 0xffff;
	val &= LED7SEG_MASK;

#if (LED7SEG_TYPE == LED7SEG_GPIO || LED7SEG_TYPE == LED7SEG_MEM_MAPPED_RW)
	tmp = LED7SEG_REG & ((unsigned short)~LED7SEG_MASK);
	LED7SEG_REG = tmp | val;
#endif

#if (LED7SEG_TYPE == LED7SEG_MEM_MAPPED_RO)
	tmp = LED7SEG_REG & (~LED7SEG_MASK);
	led7seg_shadow = tmp | val;
	LED7SEG_REG = led7seg_shadow;
#endif
}

#else  /* No 7-segment LED. */

void led_7seg_set(unsigned char digit)
{
}

#endif

