/******************************************************************************
*  File:	acbm.c
*  
*  This is ACCESS.bus peripheral driver for a master device.
******************************************************************************/

#include "chipdef.h"

#define ACB_READ 1
#define ACB_WRITE 0

/* The following structure captures information on the state of a particular
   Access.bus block.
*/
typedef struct {
	volatile boolean tx_in_progress;
	volatile boolean rx_in_progress;
	boolean start;
	boolean stop;
	unsigned char *tr_buf;
	unsigned char tr_address;
	unsigned int tr_count;
	void (*acb_callback_p)(void);
} acbstate_t;

static acbstate_t acbs_state[N_ACBS];

/* Access.bus base address table */
static acb_t * const acb_tab[] = {
	ACB_BASE_ADDRESSES
};

/* Table of IRQs for each of the Access.bus blocks */
static unsigned char const acb_irqs[] = {
	ACB_IRQS
};

static void acb_isr(unsigned int);
static void tx_end(unsigned int);
static void rx_end(unsigned int);

/******************************************************************************  
    Function: acbm_init

	Description
	-----------
    This function initializes the ACCESS.bus hardware and the ACCESS.bus driver.
  
	Arguments
	---------
	acb_num:
	The number of the Access.bus block to be initialized (some chips have
	multiple Access.bus block).

	divisor:
	The ratio between the system clock frequency and the desired ACCESS.bus
	clock frequency.  For accuracy this number should be a multiple of 4. It
	also most be in the range of 32 to 2044.

	callback_p:
	A pointer to a callback function that will be called once a transmit 
	or receive transaction is completed or aborted.
*******************************************************************************/

void acbm_init(unsigned int acb_num, unsigned short divisor, void (*callback_p)(void))
{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	/* Fix the divisor value if needed. */
	if (divisor < 32) divisor = 32;
	if (divisor > 2044) divisor = 2044;

	/* The actual value of the SCLFRQ field in the ACBTL2/ACBTL3 registers
	   is 4 times smaller than the actual frequency ratio. Therefore divide
	   original value by 4.
	*/
	divisor >>= 2;

	/* Save the callback function pointer. */
	acbstate->acb_callback_p = callback_p;

	/* Shift left by 1 since the SCLFRQ field starts at bit 1 of ACBTL2. */
	ACBCTL2 = divisor << 1;

	/* Write the most significat bits of SCLFRQ to ACBTL3.*/
	ACBCTL3 = divisor >> 7;

	/* Enable the ACCESS.bus block. */
	ACBCTL2 |= ENABLE;


	/* Enable ACCESS.bus interrupts. */
	//ACBCTL1 |= INTEN;

	/* Unmask the ACCESS.bus interrupt. */
	icu_unmask(acb_irqs[acb_num]);
}

/******************************************************************************  
    Function: acbm_tx

	Description
	-----------
    This function transmits data to a specified slave device.
  
	Arguments
	---------
	acb_num:
	The number of the Access.bus block on which this operation should be
	performed.

	address:
	The ACCESS.bus address of the slave device to be addressed.
	 
	data:
	A pointer to a memory buffer where the data to be transmitted is stored.
	
	size:
	The size of the data to be transmitted in bytes.

	stop:
	A boolean flag indicating whether a STOP condition should be generated
	at the end of data transmission.
*******************************************************************************/
void acbm_tx(unsigned int acb_num, unsigned char address, unsigned char *data,
				unsigned short size, boolean stop_flag)
{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	/* Save the address, data pointer, number of bytes and stop flag. */
	acbstate->tr_buf = data;
	acbstate->tr_count = size;
	acbstate->tr_address = address;
	acbstate->stop = stop_flag;

	acbstate->tx_in_progress = TRUE;
	acbstate->start = TRUE;

	ACBCTL1 |= START;
	ACBCTL1 |= INTEN;

	/* If there is no callback function - wait until transmission is complete. */
	if (acbstate->acb_callback_p == NULL)
		while(acbstate->tx_in_progress);

}

/******************************************************************************  
    Function: acbm_rx

	Description
	-----------
    This function receives data from a specified slave device.
  
	Arguments
	---------
	acb_num:
	The number of the Access.bus block on which this operation should be
	performed.

	address:
	The ACCESS.bus address of the slave device to be addressed.
	 
	inbuf:
	A pointer to a memory buffer where the data to be transmitted is stored.
	
	size:
	The size of the data to be transmitted in bytes.

	stop:
	A boolean flag indicating whether a STOP condition should be generated
	at the end of data transmission.
*******************************************************************************/
void acbm_rx(unsigned int acb_num, unsigned char address, unsigned char *inbuf,
				unsigned short size, boolean stop_flag)
{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	/* Save the address, data pointer, number of bytes and stop flag. */
	acbstate->tr_buf = inbuf;
	acbstate->tr_count = size;
	acbstate->tr_address = address;
	acbstate->stop = stop_flag;

	acbstate->rx_in_progress = TRUE;
	acbstate->start = TRUE;

	ACBCTL1 |= START;
	ACBCTL1 |= INTEN;

	/* If there is no callback function - wait until reception is complete. */
	if (acbstate->acb_callback_p == NULL)
		while(acbstate->rx_in_progress);
}

/******************************************************************************  
    Function: acb_isr

	Description
	-----------
    This function is the core of the ACCESS.bus interrupt service routine.
	It is called from the top-level interrupt handler of one one of the
	Access.bus block and handles the interrupt for that particular block.

	Arguments
	---------
	acb_num:
	The number of the Access.bus block on which this operation should be
	performed.
	
	Return value
	------------
	None.
*******************************************************************************/
static void acb_isr(unsigned int acb_num)

{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	if(acbstate->tx_in_progress) {
		if (acbstate->start) {
			if ((ACBST & (MASTER | SDAST | BER)) == (MASTER | SDAST)) {
				ACBSDA = (acbstate->tr_address << 1) | ACB_WRITE;
				acbstate->start = FALSE;
			}
			else
			   tx_end(acb_num);
		}
		else {
			if ((ACBST & (NEGACK | BER | SDAST)) == SDAST) {
				if (acbstate->tr_count) {
					ACBSDA = *(acbstate->tr_buf)++;
					acbstate->tr_count--;
				}
				else {
					tx_end(acb_num);
					if (acbstate->stop)
						ACBCTL1 |= STOP;
					//else
					//	ACBCTL1 &= ~INTEN;
				}
			}
			else
				tx_end(acb_num);
		}

	}

	if (acbstate->rx_in_progress) {
		if (acbstate->start) {
			if ((ACBST & (MASTER | SDAST | BER)) == (MASTER | SDAST)) {
				if (acbstate->tr_count == 1)
					ACBCTL1 |= ACK;
				ACBSDA = (acbstate->tr_address << 1) | ACB_READ;
			}
			else
				rx_end(acb_num);
			acbstate->start = FALSE;
			return;
		}

		if (ACBST & STASTR) {
			ACBST = STASTR;
			return;
		}
		if ((ACBST & (BER | SDAST)) == SDAST) {
			if (acbstate->tr_count) {
				if (acbstate->tr_count == 2)
					ACBCTL1 |= ACK;
				*(acbstate->tr_buf)++ = ACBSDA;
				acbstate->tr_count--;
				if (acbstate->tr_count == 0) {
					rx_end(acb_num);
					if (acbstate->stop)
						ACBCTL1 |= STOP;
				}
			}
			else
				panic();
		}
		else
			rx_end(acb_num);	
	}

}

/******************************************************************************  
    Function: acb0_isr

	Description
	-----------
    This function is the ACCESS.bus interrupt service routine for
    Acess.bus 0.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef ACB0_IRQ

#ifdef __NSC_CRCC__
#pragma interrupt (acb0_isr)
void acb0_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+ACB0_IRQ)
__interrupt void acb0_isr(void)
#endif
{
	acb_isr(0);
}

#endif

/******************************************************************************  
    Function: acb1_isr

	Description
	-----------
    This function is the ACCESS.bus interrupt service routine for
    Acess.bus 1.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef ACB1_IRQ

#ifdef __NSC_CRCC__
#pragma interrupt (acb1_isr)
void acb1_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+ACB1_IRQ)
__interrupt void acb1_isr(void)
#endif
{
	acb_isr(1);
}

#endif


/******************************************************************************  
    Function: tx_end

	Description
	-----------
    This function is called upon and end of and Access.bus transmit sequence
    and takes the necessary actions for this condistion.

	Arguments
	---------
	acb_num:
	The number of the Access.bus block on which this operation should be
	performed.
	
	Return value
	------------
	None.
*******************************************************************************/
static void tx_end(unsigned int acb_num)
{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	acbstate->tx_in_progress = FALSE;
	ACBCTL1 &= ~INTEN;	
	if (acbstate->acb_callback_p != NULL)
		(*(acbstate->acb_callback_p))();
}

/******************************************************************************  
    Function: rx_end

	Description
	-----------
    This function is called upon and end of and Access.bus receive sequence
    and takes the necessary actions for this condistion.

	Arguments
	---------
	acb_num:
	The number of the Access.bus block on which this operation should be
	performed.
	
	Return value
	------------
	None.
*******************************************************************************/
static void rx_end(unsigned int acb_num)
{
	acb_t * acb = acb_tab[acb_num];
	acbstate_t *acbstate = &acbs_state[acb_num];

	acbstate->rx_in_progress = FALSE;
	ACBCTL1 &= ~INTEN;	
	if (acbstate->acb_callback_p != NULL)
		(*(acbstate->acb_callback_p))();
}
