/******************************************************************************
*  File:	acbs.c
*  
*  This is ACCESS.bus peripheral driver for a slave device.
******************************************************************************/

#include "chipdef.h"
#ifdef UCOS
#  include "ucos.h"
#endif

#define TX_BUF_SIZE 64
#define RX_BUF_SIZE 64

acb_t * const acb = (acb_t *)ACB0_BASE_ADDR;

static unsigned char tx_buf[TX_BUF_SIZE];
static unsigned char rx_buf[RX_BUF_SIZE];

static unsigned char *tx_ptr, *rx_ptr,*client_tx_ptr, *client_rx_ptr;
static dl_t tx_dl_s, rx_dl_s;
static dl_t *rx_dl, *tx_dl;

/******************************************************************************  
    Function: acbs_init

	Description
	-----------
    This function initializes the ACCESS.bus hardware and the ACCESS.bus driver.
  
	Arguments
	---------
	
	address:
	The 7-bit ACCESS.bus address of the local device. 

*******************************************************************************/

void acbs_init(unsigned char address)
{
	/* Initialize data links (circular buffers) and data link pointers */
	tx_dl = &tx_dl_s;
	rx_dl = &rx_dl_s;       
	dl_init(tx_dl, tx_buf, TX_BUF_SIZE);
	dl_init(rx_dl, rx_buf, RX_BUF_SIZE);
	dl_init_ptr(rx_dl, rx_ptr);
	dl_init_ptr(rx_dl, client_rx_ptr);
	dl_init_ptr(tx_dl, tx_ptr);
	dl_init_ptr(tx_dl, client_tx_ptr);

	/* Enable the ACCESS.bus peripheral. */
	ACBCTL2 |= ENABLE;

	/* Set the slave address and enable slave mode. */
	ACBADDR = address |	SAEN;
	/* Enable interrupts on address match. */
	ACBCTL1 |= NMINTE;

	/* Enable ACCESS.bus interrupts. */
	ACBCTL1 |= INTEN;

	/* Unmask the ACCESS.bus interrupt. */
	icu_unmask(ACB0_IRQ);
}

/******************************************************************************  
    Function: acbs_tx

	Description
	-----------
    This function deposits data to be transmitted by the slave the next time
    the master requests the slave to transmit.
  
	Arguments
	---------
	data:
	A pointer to a memory buffer where the data to be transmitted is stored.
	
	size:
	The size of the data to be deposited in bytes. Note: when the master ask
	the slave to transmit it is up to the master to decide how many bytes it
	wants to pull. The application is responsible to deposit enough data so
	that the master will not starve. The amount of data needed by the master
	is application dependent.

	Return value
	------------
	The number of bytes that were actually transmitted. Could be less than
	the requested number.

*******************************************************************************/
unsigned int acbs_tx(unsigned char *data, unsigned short size)
{
	unsigned int i, avail;
	do {
		avail = dl_p_space_available(tx_dl);
		if (size > avail)
#ifdef UCOS
			OSTimeDly(1);
#else
			size = avail;
#endif
	} while (size > avail);

	for (i = 0; i < size; i++) {
		*client_tx_ptr = *data;
		dl_advance(tx_dl, client_tx_ptr);
		data++;
	}

	dl_p_commit(tx_dl, size);

	return size;
}

/******************************************************************************  
    Function: acbs_rx

	Description
	-----------
    This function reads a specified amount (or less) of received bytes
	into a specified memory location.

	Arguments
	---------
	inbuf:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be read.
	
	Return value
	------------
	The number of bytes that were actually read. Could be less than
	the requested number.
*******************************************************************************/
unsigned int acbs_rx(unsigned char *inbuf, unsigned short size)
{
  unsigned int i, avail;

  avail = dl_cnt(rx_dl);
  if (size > avail)
	size = avail;

  for (i = 0; i < size; i++) {
	*inbuf = *client_rx_ptr;
	dl_advance(rx_dl, client_rx_ptr);
	inbuf++;
  }

  dl_c_commit(rx_dl, size);

  
  return size;
}

/******************************************************************************  
    Function: acbs_rx_count

	Description
	-----------
    This function returns the number of bytes that can be currently read from
	the ACCESS.bus slave driver.

	Arguments
	---------
	None.
	
	Return value
	------------
	The number of bytes that can be currently read from the USART driver.
*******************************************************************************/
unsigned int acbs_rx_count(void)
{
  return(dl_cnt(rx_dl));	
}


/******************************************************************************  
    Function: acb_isr

	Description
	-----------
    This function is the ACCESS.bus slave interrupt service routine.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef __NSC_CRCC__
#pragma interrupt (acb_isr)
void acb_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+ACB0_IRQ)
__interrupt void acb_isr(void)
#else
__interrupt[(16+ACB0_IRQ)*IDT_WIDTH] void acb_isr(void)
#endif
#endif
{
	if (ACBST & BER) {
		/* Bus error - clear this condition and return. */
		ACBST = BER;
		return;
	}

	if (ACBST & NMATCH) {
		/* Address match - this means we are being addressed by the master.
		   Clear the NMATCH bit for future use.
		*/
		ACBST = NMATCH;

		/* Make sure the NEGACK bit is clear. */
		ACBST = NEGACK;

		/* If trasmit - attempt to send transmit first byte immediately. */
		if ((ACBST & (SDAST | XMIT)) == (SDAST | XMIT) && dl_cnt(tx_dl)) {
			ACBSDA = *tx_ptr;
			dl_c_advance_commit(tx_dl, tx_ptr, 1);
		}

		return;
	}

	if (ACBST & NEGACK) {
		/* Negative acknowledgement - master does not want to receive
		   anymore data. Clear condition and return.
		*/
		ACBST = NEGACK;
		return;
	}

	if (ACBST & XMIT) {
		/* If there's data to transmit and ACCESS.bus is ready to transmit -
		   send next byte.
		*/
		if (dl_cnt(tx_dl) && (ACBST & SDAST)) {
			/* More data is available and ACB buffer is ready - send next byte */
			ACBSDA = *tx_ptr;
			dl_c_advance_commit(tx_dl, tx_ptr, 1);
		}

	} else {
		/* Receive mode - if data is waiting then receive next byte. */
		if (ACBST & SDAST) {
			if (dl_p_space_available(rx_dl) <= 2)
				/* Only one byte available in receive buffer - send negative ACK on
				   next acknowledge cycle.
				*/
				ACBCTL1 |= ACK;
			/* Receive byte. */
			*rx_ptr = ACBSDA;
			/* Update data link info. */
			dl_p_advance_commit(rx_dl, rx_ptr, 1);
		}
		else
			if (ACBST & SLVSTP)
				ACBST = SLVSTP;
	}
}
