/******************************************************************************
*  File:	adc.c
*  
*  This is the Analog to Digital Converter (ADC) driver.
******************************************************************************/

#include "chipdef.h"
#ifdef UCOS
#include "ucos.h"
#endif

/* Generate code only if the peripheral exists on the chip */
#ifdef ADC_BASE_ADDR

adc_t * const adc = (adc_t *)ADC_BASE_ADDR;

static void (* adc_cb_p)(volatile short *Result);
static boolean TouchScreenMode = FALSE;
static volatile short AdcResult;
static volatile boolean ConvInProgress = FALSE;

/*******************************************************************************

	Function: AdcInit

	Description
	-----------
    This function initializes the ADC peripheral and the ADC driver.

	Arguments
	---------
	boolean Diff:
	This indicates if the ADC is to be operated in Differential or Single
	Ended Mode. TRUE means that the ADC is in Differential Mode and FALSE
	means that the ADC is in Single Ended Mode.

	Channel:
	This parameter specifies which Channel of the ADC is being used. If the 
	ADC is being used in Differential Mode then this will select which channel 
	pair is being used based on the spec.

	PRef:
	This parameter specifies the positive Reference source. The values that it is
	allowed to have are: PREF_CFG_INTERNAL, PREF_CFG_VREFP, PREF_CFG_ADC0, PREF_CFG_ADC1.

	NRef:
	This parameter specifies the negative Reference source. The values that it is
	allowed to have are: NREF_CFG_INTERNAL, NREF_CFG_ADC2, NREF_CFG_ADC3.


	ExternalWakeup:
	This paramter specifies if the Analog to Digital Conversion is being initiated
	from an External Trigger or not. A value of TRUE means that the conversion is being 
	intiated from an external signal.

	Repeat:
	This boolean parameter indicates whether the ADC block in External Mode and when this 
	parameter is TRUE, an analog to digital conversion is initiated on every ASYNC trigger, 
	with no CPU intervention.

	TouchScreen:
	This parameter specifies if the ADC is being used in a Touch Screen application
	or not. If it is being used in a Touch Screen application, then the ADC block is
	configured to be in Pen Down Detection mode at the end of the initialization.

	tick_callback_p:
	A pointer to a function that will be called on every time a result is available.
	If this is not used (set to NULL), then the ADC interrupt will not be generated
	at the end of every conversion. The application can get the result using the 
	blocking function AdcReadResult().

*************************************************************************************/


void AdcInit(boolean Diff, unsigned char Channel, unsigned short PRef,unsigned short NRef,
				boolean ExternalWakeup, boolean TouchScreen, boolean Repeat,
				void (*adc_result_callback_p)(volatile short *Result))
{
	//Wait for any existing conversion to finish
	while(ConvInProgress)
#ifdef UCOS
		OSTimeDly(1);
#else
	;
#endif

	//If Diff then ADC in differential Mode else in Single Ended Mode
	if (Diff)
	{
		ADCGCR |= DIFF;
		//DifferentialMode = TRUE;
	}
	else
		ADCGCR &= ~DIFF;

	
	//Setup the Reference Voltages
	ADCGCR |= PRef;
	ADCGCR |= NRef;

	//If ExternalWakeup is 0 then ADC has to configured to be started by software
	//else it should be configured for external wakeup
	if (ExternalWakeup)
	{
		ADCCNTRL |= EXT;
		//More to do here to set up appropriate delays
		
		//To take the clock from the ADCIN pin
		//ADCGCR |= ADCIN;
	}
	else	//Internal Mode
	{
		/*
		ADCGCR |= PREF_CFG_INTERNAL;	//Use System Vcc
		ADCGCR |= NREF_CFG_INTERNAL;	//Use System Gnd
		*/
	}
	//If ExternalWakeup and Repeat then setup auto mode
	if (ExternalWakeup && Repeat)
	{
		ADCCNTRL |= AUTO;
	}

	//If Touchscreen then based on the algorithm set for Pen-Down Detection
	if (TouchScreen)
	{
		TouchScreenMode = TRUE;			//Indicate to the ISR/ADCReadResult
		ADCGCR |= PEN_DOWN_DETECT;
	}

	
	//Set the adc_result_callback_p to the return the result to that function
	//If the callback function pointer is set to NULL, then it is up to the application
	//to check for the results
	adc_cb_p = adc_result_callback_p;

	//Setup the Clock Divisor and Delays
	ADCACR |= CLK_DIV_0;	//Set it up for a 12 Mhz clock

	//ADCACR |= CLKSEL;	//For auxillary clock only

	//Enable the interrupt if there is a callback function to read the result
	if (adc_cb_p != NULL)
	{
  		ADCGCR |= ADCINTEN;
		icu_unmask(ADC_IRQ);
	}  
	
	//Clear the existing Channel number or pair to be used as Input
	ADCGCR &= 0xFFC7;

	//Get the channel number or channel pair according to the Channel input
	switch (Channel)
	{
	case 0:
		ADCGCR |= MUX_CFG_0;
		break;
	case 1:
		ADCGCR |= MUX_CFG_1;
		break;
	case 2:
		ADCGCR |= MUX_CFG_2;
		break;
	case 3:
		ADCGCR |= MUX_CFG_3;
		break;
	case 4:
		ADCGCR |= MUX_CFG_4;
		break;
	case 5:
		ADCGCR |= MUX_CFG_5;
		break;
	case 6:
		ADCGCR |= MUX_CFG_6;
		break;
	case 7:
		ADCGCR |= MUX_CFG_7;
		break;
	default:
		panic();	//Invalid Channel Number
	}

	//Enable the ADC
	ADCGCR |= ADCCLKEN;
}

/******************************************************************************
	
	Function:	AdcShutDown

	Description
	-----------
    This function shuts down the ADC block and disables the ADC interrupt.

	Arguments
	---------
	None.

******************************************************************************/

void AdcShutDown(void)
{

	//Wait for any existing conversion to finish
	while(ConvInProgress)
#ifdef UCOS
		OSTimeDly(1);
#else
	;
#endif
	//Clear the existing Channel number or pair to be used as Input
	ADCGCR &= 0xFFC7;
	
	//Disable the ADC Interrupt
	ADCGCR &= ~ADCINTEN;
	icu_mask(ADC_IRQ);
	  
	//Disable the clock to the ADC block
	ADCGCR &= ~ADCCLKEN;
	
}

/*************************************************************************************

	Function: AdcTouchScreenConfig

	Description
	-----------
    This function sets up the ADC block to sample the appropriate Co-ordinate for 
	the Touch Screen application.

	Arguments
	---------
	CoOrdinate:
	This paramter indicates which Co-ordinate has to be sampled next. The possible
	values are: SAMPLE_X, SAMPLE_Y, SAMPLE_Z, PEN_DOWN_DETECT and SAMPLE_Z_PRE_PENDOWN.


!- Note: Once a round of X, Y and/or Z co-ordinates have been sampled, it is up to the
		application to configure the ADC block to Pen Down Detection mode.

*************************************************************************************/


void AdcTouchScreenConfig(unsigned short CoOrdinate)
{
	if (TouchScreenMode)
	{
		//Based on CoOrdinate configure to read the value for the appropriate
		//Co-ordinate
		switch (CoOrdinate)
		{
		case SAMPLE_X:
			ADCGCR |= SAMPLE_X;
			break;
		case SAMPLE_Y:
			ADCGCR |= SAMPLE_Y;
			break;
		case SAMPLE_Z:
			ADCGCR |= SAMPLE_Z;
			break;
		case PEN_DOWN_DETECT:
			ADCGCR |= PEN_DOWN_DETECT;
			break;
		case SAMPLE_Z_PRE_PENDOWN:
			ADCGCR |= SAMPLE_Z_PRE_PENDOWN;
			break;
		default:
			panic();
		}
	}
	else	//The touchscreen mode has not been initialized
		panic();
}

/*************************************************************************************

	Function: AdcStartConversion

	Description
	-----------
    This function Starts the Analog to Digital Conversion.

	Arguments
	---------

!- Note: If the External Mode is not chosen, this is the way the A to D conversion has
		to be initiated. If the External Mode is chosen, this function will Prime the 
		ADC to start conversion on the next External signal.

*************************************************************************************/


void AdcStartConversion(void)
{
	//Wait for any existing conversion to finish
	while(ConvInProgress)
#ifdef UCOS
		OSTimeDly(1);
#else
	;
#endif
		//Start the conversion
		ADCSTART = 0x1;

#ifdef UCOS
		OS_ENTER_CRITICAL();
#endif
		ConvInProgress = TRUE;

#ifdef UCOS
		OS_EXIT_CRITICAL();
#endif
}

/*************************************************************************************

	Function: AdcReadResult

	Description
	-----------
    This function Reads the Result after the Analog to Digital Conversion has finished.

	Arguments
	---------
	None

	Returns
	---------
	Result: This is the 12-bit result of the Analog to Digital Conversion.
	
!- Note: If during the ADC initialization, there was no callback function provided to 
		read the results from the ADC, then this is another way for the application to
		read the results in a busy waiting loop.

*************************************************************************************/

short AdcReadResult(void)
{
	//if the conversion is finished Read the result and put it in Result
	do
	{
		AdcResult = ADCRESLT;

    }while (!(AdcResult & RSLT_ADC_DONE));
  
	AdcResult &= 0x1FFF;

#ifdef UCOS
		OS_ENTER_CRITICAL();
#endif
		ConvInProgress = FALSE;

#ifdef UCOS
		OS_EXIT_CRITICAL();
#endif
  
	//If the ADC is operating in TouchScreenMode, then re-enable the Pendown Detection
	/*if (TouchScreenMode)
	{
		ADCGCR |= PEN_DOWN_DETECT;
	}*/

	return (short)AdcResult;
}

/******************************************************************************  
    Function: adc_isr

	Description
	-----------
    This interrupt is generated when the ADC has finished a conversion and the
	result is available to be read.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.

!-Note: The callback function from inside the interrupt has a retun value:
		Result: This a pointer that holds the result of the Analog to Digital 
		Conversion.
*******************************************************************************/
#ifdef __CR__
#pragma interrupt (adc_isr)
void adc_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+ADC_IRQ)
__interrupt void adc_isr(void)
#else
__interrupt[(16+ADC_IRQ)*IDT_WIDTH] void adc_isr(void)
#endif
#endif
{
	//Read the result to clear the Interrupt 
	AdcResult = ADCRESLT;
    
	AdcResult &= 0x1FFF;

	ConvInProgress = FALSE;

	//Indicate that the result is available and if decided call the callback function
	/* Call the callback function to indicate the result is available */
	if (adc_cb_p != NULL)
  		(*adc_cb_p)(&AdcResult);

	//If the ADC is operating in TouchScreenMode, then re-enable the Pendown Detection
	/*if (TouchScreenMode)
	{
		ADCGCR |= PEN_DOWN_DETECT;
	}*/

}

#endif /* ADC_BASE_ADDR */
