/******************************************************************************
*  File:	audio.c
*  
*  This is the Advanced Audio Interface peripheral drivers.
******************************************************************************/

#include "chipdef.h"
#include "dl.h"

#define RX_FIFO_THRESHOLD 12
#define TX_FIFO_THRESHOLD 4
#define AAI_TX_FIFO_SIZE 16


aai_t * const aai = (aai_t *)AAI_BASE_ADDR;
cvsd_t * const cvsd = (cvsd_t *)CVSD_BASE_ADDR;

static dl_t *audio_tx_dl, *audio_rx_dl;
static unsigned char *audio_tx_ptr, *audio_rx_ptr;
static boolean cvsd_flag;
static void (*audio_rx_callback_p)(unsigned int);
static unsigned short control_bits;
unsigned short audio_dummy_read;


/******************************************************************************  
    Function: audio_init

	Description
	-----------
    This function initializes the AAI hardware and the AAI driver.
  
	Arguments
	---------
	config:
	A pointer to an audio configuration structure. Here are the fields and
	their meaning:
	- async_mode: a boolean indicating whether asynchronous mode should
	  be used for the codec (separate tx/rx clocks). False means
	  synchronous mode (single clock for tx and rx).
	- sample_length: a byte indicating the length in bits of a single voice sample 
	  used with the codec.
	- long_fs: a boolean indicating whether a long frame sync pulse should be used
	  with the codec. FALSE means short frame sync.
	- frame_sync_length: a byte indicating the length of the frame sync pulse in
	  bits. Applicable only if long frame sync pulse is used and if the sample
	  length is more than 8 bits. Valid value are 13-16 bits.
	- invert_fs: a boolean indicating whether the frame sync pulse should
	  be inverted. FALSE means not inverted.
	- num_of_slots: a byte indicating number of slots per frame. Valid value
	  are 1-4.
	- bitclock_prescalar: a byte which specifies the prescalar used to divide
	  the system clock to generate the codec bit clock.
	- bits_per_frame: A byte specifying the number of bits per codec frame. This
	  is used the generate the frame clock from the bit clock (using this number
	  as a as divisor). 
	- assigned_tx_slots: a bit mask indicating on which slots in the frame
	  voice transmission will be active. Bit n corresponds to slot n+1. Only
	  bits 0-3 should be used.
	- assigned_rx_slots: a bit mask indicating on which slots in the frame
	  voice reception will be active. Bit n corresponds to slot n+1. Only
	  bits 0-3 should be used.
	- cvsd_conv: indicates if and what type of conversion should be
	  done by the CVSD conversion block. Values are: 
	  * NOCVSDCONV (no CVSD conversion)
	  * CVSD2ULAW (CVSD <--> ulaw)
	  * CVSD2ALAW (CVSD <--> alaw)
	  * CVSD2LINEARPCM (CVSD <--> linear PCM).

	transmit_dl:
	A pointer to data link (circular buffer) control structure to be used for
	voice transmission (audio output). A NULL pointer means that no transmission
	is required.
	
	receive_dl:
	A pointer to data link (circular buffer) control structure to be used for
	voice reception (audio input).  A NULL pointer means that no reception
	is required.

	rx_callback:
	A pointer to a callback function to be called whenever input voice
	samples are written to the receive data link circular buffer.
	  
 *******************************************************************************/
 void audio_init(audio_config_t *config, dl_t *transmit_dl, dl_t *receive_dl,
 					void (*rx_callback_p)(unsigned int))
 {
	unsigned int i;
	audio_tx_dl = transmit_dl;
	audio_rx_dl = receive_dl;

	audio_rx_callback_p = rx_callback_p;	

	/* Need to enable clock before any other register access. */
	AGCR  |=  AACLKEN;

	/* Program global config register */
	AGCR |= (
			 (config->async_mode ? ASS : 0 ) |
			 (config->sample_length > 8 ? DWL : 0) |
			 (config->long_fs ? FSS : 0) |
			 ((unsigned short)(config->frame_sync_length - 13) << 10) |
			 (config->invert_fs ? IFS : 0) |
			 ((config->num_of_slots - 1) << 3) |
			 CRF |
			 CTF
			 );

	if (config->audio_conv == NOCVSDCONV) {
		/* No CVSD conversion required. AAI FIFO will be used. */

		cvsd_flag = FALSE;

		/* Program RX slots and RX FIFO interrupt threshold. */
		ARSCR = (ARSCR & 0x0f0f) | ((unsigned short)(config->assigned_rx_slots) << 4) |
					(unsigned short)(RX_FIFO_THRESHOLD << 12);
		/* Program TX slots and TX FIFO interrupt threshold. */
		ATSCR = (ATSCR & 0x0f0f) | ((unsigned short)(config->assigned_tx_slots) << 4) |
					(unsigned short)(TX_FIFO_THRESHOLD << 12);

		/* Fill AAI Tx FIFO with initial samples. This is absolutely necessary:
		   if it's not done the the Tx FIFO almost empty flag is never set by the
		   hardware and we don't get an interrupt.
		*/
		while (!(ATSCR & TXF))
			ATFR = 0;

		/* Enable or disable AAI interrupts on Tx/Rx. */
		if (transmit_dl != NULL)
			AISCR |= TXIE;
		else
			AISCR &= ~TXIE;
		if (receive_dl != NULL)
			AISCR |= RXIE;
		else
			AISCR &= ~RXIE;

		AISCR |= (TXIC | RXIC);

	} else {
		/* CVSD conversion required. DMA will be used to transfer PCM data betwen AAI
		   and CVSD blocks. CVSD FIFO and interrupt will be used at the CVSD end.
		*/

		cvsd_flag = TRUE;


		/* Program rx/tx slot assignment */
		ARSCR  =  (ARSCR & 0xf00f) | ((unsigned short)(config->assigned_rx_slots) << 4) |
									 ((unsigned short)(config->assigned_rx_slots) << 8);
		ATSCR  =  (ATSCR & 0xf00f) | ((unsigned short)(config->assigned_tx_slots) << 4) |
									 ((unsigned short)(config->assigned_tx_slots) << 8);

		/* Program rx/tx slot assignment for DMA*/
		ADMACR = (config->assigned_rx_slots) | (config->assigned_tx_slots << 4);

		/* Configure DMA channel 2: AAI->CVSD. */

		/* Set DMA target and destination addresses. */
		ADCA2 = ADRA2 = (void FAR *)(&ARDR0);
		ADCB2 = ADRB2 = (void FAR *)(&PCMIN);

		/* DMA transfer length is one word at a time. */
		BLTC2 = BLTR2 = 1;

		/* DMA options: word (16-bit) transfers, indirect, auto-initialize. */
		DMACNTL2 = (TCS | IND | OT);

		/* Enable DMA channel. */
		DMACNTL2 |= CHEN;

		/* Configure DMA channel 3: CVSD->AAI. */

		/* Set DMA target and destination addresses. */

		ADCA3 = ADRA3 = (void FAR *)(&ATDR0);
		ADCB3 = ADRB3 = (void FAR *)(&PCMOUT);

		/* DMA transfer length is one word at a time. */
		BLTC3 = BLTR3 = 1;

		/* DMA options: word (16-bit) transfers, direction device B->device A,
		   indirect, auto-initialize.
		*/
		DMACNTL3 = (TCS | IND | DIR | OT);
		DMACNTL3 |= CHEN;

		/* Enable CVSD conversion block. */
	    CVCTRL =  CVEN;

		/* Configure CVSD block. */

		CVCTRL |= CVSDINT | /*CVSDERRINT |*/
					(((unsigned short)(config->audio_conv)) << 9) | 
					((unsigned short)(16 - config->sample_length) << 12);

		/* Fill CVSD input FIFO with initial samples. This is absolutely necessary:
		   if it's not done the the CVSD FIFO nearly empty flag is never set by the
		   hardware and we don't get an interrupt.
		*/
		for (i = 0; i < 8; i++)
			CVSDIN = 0x5555;

		/* Set auxiliary clock 2 prescalar to 6 to gnenerate a 2MHz clock 
		   required by the CVSD block.
		*/
		PRSAC   =  (PRSAC & 0x0f) | 0x50;
		/* Enable auxiliary clock 2. */
    	CRCTRL  |=  ACE2;

	}


  	/* Enable auxiliary clock 1. This clock is used by the AAI. 
  	   Set the auxilary clock 1 prescalar to 1.
	*/	
	PRSAC &= 0xf0;
	CRCTRL |= ACE1;

	ACCR =
		/* Bit clock prescalar value minus 1 shifted to its right position. */ 
		((config->bitclock_prescalar - 1) << 8) |
		/* Frame clock prescalar (= number of bits per frame) minus 1 shifted
		   to its right position.
		*/
		((config->bits_per_frame - 1) << 1);

	/* Configure the AAI pins to their alternate (audio) function. */
	AAI_MUX_PORT->pxalt |= AAI_MUX_MASK;
 }

/******************************************************************************  
    Function: audio_set_control_bits

	Description
	-----------
    This function sets the number and contents of the control bits that are
	appended to each transmitted word. Typically this is used by codecs for
	volume control.
  
	Arguments
	---------
	n_bits:
	The number of bits to be appended (0-3).

	value:
	The value or contents of the control bits (0x0-0x7).
	None.
 *******************************************************************************/
 void audio_set_control_bits(unsigned int n_bits, unsigned char value)
 {
    if (cvsd_flag)
		/* In this case samples are transferred by the DMA. Control bits have
		   to be programmed in the ADMACR register.
		*/
		ADMACR = ((ADMACR & 0xe0ff) | ((n_bits & 0x3) << 11)
									| ((value & 0x7) << 8));
	else
		/* Save control bits in a static variable. The value is and'ed with
		   a mask has 1's in bit positions 0 to n_bits - 1.
		*/
		control_bits = value & ((1 << n_bits) - 1);
 }

/******************************************************************************  
    Function: audio_start

	Description
	-----------
    This function starts the AAI/CVSD hardware.
  
	Arguments
	---------
	None.
 *******************************************************************************/
 void audio_start(void)
 {
	if (audio_tx_dl != NULL) 
		dl_init_ptr(audio_tx_dl, audio_tx_ptr);
	if (audio_rx_dl != NULL)
		dl_init_ptr(audio_rx_dl, audio_rx_ptr);

	/* Start the AAI module.*/
	AGCR |= AAIEN;


	if (cvsd_flag) {
		/* Start the CVSD module. */
		CVCTRL |= CVCLKEN;
		/* Unmask the CVSD interrupt. */
		icu_unmask(CVSD_IRQ);
	}
	else
		/* Unmask AAI interrupt */
		icu_unmask(AAI_IRQ);
 }


 /******************************************************************************  
    Function: audio_stop

	Description
	-----------
    This function stops the AAI/CVSD hardware.
  
	Arguments
	---------
	None.
 *******************************************************************************/
 void audio_stop(void)
 {
	/* Stop the AAI module. */
	AGCR &= ~(AAIEN);

	/* Stop the CVSD module. */
	CVCTRL &= ~(CVCLKEN);

	/* Mask interrupts */
	icu_mask(CVSD_IRQ);
	icu_mask(AAI_IRQ);
 }

/******************************************************************************  
    Function: aai_isr

	Description
	-----------
    This function is the AAI interrupt service routine.
  
	Arguments
	---------
	None.
 *******************************************************************************/
#ifdef __NSC_CRCC__
#pragma interrupt (aai_isr)
void aai_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+AAI_IRQ)
__interrupt void aai_isr(void)
#endif
{
	unsigned int sample_count;

	if ((AISCR & RXIE) /*&& (ARSCR & RXAF)*/) {
		/* Rx FIFO almost full */

		/* Clear Rx interrupt. */
		AISCR |= RXIC;

		sample_count = 0;


		/* Empty Rx FIFO. */
		while (!(ARSCR & RXE)) {
			if (!dl_full(audio_rx_dl)) {
				/* If there is room copy sample to data link buffer. */
				*((unsigned short *)audio_rx_ptr) = ARFR >> 3;
				dl_advance_multiple(audio_rx_dl, audio_rx_ptr, 2);
				sample_count++;
			} else
				/* Otherwise make a dummy read from the FIFO to avoid overrun. */
				audio_dummy_read = ARFR;
		}

		/* Commit the produced data. */
		dl_p_commit(audio_rx_dl, sample_count << 1);

		/* Call callback function if necessary. */
		if (audio_rx_callback_p != NULL)
			(*audio_rx_callback_p)(sample_count << 1);
	}

	if ((AISCR & TXIE)/* && (ATSCR & TXAE)*/) {
		/* Tx FIFO almost empty */

		/* Clear Tx interrupt. */
		AISCR |= TXIC;

		sample_count = 0;

		/* Fill Tx FIFO. */
		while (!(ATSCR & TXF)) {
			if (dl_empty(audio_tx_dl))
				/* If no data available in data link buffer write 0 to FIFO. */
				ATFR = 0 | control_bits;
			else {
				/* Otherwise copy data from data link buffer into FIFO. */
				ATFR = ((*((unsigned short *)audio_tx_ptr)) << 3) | control_bits;
				dl_advance_multiple(audio_tx_dl, audio_tx_ptr, 2);
				sample_count++;
			}
		}
		/* Commit the consumed data. */
		dl_c_commit(audio_tx_dl, sample_count << 1);
	}
}


/******************************************************************************  
    Function: pcm2cvsd_isr

	Description
	-----------
    This function is the CVSD interrupt service routine.
  
	Arguments
	---------
	None.
 *******************************************************************************/
#ifdef __NSC_CRCC__
#pragma interrupt (pcm2cvsd_isr)
void pcm2cvsd_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+CVSD_IRQ)
__interrupt void pcm2cvsd_isr(void)
#endif
{
	unsigned short cvstat;
	unsigned int i, fifo_count, avail, fill;

	/* Save CVSD status register value. */
	cvstat = CVSTAT;

	/* If CVSD input FIFO is nearly empty - fill it up. */
	if (cvstat & CVNE) {
		/* Get number of empty FIFO entries. Note: the CVINST field shows
		   X-1 empty locations for X=1..8 and 0 for 0 empty location. As
		   a results there is no way to distinguish between 0 and 1 empty
		   locations (CVINST shows 0 in both cases). To avoid overflow, we
		   take a defensive approach here and fill exactly CVINST locations
		   even though in most cases it one less than what we could fill.
		*/
		fifo_count = (cvstat & CVINST) >> 5;
		/* Get number of words we can read from the circular buffer. */
		avail = dl_cnt(audio_tx_dl) >> 1;
		/* Take the minimum of the two. */
		if (avail < fifo_count) {
			fill = fifo_count - avail;
			fifo_count = avail;
		}
		else
			fill = 0;

		/* Copy this number of words from circular buffer to CVSD FIFO. */
		for (i = 0; i < fifo_count; i++) {
			CVSDIN = *((unsigned short *)audio_tx_ptr);
			dl_advance_multiple(audio_tx_dl, audio_tx_ptr, 2);
		}
		/* Fill the remaining FIFO empty words with checkboard pattern. */
		for(i = 0; i < fill; i++)
			CVSDIN = 0x5555;

		/* Update data link (circular buffer) count. */
		dl_c_commit(audio_tx_dl, fifo_count << 1);
	}

	/* If CVSD output FIFO is nearly full - empty it. */
	if (cvstat & CVNF) {
		/* Get number of words currently in FIFO Note that the CVOUTST field
		   has no way to indicate 8 words in the FIFO. It shows 7 for both
		   7 and 8 words. As a result if have 8 words in the FIFO, we will 
		   read 7 and thus will not completely empty the FIFO.
		*/
		fifo_count = (cvstat & CVOUTST) >> 8;
		/* Get number of words we can write to the circular buffer. */
		avail = dl_p_space_available(audio_rx_dl) >> 1;
		/* Take the minimum of the two. */
		if (avail < fifo_count) {
			fill = fifo_count - avail;
			fifo_count = avail;
		}
		else
			fill = 0;
		/* Copy this number of words from CVSD FIFO to circular buffer. */
		for (i = 0; i < fifo_count; i++) {
			*((unsigned short *)audio_rx_ptr) = CVSDOUT;
			dl_advance_multiple(audio_rx_dl, audio_rx_ptr, 2);
		}
		/* Empty the remaining words from the FIFO. */
		for(i = 0; i < fill; i++)
			audio_dummy_read = CVSDOUT;

		/* Update data link (circular buffer) count. */
		dl_p_commit(audio_rx_dl, fifo_count << 1);
		/* Call callback function if necessary. */
		if (audio_rx_callback_p != NULL)
			(*audio_rx_callback_p)(fifo_count << 1);
	}
}
