
/************************************************************************
	File Name : can.c

	CAN bus interface driver

	The public API consists of:

	can_init()
	can_config_tx()
	can_config_rx()
	can_read_msg()
	can_shutdown()

	All other functions are internal to the driver.

************************************************************************/

#include <stdio.h>
#include "crcommon.h"
#include "chipdef.h"
#include "drivers.h"

#ifdef UCOS
#include "ucos.h"
#endif

// Base address table for the message buffers
static can_mbuf_t * const can_mbuf[] = {
	CAN_MBUF_BASE_ADDRS };

// Base address table of control registers
static can_t * const can_reg[] = {
	CAN_REG_BASE_ADDRS };

// Base address table of ports muxed with each CAN interface. 
static gpio_t * const can_mux_port_tab[] = {
	CAN_MUX_PORTS };

// Table of bit masks of pins of ports that are masked with the pins of each
// CAN interface
static unsigned char const can_mux_mask_tab[] = {
	CAN_MUX_MASKS };

// Table of transmit interrupt IRQs for each CAN interface
static unsigned char const can_irqs[] = {
	CAN_IRQS }; 

// Table of callbacks, one per CAN block
void   (*mbuf_callback_fn[N_CANBLOCKS])(
			    unsigned short mbuf_num, 
			    unsigned short cediag );

/******************************************************************************  
    Function: calc_CAN_bus_bit_timing

	Description
	-----------
	Calculates the value for CTIM, according to bus frequency required
	relative to CPU frequency. Places bit sampling point approximately half way
	along the bit period. Internal to driver.

	Arguments
	---------
	bus_freq_ratio:
	Value is ratio (system freq)/(CAN bus freq). eg for 12 MHz CPU 
	and 50 KHz CAN bus bit rate, bus_freq_ratio=240.

	Return value
	------------
	Returns the value for CTIM.

*******************************************************************************/
static unsigned short calc_CAN_bus_bit_timing( unsigned int bus_freq_ratio )
{
    short  Presc, Tseg1, Tseg2, N, P, Sjw;

    Presc   =   2;   
	P = bus_freq_ratio;                              

    while (1) 
    {
        N       =   P;
        Tseg1   =   0;
        Tseg2   =   -1;       /* This accounts for the SYNC time... */
        Sjw     =   1;
                                                    
        while (Tseg2 < 8 && N >= Presc)	/* Tseg2 register may not exceed 8  */
        {                               /* Alternately increment Tseg values, 
										   checking for best divisor... */
            if (Sjw)                    
            {
                Tseg2   +=  1;
                --Sjw;
            }
            else 
            {
                Tseg1   +=  1;
                ++Sjw;
            }
                                                    
            N   -=  Presc;	    /* Subtract latest addition */
        }
                                                    
        if ((Tseg2 == 8 || N))	/* Try bumping the prescaler... */
            ++Presc;                                           
	else
            break;              /* Otherwise, we're done */
    }
                                                    

    if (Tseg1 > 1) /* Tseg1 must NOT be less than 2!! */
    {                 
	  /* Sync Jump Width must not be negative!! */
        Sjw =   (Tseg2 - 1) ? Tseg2 - 1: Tseg2;	
                                                    
		Presc -=  2;
		Presc <<= 9;             /* Shift to proper position */

		Sjw--;                   /* SJW is val-1... */
		Sjw <<= 7;               /* Shift to its place */
      
		Tseg1--;                 /* Tseg1,2 is val-1... */
		Tseg1   <<= 3;
		Tseg2--;
                                                    
       return (Presc | Sjw | Tseg1 | Tseg2);	/* All Values for CTIM */
    }
	else
		return 0;
}

/******************************************************************************  
    Function: can_init

	Description
	-----------
    This function initialises the CAN hardware and CAN driver. After
	returning from this function the individual CAN message buffers are ready
	to be configured for transmission and reception. Puts message buffers into
	benign state and zeros all contents.

	Arguments
	---------
	can_id:
	Which CAN HW block to use, starting at 0. Got only one? Use 0.

	global_config:
	Set message buffers into locking mode, and configures 
	direction of data byte. Values ORed together.

	bus_freq_ratio:
	Set CAN bus bit rate relative to CPU frequency. Value is
	ratio (system freq)/(CAN bus freq). eg for 12 MHz CPU and 50 KHz CAN bus
	bit rate bus_freq_ration=240.

	global_mask:
	Sets standard message mask for message buffers 0-14, 1 indicates 
	don't care for similar bit in ID1. Note that RTR bit is automatically
	masked when a remote message is sent to ensure reception of response

	global_mask_ext:
	Sets extended message mask for message buffers 0-14, 1 indicates 
	don't care for similar bit in ID0. Note that RTR bit is automatically
	masked when a remote message is sent to ensure reception of response

	basic_mask:
	Sets standard message mask for message buffer 15, 1 indicates 
	don't care for similar bit in ID1.

	basic_mask_ext:
	Sets extended message mask for message buffer 15, 1 indicates 
	don't care for similar bit in ID0.

	mbuf_callback:
	Callback function invoked whenever interrupt on ANY message buffer is
	activated. Setting this to non-null causes setup of interrupts for this
	CAN block. Null value prevents use of interrupts.
		Parameters:
		mbuf_num: IF cediag 0 indicates which message buffer caused the 
			interrupt, could be due to rx or tx.
		cediag: if non-zero is copy of CEDIAG register indicating error

	Return value
	------------
	None.

*******************************************************************************/
void can_init( 
		unsigned int can_id,
		can_config_t global_config,
		unsigned int  bus_freq_ratio, 
		unsigned short global_mask, 
		unsigned short global_mask_ext, 
		unsigned short basic_mask, 
		unsigned short basic_mask_ext, 
		void   (*mbuf_callback)(unsigned short mbuf_num, 
								unsigned short cediag) )
{
	unsigned int i,j;
	can_t *can;

	for( j = 0; j < (15 * sizeof(can_mbuf_t)); j++)
		((unsigned char *) can_mbuf[can_id])[j] = 0;

	can = can_reg[can_id];

	/* Set shared GPIO pins to CAN */
	(can_mux_port_tab[can_id])->pxalt |= can_mux_mask_tab[can_id];
	(can_mux_port_tab[can_id])->pxalts &= ~can_mux_mask_tab[can_id];

	/* Set masks */
	GMSKB = global_mask;
	GMSKX = global_mask_ext;
	BMSKB = basic_mask;
	BMSKX = basic_mask_ext;
    
	CTIM = calc_CAN_bus_bit_timing( bus_freq_ratio );

	mbuf_callback_fn[can_id] = mbuf_callback;
	if (mbuf_callback) /* Caller wants interrupts if there's a callback */
	{
		CICLR = 0xffff; /* Clear all pending interrupts before enabling */
		icu_unmask(can_irqs[can_id]); /* Unmask Interrupt controller line */
	}

	/* In the Device Spec it says to only enable the block once the message
	buffers are configured. Actually it is sufficient to ensure any random data
   in CNSTAT is set to a benign state eg RX/TX_NOT_ACTIVE so junk isn't tx'd */
	for( i=0; i<CAN_NUM_MBUFS; i++)
		can_mbuf[can_id][i].cntstat = 0; // Actually RX_NOT_ACTIVE

	CGCR = CANEN | global_config;
//	CGCR |= 0x0780; /* Loopback etc... for debugging */
}

/******************************************************************************  
    Function: can_config_tx

	Description
	-----------
	Configure a message buffer to transmit a packet, any previous configuration
	is completely overridden. There are five modes (see tx_type parameter and
	can_tx_type_t)

	Arguments
	---------
	can_id:
	Which CAN HW block to use, starting at 0. Got only one? Use 0.

	mbuf_num:
	Which message buffer to set into this mode, 0-14.

	id:
	The Message Identifier

	msg:
	Pointer to array of max 8 bytes of payload:

	len:
	length of payload

	msg_type:
	Normal or extended

	priority:
	The priority to set the message to.

	tx_type:
		SEND_DATA: Send a single data frame.
		REMOTE_REQ: Send a remote frame and setup to receive the response, user 
			calls can_read_msg() to retrieve the response from the message buff
		RESPOND_TO_REMOTE: Loads the message buffer with data etc and waits for
			a matching Remote frame before transmitting.
		SEND_THEN_RTR: Same as RESPOND_TO_REMOTE but send the frame once before
			waiting for an incoming Remote frame.
		TX_CANCEL: Cancel any pending transmit or RESPOND_TO_REMOTE setting

	Return value
	------------
	ERR_MBUF_BUSY if the buffer is busy already. To force a new setup, call fn
	with TX_CANCEL first.

*******************************************************************************/
unsigned int can_config_tx( 
		unsigned int can_id,
		unsigned char mbuf_num, 
		unsigned long id, 
		unsigned char *msg, 
		unsigned int  len,
		msg_type_t    msg_type,
		unsigned int  priority,
		can_tx_type_t tx_type )
{
	can_mbuf_t *curr_mbuf;
	unsigned char *mbuf_data;
	int i;
	can_t *can;	 

	can = can_reg[can_id];
	curr_mbuf = &can_mbuf[can_id][mbuf_num];

	if (tx_type == TX_CANCEL) /* Stop pending TX as requested */
	{
		curr_mbuf->cntstat = (curr_mbuf->cntstat & ((unsigned short)~ST)) | TX_CANCEL;
		return 0;
	}

	curr_mbuf->cntstat = (curr_mbuf->cntstat & ((unsigned short)~ST)) | ST_TX_NOT_ACTIVE;

	/* Don't update mbuf if transmission still occuring */
	if (curr_mbuf->cntstat & MBUF_BUSY)
		return(ERR_MBUF_BUSY);

	if ( msg_type == CAN_STD)
	{
	    curr_mbuf->id1 = id << 5;   /* Place STD mesg id in ID1 */

		if (tx_type == REMOTE_REQ)
		{
		    GMSKB |= ID1_RTR;		   // Ensure that mbuf can receive RTR
			curr_mbuf->id1 |= ID1_RTR; // This is a remote message
		    tx_type = SEND_DATA;       // Actual Tx state set in HW 
		}
		else if (tx_type == RESPOND_TO_REMOTE)
		    GMSKB |= ID1_RTR;		   // Ensure that mbuf can receive Remote

		curr_mbuf->id1 &= ~IDE;	/* Indicate STD message */
	}
	else /* CAN_EXT message */
	{
	    /* Get mesg id into right positions of ID0,1 registers */
		curr_mbuf->id0 = id << 1;    
		curr_mbuf->id1 = ((id >> 13) && ID28_18) | ((id >> 15) && ID17_15);

		if (tx_type == REMOTE_REQ)
		{
		    GMSKX |= ID0_RTR;		   // Ensure that mbuf can receive RTR
			curr_mbuf->id0 |= ID0_RTR; // This is a remote message
		    tx_type = SEND_DATA;       // Actual Tx state set in HW
		}
		else if (tx_type == RESPOND_TO_REMOTE)
		    GMSKX |= ID0_RTR;		   // Ensure that mbuf can receive Remote

		curr_mbuf->id1 |= ID1_RTR; /* Set SRR in extended message */
		curr_mbuf->id1 |= IDE; /* Indicate Extended message */
	}

	// If callback is defined setup for interrupt mode.
	if (mbuf_callback_fn[can_id])
	{
		CICLR |= (1 << mbuf_num); /* Clear pending interrupts */
		CIEN |= (1 << mbuf_num);  /* Enable int for message buffer */
		CICEN |= (1 << mbuf_num);  /* Enable int code for message buffer */
	}

	if (msg)
	{
		/* Copy payload data */
		mbuf_data = ((unsigned char *) &curr_mbuf->data0) + 1; 
		for( i = 0; i < len; i++ )
			*mbuf_data-- = *msg++;
	}

	/* Rx/Tx direction set and initiated here */
	curr_mbuf->cntstat |= (len << 12) | (priority << 4) | tx_type;

	return 0;
}

/******************************************************************************  
    Function: can_config_rx

	Description
	-----------
	Configure the message buffer to receive a Data Frame only. Call 
	can_read_msg() to retrieve the payload data and details of message
	actually received. 

	Arguments
	---------
	can_id:
	Which CAN HW block to use, starting at 0. Got only one? Use 0.

	mbuf_num:
	Which message buffer to setup for receiving, 0-14.

	msg:
	Pointer to array of max 8 bytes of payload:

	id:
	The Message Identifier

	msg_type:
	Normal or extended

	Return value
	------------
	ERR_MBUF_BUSY if the buffer is busy already otherwise zero.

*******************************************************************************/
unsigned int can_config_rx( 
		unsigned int can_id,
		unsigned char mbuf_num, 
		unsigned long id, 
		msg_type_t    msg_type )
{
	can_mbuf_t *curr_mbuf;
	can_t *can;

	can = can_reg[can_id];
	curr_mbuf = &can_mbuf[can_id][mbuf_num];

	/* Don't update mbuf if BUSY */
	if (curr_mbuf->cntstat & MBUF_BUSY)
		return(ERR_MBUF_BUSY);

	if ( msg_type == CAN_STD)
	{
		curr_mbuf->id1 = id << 5;   /* Place STD mesg id in buffer */
		curr_mbuf->id1 &= ~IDE;	/* Indicate STD message */
	}
	else /* CAN_EXT message */
	{
	  /* Get mesg id into right positions of ID0,1 registers */
		curr_mbuf->id0 = id << 1;    
		curr_mbuf->id1 = ((id >> 13) && ID28_18) | ((id >> 15) && ID17_15);

		curr_mbuf->id1 |= (IDE | ID1_RTR); /* Indicate Extended message */
	}

	if (mbuf_callback_fn[can_id])
	{
		CICLR |= (1 << mbuf_num); /* Clear pending interrupts */
		CIEN |= (1 << mbuf_num);  /* Enable int for message buffer */
		CICEN |= (1<< mbuf_num);  /* Enable int code for meassage buffer */
	}

	curr_mbuf->cntstat = ST_RX_READY;

	return 0;
}

/******************************************************************************  
    Function: read_msg

	Description
	-----------
	Internal to driver (can_read_msg() is caller). Reads data from message
	buffer and gets message Id etc... Caller is expected to check and handle
	state of the buffer.

	Arguments
	---------
	curr_mbuf:
	Pointer to message buffer to get data from.

	msg:
	Pointer to array of max 8 bytes to place payload.

	msg_type:
	Normal or extended message type, read from buffer and placed in argument.

	id:
	The Message Identifier collected from the bits spread around the registers
	and placed in argument.

	Return value
	------------
	Length of payload.

*******************************************************************************/ 
static unsigned int read_msg(
		can_mbuf_t *curr_mbuf, 
		unsigned char *msg,
		msg_type_t *msg_type,	
		unsigned long *id )
{
	unsigned char *mbuf_data;
	unsigned int len;
	int i;

	len = (curr_mbuf->cntstat >> 12);

	// Extract the ID bits that are spread around different registers.
	if (curr_mbuf->id1 & IDE)
	{
		if (msg_type)
			*msg_type = CAN_EXT;

		if ((id))
			*id = ((ID28_18 & curr_mbuf->id1) << 13)| 
				  ((ID17_15 & curr_mbuf->id1) >> 15)|
				  (curr_mbuf->id0 >> 1);
	}
	else
	{
		if (msg_type)
			*msg_type = CAN_STD;

		if (id)
			*id = curr_mbuf->id1 >> 5;
	}
	
	/* Copy payload data */
	if (msg)
	{
		mbuf_data = ((unsigned char *) &curr_mbuf->data0) + 1; 
		for( i = 0; i < len; i++ )
			*msg++ = *mbuf_data--;
	}

	return len;
}

/******************************************************************************  
    Function: can_read_msg

	Description
	-----------

	Arguments
	---------
	can_id:
	Which CAN HW block to use, starting at 0. Got only one? Use 0.

	mbuf_num:
	Which message buffer to read message from, 0-14.

	id:
	The Message Identifier is returned in this parameter.

	msg:
	Pointer to array of max 8 bytes to write payload into priority.

	msg_type:
	Normal or extended is returned in this pointer parameter.

	priority:
	Pointer set to the priority of the message received.

	tstamp:
	Bit time stamp of message reception is returned in this pointer parameter.

	Return value
	------------
	Returns length of payload.

*******************************************************************************/
unsigned int can_read_msg( 
		unsigned int   can_id,
		unsigned char  mbuf_num, 
		unsigned long *id, 
		unsigned char *msg,
		msg_type_t    *msg_type,
		unsigned int *priority,
		unsigned short *tstamp )
{
	can_mbuf_t *curr_mbuf;
	unsigned int len;
	can_t *can;

	can = can_reg[can_id];
	curr_mbuf = &can_mbuf[can_id][mbuf_num];

	if (CGCR & BUFFLOCK)
	{
		len = read_msg( curr_mbuf, msg, msg_type, id );
		curr_mbuf->cntstat = ST_RX_READY | (curr_mbuf->cntstat & ((unsigned short)~ST));	
	}
	else /* Without locking have to keep checking just in case mesg 
			gets overwritten part way through */
		do{
			/* If still in READY state nothing received */
			if ((curr_mbuf->cntstat & ST) == ST_RX_READY)
				return(0);

			while(curr_mbuf->cntstat & MBUF_BUSY); /* Message currently being 
													copied into mbuf by HW, wait */

			curr_mbuf->cntstat = ST_RX_READY | (curr_mbuf->cntstat & ((unsigned short)~ST));	

			len = read_msg( curr_mbuf, msg, msg_type, id ); 

			while(curr_mbuf->cntstat & MBUF_BUSY) ; /* Wait again if necessary */

		} while (((curr_mbuf->cntstat & ST) == ST_RX_FULL) || 
				 ((curr_mbuf->cntstat & ST) == ST_RX_OVERRUN));

	if (priority)
		*priority = (curr_mbuf->cntstat & PRI) >> 4;

	if (tstamp)
		*tstamp = curr_mbuf->tstp;

	return len;
}

/******************************************************************************  
    Function: can_shutdown

	Description
	-----------
	Shuts down the CAN driver and given HW interface, turns off interrupts. 
	To use given CAN interface subsequently, can_init() should be called.

	Arguments
	---------
	can_id:
	Which CAN HW block to use, starting at 0. Got only one? Use 0.

	Return value
	------------
	None.

*******************************************************************************/
void can_shutdown(unsigned int can_id)
{
	can_t *can;
	can = can_reg[can_id];

	// Disable the block
	CGCR &= ~CANEN;
	
	icu_mask(can_irqs[can_id]); /* Mask Interrupt controller line */
}

/******************************************************************************  
    Function: can_isr

	Description
	-----------
	This one function gets called for EVERY CAN interrupt, independent of CAN
	block. It distinguishes between error interrupts and message buffer rx/tx
	interrupts. If an error caused the interrupt then the contents of the 
	CEDIAG register is passed to the callback. If an rx/tx interrupt the 
	mesage buffer id is passed to the callback. Called by canX_isr().

	Arguments
	---------
	can_id:
	Indicates the CAN block with pending interrupt.

	Return value
	------------
	None.

*******************************************************************************/
static void can_isr(unsigned int can_id)
{
	unsigned int flag = 1, mbuf_num = 0;
	can_t *can;

	can = can_reg[can_id];

	if ( CIPND & EIPND) /* Interrupt due to Error */
	{
		if (mbuf_callback_fn[can_id])
			mbuf_callback_fn[can_id]( 0, CEDIAG );

		/* Clear error interrupt */
		CICLR = EICLR;
	}
	else /* Interrupt due to tx/rx activity */
	{
		mbuf_num = (CSTPND & IST) - 1;
		flag = 1 << mbuf_num;

		if (mbuf_callback_fn[can_id])
			mbuf_callback_fn[can_id]( mbuf_num, 0 );

		/* Clear interrupt */
		CICLR = flag;
	}
}

/******************************************************************************  
    Function: can0_isr

	Description
	-----------
	Just calls can_isr() the real handler
	Arguments
	---------
	None.

	Return value
	------------
	None.

*******************************************************************************/
#ifdef CAN0_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(can0_isr)
void can0_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+CAN0_IRQ)
__interrupt void can0_isr( void )
#endif
{
	can_isr(0);
}
#endif

/******************************************************************************  
    Function: can1_isr

	Description
	-----------
	Just calls can_isr() the real handler
	Arguments
	---------
	None.

	Return value
	------------
	None.

*******************************************************************************/
#ifdef CAN1_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(can1_isr)
void can1_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+CAN1_IRQ)
__interrupt void can1_isr( void )
#endif
{
	can_isr(1);
}
#endif

