
/************************************************************************
	File Name : flash.c

	Function:	This is a flash memory driver.

  Note: This driver draws on several source of information and 
        code example. 
        It first and foremost relies on the specs as described in 
        CP3BT10 Reprogrammable Connectivity Processor
        with Bluetooth TM and USB Technologies(CP3BT10.pdf or CP3BT13.pdf)
        then earlier work from Stuart Corning with flash.c and
        finally flash.c from the released BTDK.
        The present driver attempt to cover the full section 8.0 of the 
        specs and combine features and implementation decisions made in 
        earlier work. It tries to generalize the constructs and apply same
        code for Flash Data and Flash Program abstracted from addressing 
        differences.


************************************************************************/
#include <stdio.h>
#include <stdlib.h>		  
#include <string.h>

#include "chipdef.h"


/******************************************************************************
  Functions
       FlashInit - Flash Timing intialization for specified type and frequency
       FlashInformationBlockRead - Read data from  specified offset within the info Block
       FlashInformationBlockWrite - Write data at spefied offset with the info block
       FlashInformationBlockErase - Erase Info block specifed by address
************************************************************************/

fm_t * const fm = (fm_t *)PROGFM_BASE_ADDR;
fsm_t * const fsm = (fsm_t *)DATAFSM_BASE_ADDR;

int flashStatus = 0;
/* Defines */

/******************************************************************************
	
	Function: FlashInit
	
	Description
	-----------
        This function initializes the data flash, to make sure that the timing for
	accessing and writing to data flash, to appropriate frequency.
  
	Arguments
	---------
	Type : a memory type for which frequencies get intialized
        Frequency: a frequency identificator to access the appropriate constants

	Return value
	------------
	None
	
        Note: 
        ----
             
*******************************************************************************/

// Need to resolve space issue before puting freq support 
typedef struct freq_data {
  uint8 pslr;
  uint8 start;
  uint8 tran;
  uint8 prog;
  uint8 perase;
  uint8 merase0;
  uint8 merase1;
  uint8 end;
  uint8 mend;
  uint8 rcv;
} freq_data_t;

static freq_data_t freqDataTable[5] = { 
 {
      FMPSLR_8Mhz, 
      FMSTART_8Mhz, 
      FMTRAN_8Mhz, 
      FMPROG_8Mhz, 
      FMPERASE_8Mhz,  
      FMMERASE0_8Mhz,  
      FMMERASE1_8Mhz,  
      FMEND_8Mhz,  
      FMMEND_8Mhz,  
      FMRCR_8Mhz  
   },
   {
      FMPSLR_12Mhz, 
      FMSTART_12Mhz, 
      FMTRAN_12Mhz, 
      FMPROG_12Mhz, 
      FMPERASE_12Mhz,  
      FMMERASE0_12Mhz,  
      FMMERASE1_12Mhz,  
      FMEND_12Mhz,  
      FMMEND_12Mhz,  
      FMRCR_12Mhz  
   },
   { 
      FMPSLR_16Mhz, 
      FMSTART_16Mhz, 
      FMTRAN_16Mhz, 
      FMPROG_16Mhz, 
      FMPERASE_16Mhz,  
      FMMERASE0_16Mhz,  
      FMMERASE1_16Mhz,  
      FMEND_16Mhz,  
      FMMEND_16Mhz,  
      FMRCR_16Mhz  
   },
   {
      FMPSLR_20Mhz, 
      FMSTART_20Mhz, 
      FMTRAN_20Mhz, 
      FMPROG_20Mhz, 
      FMPERASE_20Mhz,  
      FMMERASE0_20Mhz,  
      FMMERASE1_20Mhz,  
      FMEND_20Mhz,  
      FMMEND_20Mhz,  
      FMRCR_20Mhz  
    },
    {
      FMPSLR_24Mhz, 
      FMSTART_24Mhz, 
      FMTRAN_24Mhz, 
      FMPROG_24Mhz, 
      FMPERASE_24Mhz,  
      FMMERASE0_24Mhz,  
      FMMERASE1_24Mhz,  
      FMEND_24Mhz,  
      FMMEND_24Mhz,  
      FMRCR_24Mhz  
    }
};


void FlashInit(memory_t Type, frequency_t Frequency)
{
  if ( Type == FLASH_DATA) {
    FSMPSR = freqDataTable[Frequency].pslr; 
    FSMSTART =  freqDataTable[Frequency].start;   
    FSMTRAN =  freqDataTable[Frequency].tran;  
    FSMPROG =  freqDataTable[Frequency].prog; 
    FSMPERASE =  freqDataTable[Frequency].perase; 
    FSMMERASE0 =  freqDataTable[Frequency].merase0; 
    FSMMERASE1 =  freqDataTable[Frequency].merase1;
    FSMEND =  freqDataTable[Frequency].end; 
    FSMMEND =  freqDataTable[Frequency].mend;
    FSMRCV =  freqDataTable[Frequency].rcv;
  }
  if ( Type == FLASH_PROGRAM) {
    FMPSR = freqDataTable[Frequency].pslr; 
    FMSTART =  freqDataTable[Frequency].start;   
    FMTRAN =  freqDataTable[Frequency].tran;  
    FMPROG =  freqDataTable[Frequency].prog; 
    FMPERASE =  freqDataTable[Frequency].perase; 
    FMMERASE0 =  freqDataTable[Frequency].merase0; 
    FMMERASE1 =  freqDataTable[Frequency].merase1;
    FMEND =  freqDataTable[Frequency].end; 
    FMMEND =  freqDataTable[Frequency].mend;
    FMRCV =  freqDataTable[Frequency].rcv;
  }

} /* end FlashInit */

/******************************************************************************  
    Function: FlashGetPageSize

	Description
	-----------
        This function returns the flash page size information to the calling
        function.
  
	Arguments
	---------
        Type:
        A byte containing the type of memory to read. 
        Either FLASH_PROGRAM or FLASH_DATA

 
	Return value
	------------
		The page size of the type of flash memory.
  
*******************************************************************************/

unsigned short FlashGetPageSize(memory_t Type)
{
	/* Return the Flash Page Size for the type of Flash Memory */
	return(Type == FLASH_PROGRAM ? FLASH_PROGRAM_PAGE_SIZE : FLASH_DATA_PAGE_SIZE);
}

/******************************************************************************  
    Function: FlashGetPageNum

	Description
	-----------
        This function returns the number of flash pages information to the calling
        function.
  
	Arguments
	---------
        Type:
        A byte containing the type of memory to read. 
        Either FLASH_PROGRAM or FLASH_DATA

 
	Return value
	------------
		The number of pages of the type of flash memory.
  
*******************************************************************************/

unsigned short FlashGetPageNum(memory_t Type)
{
	/* Return the Number of Flash Pages for the type of Flash Memory */
	return(Type == FLASH_PROGRAM ? FLASH_PROGRAM_PAGE_NUM : FLASH_DATA_PAGE_NUM);
}


/******************************************************************************  
    Function: FlashGetStartAddress

	Description
	-----------
        This function returns the starting address of flash to the calling
        function.
  
	Arguments
	---------
        Type:
        A byte containing the type of memory to read. 
        Either FLASH_PROGRAM or FLASH_DATA

 
	Return value
	------------
		The starting address of the type of flash memory.
  
*******************************************************************************/

unsigned long FlashGetStartAddress(memory_t Type)
{
	/* Return the Number of Flash Pages for the type of Flash Memory */
	return(Type == FLASH_PROGRAM ? PROG_FLASH_START_ADDR : DATA_FLASH_START_ADDR);
}

/******************************************************************************  
    Function: FlashInformationBlockRead

	Description
	-----------
        This function reads a word from a given location in the flash information
        block.
        Information block data is read through the register-based interface.
        Only word read operations are supported and the read address 
        must be word-aligned (LSB = 0). 
  
	Arguments
	---------

        Type:
        A byte containing the type of memory to read. 
        Either FLASH_PROGRAM or FLASH_DATA
	
        Offset:
		A byte containing the offset in the information block to read from.


        Size:
        Number of words to read from the flash 


        Data:
        A pointer to a buffer of words to store the data read from the info block

	Return value
	------------
	None.

!-Note:	Further enhancements to handle Read Protected Flash Memory.
  
*******************************************************************************/

void FlashRead(memory_t Type, 
				unsigned long Offset,
				unsigned short Size,
				unsigned char *Data )
{
	/* Get the exact location in Flash Memory */
	const unsigned long FlashStartAddress = (Type == FLASH_PROGRAM ? 
									PROGRAM_FLASH_START_ADDR : DATA_FLASH_START_ADDR);
	while (Size--)
		*Data++ = *(uint8 *)(FlashStartAddress + (Offset++));
}

/******************************************************************************
  Information Block Operations:

        FlashInformationBlockRead - Read data from  specified offset within the info Block
        FlashInformationBlockWrite - Write data at spefied offset with the info block
        FlashInformationBlockErase - Erase Info block specifed by address
       

  Notes:
        Refer to 8.2.2 8.23 and 8.2.5 of the Data Sheet for details  
******************************************************************************/


/******************************************************************************  
    Function: FlashInformationBlockRead

	Description
	-----------
        This function reads a word from a given location in the flash information
        block.
        Information block data is read through the register-based interface.
        Only word read operations are supported and the read address 
        must be word-aligned (LSB = 0). 
  
	Arguments
	---------

        Type:
        A byte containing the type of memory to read. 
        Either FLASH_PROGRAM or FLASH_DATA
	
        Offset:
	A byte containing the offset in the information block to read from.

        Data:
        A pointer to a buffer of words to store the data read from the info block

        Size:
        Number of words to read from the flash 

	Return value
	------------
	The number  of  flash words  read.

        Pseudo Code:
        -----------
        The following steps are used to read from an information block:

        Step 1: Load *IBAR and *IBDR registers with reg address 
                in function of the type of flash to access

        Step 2: Repeat the read cycle (step 3 4 ) for Size words
        Step 3: Load the word address in the Flash Memory Information
                Block Address (FMIBAR) or Flash Slave Memory Information
                Block Address (FSMIBAR) register.
        Step 4: Read the data word by reading out the Flash Memory Information
                Block Data (FMIBDR) or Flash Slave Memory
                Information Block Data (FSMIBDR) register.

*******************************************************************************/

unsigned short FlashInformationBlockRead(memory_t Type, 
					 unsigned short Offset, 
					 unsigned short *Data, 
					 unsigned short Size  )
{
  unsigned short  volatile FARIO *ibar; /* address reg */
  unsigned short  volatile FARIO *ibdr; /* data reg */
  unsigned short  count = 0;            /* number of words processed */
  
  /*!
    Address the *IBAR and *IBDR registers as per specified type of Flash 
  */
  if ( Type == FLASH_PROGRAM) {
    ibar = &FMIBAR;
    ibdr = &FMIBDR;
  }
  else {
    ibar = &FSMIBAR;
    ibdr = &FSMIBDR;
  }

  __disable_interrupt();

  while ( Size--) {
    /*!
      Address data through address and data registers
    */
    *ibar = Offset;
    *Data = *ibdr;
    Data++;
    count++;
    Offset +=2;
  }
  __enable_interrupt();
  return count;
} /*end FlashInformationBlockRead */              

/******************************************************************************  
    Function: FlashInformationBlockWrite

	Description
	-----------
                  Information Block Write
                  Writing is only allowed when global write protection is disabled.
                  Writing by the CPU is only allowed when the write enable
                  bit is set for the sector which contains the word to be
                  written. The CPU cannot write Information Block 0. 
                  Only word-wide write access to word-aligned addresses is supported.
    
  
	Arguments
	---------

        Type:
        A byte containing the type of memory to write to. 
        Either FLASH_PROGRAM or FLASH_DATA
	
        Offset:
	A byte containing the offset in the information block to write to.

        Data:
        A pointer to a buffer of words to extract  the data to write to the info block

        Size:
        Number of words to write to the flash 

	Returns
	-------

        Number of words transfered to the info block

        Pseudo Code:
        -----------
                  The following steps are performed to write in the info block:
                  Step 1: Load *STAT, *CTRL, *IBAR and *IBDR registers with reg address 
                          in function of the type of flash to access
                  Step 2: Verify that the Flash Memory Busy (FMBUSY) bit is
                          clear. The FMBUSY bit is in the FMSTAT or FSMSTAT register.
                  Step 3: Prevent accesses to the flash memory while the write is in progress.
                  Step 4: Set the Program Enable (PE) bit. The PE bit is in the FMCTRL or FSMCTRL register.
                  Step 5: Write the desired target address into the FMIBAR or FSMIBAR register.
                  Step 6: Write the data word into the FMIBDR or FSMIBDR register. 
                          This starts a  new pipelined programming sequence. 
                          The FMBUSY bit becomes set while the write operation is in progress. 
                          The FMFULL bit in the FMSTAT or FSMSTAT register becomes set if a 
                          previous write operation is still in progress.
                  Step 7: Wait until the FMFULL bit becomes clear.
                 Step 8: Repeat steps 5 through 7 for additional words.
                  Step 9: Wait until the FMBUSY bit becomes clear again.
                  Step 10: Check the programming error (PERR) bit to confirm successful programming. 
                          The PERR bit is in the FMSTAT or FSMSTAT register.
                  Step 11: Clear the Program Enable (PE) bit.

         Notes:
         ----- 
               We need to be aware of the tight loops in this code and plan for
               some timeout mechanisms either by counting or by using external hardware

*******************************************************************************/

unsigned short FlashInformationBlockWrite(memory_t Type,
					  unsigned short Offset, 
					  unsigned short *Data, 
					  unsigned short Size)
{
  unsigned short  volatile FARIO *stat; /* status reg */
  unsigned short  volatile FARIO *ctrl; /* control reg */
  unsigned short  volatile FARIO *ibar; /* address reg */ 
  unsigned short  volatile FARIO *ibdr; /* data reg */

  unsigned short count = 0;
  
  /* 
     Step 1: Address appropriate regs for memory type 
  */
  if ( Type == FLASH_PROGRAM) {
    stat = &FMSTAT;
    ctrl = &FMCTRL;
    ibar = &FMIBAR;
    ibdr = &FMIBDR;
  }
  else {
    stat = &FSMSTAT;
    ctrl = &FSMCTRL;
    ibar = &FSMIBAR;
    ibdr = &FSMIBDR;
  }
  __disable_interrupt();
  /*! 
     Step 2:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
  */
  while(*stat & FMBUSY) ;

  /*! 
     Step 3: Prevent accesses to the flash memory while the write is in progress. 
  */
  /* TBD */

  /*! 
     Step 4: Set the Program Enable (PE) bit. The PE bit is in the FMCTRL or FSMCTRL register. 
  */
  *ctrl |= PE;



  /*!
    Process the specified number of word  
   */
  while( Size--) {
    /*! 
       Step 5: Write the desired target address into the FMIBAR or FSMIBAR register. 
    */
    *ibar = Offset;
    Offset += 2;


    /*! 
       Step 6: Write the data word into the FMIBDR or FSMIBDR register. 
    */
    *ibdr = *Data;
    Data++;
    
    while(*stat & FMBUSY) ;

    /*! 
       Step 7: Wait until the FMFULL bit becomes clear. 
    */
    while(*stat & FMFULL) ;

  /*! 
     Step 8: Repeat steps 5 through 7 for additional words. 
  */  
    count++;
  } /* while Size */


  /*! 
     Step 9: Wait until the FMBUSY bit becomes clear again. 
  */
  while(*stat & FMBUSY) ;

  /*! 
     Step 9: Check the programming error (PERR) bit to confirm successful programming.
     The PERR bit is in the FMSTAT or FSMSTAT register. 
   */
  if ( (*stat & PERR) == PERR ) {
    flashStatus = FLASH_PROG_ERR;
  }

  /*! 
     Step 10: Clear the Program Enable (PE) bit. 
  */
  *ctrl &= ~PE;

  __enable_interrupt();
  return count;
} /* end FlashInfoBlockWrite */


/******************************************************************************  
    Function: FlashInformationBlockErase

	Description
	-----------
       Information Block Module Erase
       Erasing an information block also erases the corresponding
       main block. Page erase is not supported for information blocks.
       
  
	Arguments
	---------
	Type: 
              the type of information block DATA or PROGRAM
        Offset:
              Any Adress within the block

        Data:
              Any Data
	

	Returns
	-------
              None

        Pseudo Code:
        -----------

        The following steps are performed to erase an information block:
        Step 1. Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
                The FMBUSY bit is in the FMSTAT or FSMSTAT register.
        Step 2. Prevent accesses to the flash memory while erasing is in progress.
        Step 3. Set the Module Erase (MER) bit. 
                The MER bit is in the FMCTRL or FSMCTRL register.
        Step 4. Load the FMIBAR or FSMIBAR register with any address within the block, 
                then write any data to the FMIBDR or FSMIBDR register.
        Step 5. Wait until the FMBUSY bit becomes clear again.
        Step 6. Check the Erase Error (EERR) bit to confirm successful
                erase of the block. The EERR bit is in the FMSTAT or FSMSTAT register.
        Step 7. Clear the MER bit.


                                           
*******************************************************************************/

void FlashInformationBlockErase(memory_t Type, 
				unsigned short Offset, 
				unsigned char Data )
{
 
  unsigned short  volatile FARIO *stat; /* status reg */
  unsigned short  volatile FARIO *ctrl; /* ctrl reg */
  unsigned short  volatile FARIO *ibar; /* address reg */
  unsigned short  volatile FARIO *ibdr; /* data reg */

  /*! 
     Step 1: Address the appropriate register based on the type of Flash to access
  */
  if ( Type == FLASH_PROGRAM) {
    stat = &FMSTAT;
    ctrl = &FMCTRL;
    ibar = &FMIBAR;
    ibdr = &FMIBDR;
  }
  else {
    stat = &FSMSTAT;
    ctrl = &FSMCTRL;
    ibar = &FSMIBAR;
    ibdr = &FSMIBDR;
  }

  __disable_interrupt();
  /*! 
     Step 2:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
  */
  while(*stat & FMBUSY) ;

  /*!
    Step 3: Prevent access to the flash memory while erasing is in progress 
  */
  /* TBD */

  /*! 
     Step 4: Set MER  bit in FMCTRL to enable programing 
  */
  *ctrl |= MER;

  /*! 
     Step 5: Load the FMIBAR or FSMIBAR register with any address within the block, 
             then write any data to the FMIBDR or FSMIBDR register.
  */ 
  *ibar = Offset;
  *ibdr = Data;

  /*!
    Step 6:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
  */
  while(*stat & FMBUSY) ;

  /*! 
     Step 7: Check the Erase Error (EERR) bit to confirm successful erase of the block.
             The EERR bit is in the FMSTAT or FSMSTAT register.
   */
  if ( (*stat & EERR) == EERR ) {
   flashStatus = FLASH_ERASE_ERR;
  }

  /*!  
    Step 8: Clear the MER bit. 
  */
  *ctrl &= ~MER;


  __enable_interrupt();
} /* end FlashInfoBlockErase */

/**********************************************************************************

	Function: FlashPageProgramSafe

	Description
	-----------
        This function provides a generalized mechanism to write to the flash memory main 
        blocks. It assumes that the data to be written is not necessarly aligned on a 
        page boundary. The write cycle implies that an erase operation has taken place
        prior to programing. The algorythm here is as follows 
          a) locate a place to save a page worth of data 
          c) erase it to allow later writting operation
          b) copy the current page to save area
          d) erase current page
          e) write current page with data from saved and new data ( partial page write are 
             thusly supported)
	Arguments
	---------
	Type: 
             Type of memory to program
        Offset:
             offset within the block of the first byte to program
        Size:
             Number of bytes to program
        InBuf
            Buffer containing data to program to flash

	Return value
	------------
	None

        Pseudo Code:
        ------------
           Step 1: Allocate a Work Page worth of storage 
                   if dynamic allocation fails, we use the last page of the appropriate block
                   as a scratch pad.
           Step 2: Erase Work Page. 
           Step 3: Write Current Page to Work Page.
           Step 4: Erase Current Page.
           Step 5: Write Current Page with Work Page Data and Specified Input Data.
		   Step 6: Repeat steps 1 through 5 till all the data has been writen.


**********************************************************************************/

void FlashPageProgramSafe (memory_t Type, 
		  unsigned long Offset, 
		  unsigned short Size,  
		  unsigned char *InBuf)
{
  unsigned char *workPagePtr = NULL;      /* Address of the save/work area  either RAM or FLASH */
  unsigned short pageSize=0;                /* size of the page to write */
  unsigned char  *startAddress=0;           /* start address of the flash main bock(s) */
  unsigned char  *address=0;                /* address within the page */
  unsigned long pageOffset=0;               /* page address in which our data resides */ 
  unsigned long numberOfPages=0;            /* total number of pages in this flash main block(s) */
  boolean isDynamic = TRUE;
  unsigned int offset_in_page;
  unsigned int page_number;

  /*!
    Step 1: Address and calculate parameters for the specified type of Flash
            pageSize, startAddress, numberOfPages, pageOffset, address
  */
  if ( Type == FLASH_PROGRAM ) {
    pageSize = FLASH_PROGRAM_PAGE_SIZE;
    startAddress = PROGRAM_FLASH_START_ADDR ;
    numberOfPages = 256 ;
    pageOffset = Offset & 0x3FC00;
	page_number = Offset >> 10;

  }
  else {
    pageSize = FLASH_DATA_PAGE_SIZE;
    startAddress = (unsigned char *)DATA_FLASH_START_ADDR;
    numberOfPages = 16;
    pageOffset = Offset & 0xFE00; 
	page_number = Offset >> 9;
  }
  
  /*!
    Step 2: allocate or locate an appropriate save area. 
  */
  workPagePtr = (unsigned char *)malloc(pageSize);

  /* 
     if the heap data is not available try the last page of the flash main block(s)
  */
  if ( workPagePtr == NULL ) {
    
    workPagePtr = startAddress + ((numberOfPages-1) * pageSize);
    isDynamic = FALSE; /* remember we are now flash */
  }

	/* Find out how much data can be stored in the flash page */
	offset_in_page = Offset - (pageSize * page_number);
	
	/* Break the data into make it page aligned */
	while(Size)
	{
		int page_data_size;
		
		address =  pageOffset + startAddress; 

		/* Check if all the data fits into a single page or not */
		page_data_size = ((offset_in_page + Size) > pageSize ? (pageSize - offset_in_page) : Size);
			
		/*!
			Step 3: Save the contents of the current page to scratch page 
		*/
		if ( isDynamic ) {

			/* erase ram */
			memset(workPagePtr, 0xFF, pageSize);

			/* Write Current page to work Page */
			memcpy(workPagePtr, address , pageSize);
		}
		else {

			/* Erase Last Page */
			FlashPageErase(Type, (unsigned long)workPagePtr);

			/* 
			Write Current Page to Work Area 
			(This save the current content of this page on the work page) 
			*/
			FlashPageProgram(Type,(unsigned long)workPagePtr, pageSize,
								   address,(unsigned char *)pageOffset);      
		}
		/*! 
		 Step 4: Erase the current one 
		*/ 
		FlashPageErase(Type, (unsigned long)address);
	  
		/*! 
		 Step 5: Write the page with old and new data as required 
		*/
		FlashPageProgram(Type, Offset, page_data_size, workPagePtr, InBuf);	          
	
	
		/* If the data size is beyond the page boundary then make sure that the offset is from the 
		   beginning of the next page */
		if(page_data_size < Size) {
			
			offset_in_page = 0;

			/* Increment the data pointer to point to the remainder of the data */
			InBuf += page_data_size;

			/* Increment to point to the next page */
			pageOffset += pageSize;

			/* Data is going to be continued onto the first byte on the next page */
			Offset = pageOffset;	
		}

		/* Decrement the amount of data to be read from the total size of the data to be read */
		Size -= page_data_size;

	}		/* Step 6: Loop till all the data has been written */

         
  if ( (isDynamic == TRUE) && (workPagePtr != NULL)) {
    free(workPagePtr);
  }	  
  

} /* end  FlashPageProgramSafe */

/**********************************************************************************

	Function: FlashPageProgram

	Description
	-----------
        This function always writes a page worth of data. 
        The data is obtained from an input Buffer and a Reference Buffer
        The specified size indicates the number of bytes from a page offset up to the end of the 
        page and therefore can not exceed the page size. The page offset is calculated from the 
        specified offset with the specified Flash Block(s).
        The algorithm used here is complicated by the fact that the size of the modified data 
        is not always exactly a page. Since we have to write the whole page everytimes we want to 
        make a change, we need to have read the page in some other buffer, erased the page
        and modify the data as we write it back. This routine assumes that those two operations
        have been effectued before invocation. The Reference Buffer contains the copy of the
        data that was in this page.

        See 8.3.6 for Main Block Write
        Writing is only allowed when global write protection is disabled.
        Writing by the CPU is only allowed when the write enable bit is set for the sector 
        which contains the word to be written.
        The CPU cannot write the Boot Area. 
        Only wordwide write access to word-aligned addresses is supported. 
  
	Arguments
	---------
	     Offset : an offset from the start of the block(s) of Flash memory ( Program or Data)
                  This is used to determine the Page offset of the page where the data will be
                  written. 
         Size: a number of byte to be written ( this can not exceed the size of a Page)
         RefBuf: A reference buffer where the content of the page to write has been saved.
                 Possibly Null if we are writting the all page
         InBuf: A buffer containing the new data to be written to the page.

	Return value
	------------
	None

        Pseudo Code:
        ------------


         The following steps are performed to write a word:

         Step 1: Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
            The FMBUSY bit is in the FMSTAT or FSMSTAT register.

         Step 2: Prevent accesses to the flash memory while the write is in progress.

         Step 3: Set the Program Enable (PE) bit. The PE bit is in the FM-CTRL or FSMCTRL register.

         Step 4: Write a word to the desired wordaligned address. 
                 This starts a new pipelined programming sequence. 
                 The FM-BUSYbit becomes set while the write operation is in progress. 
                 The FMFULL bit in the FMSTAT or FSMSTAT register becomes set 
                 if a previous write operation is still in progress.

         Step 5: Wait until the FMFULL bit becomes clear.

         Step 6: Repeat steps 4 and 5 for additional words.

         Step 7: Wait until the FMBUSY bit becomes clear again.

         Step 8: Check the programming error (PERR) bit to confirm successful
                 programming. The PERR bit is in the FMSTAT or FSMSTAT register.

         step 9. Clear the Program Enable (PE) bit.



**********************************************************************************/
void FlashPageProgram(memory_t Type, 
		  unsigned long Offset, 
		  unsigned short Size, 
		  unsigned char *RefBuf, 
		  unsigned char *InBuf)
{

     unsigned short  volatile FARIO *stat;            /* status reg */
     unsigned short  volatile FARIO *ctrl;            /* control reg */
     unsigned short  volatile FARIO *wer;             /* write enable reg */

     unsigned char   *refPtr = RefBuf;                /* The address where the reference data lives */
     unsigned char   *modPtr = InBuf;                 /* The new data buffer */
     unsigned char   *address;                         /* address of the start of the page we want to modify */
     
     unsigned long   pageOffset=0L;                     /* offset of the page */
     int  pageSize=0;                                   /* size in bytes of a flash page */
     int  upperLimit=0;                                 /* upper limit of the range within the page */
     int  lowerLimit=0;                                 /* lower limit of the range within the page */
     int  i;                                            /* an index for iteration thru buffer */

     unsigned short  writeEnableBit;                  /* a 16 bit wide mask for the we regs */ 
     unsigned short  work=0;                          /* a word to be written to flash location */
                   
     unsigned char   curByte;                         /* current byte under processing */ 
     boolean alreadyEnableBit = FALSE;                /* indicator of some other process trying to access */
 
 /*! 
    Address appropriate register based on memory type 
 */
  if ( Type == FLASH_PROGRAM ) {
    stat = &FMSTAT;
    ctrl = &FMCTRL;
    wer  = &FM1WER;
    writeEnableBit = Offset >> 13; /* 8K */
    if ( Offset  <  FM1WER_START ) {
     wer = &FM0WER;   
    }
    else {
      writeEnableBit %=16;
    }
    
    pageSize = FLASH_PROGRAM_PAGE_SIZE;
    /* Config info as well */
    /* for program flash the offset is in the range 0x000000 - 0x03ffff (256K) */
    /* each page is 1024 bytes. 2^10. to get page offset within the block */
    pageOffset = (Offset & 0x3FC00);
    address =  (unsigned char *)(PROGRAM_FLASH_START_ADDR + pageOffset);
  } 
  else {
    stat = &FSMSTAT;
    ctrl = &FSMCTRL;
    wer  = &FSM0WER;                               
    writeEnableBit = Offset >> 9; /* 512 bytes */
    pageSize = FLASH_DATA_PAGE_SIZE;
    /* Config info as well */
    /* for data flash the offset is in the range 0x0000 - 0x1fff (8k) - start address at 0x0e0000 */
    /* each page is 512 bytes. 2^9 to get page offset within the block */
    pageOffset = (Offset & 0xFE00);
    address = (unsigned char *) (DATA_FLASH_START_ADDR + pageOffset);
  }
 
  if ( Type == FLASH_PROGRAM ) {
	  /* Let's have the CPU attention for the duration of this operation */
	  __disable_interrupt();
  }
  /*!    
	Step 1:  Enable writing in the section corresponding to this page. 
  */
  if (! (*wer & ( 1 << writeEnableBit)) ) {
    *wer |= (1 << writeEnableBit);
  }
  else {
    alreadyEnableBit = TRUE;
  }

  /*! 
     Step 2:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
              The FMBUSY it is in the FMSTAT or FSMSTAT register. 
  */
  while(*stat & FMBUSY) ;

  /*! 
     Step 3:  Prevent accesses to the flash memory while write  is in progress. 
  */
  /* TBD */

  /*! 
    Step 4: Set the Program Enable (PE) bit. The PE bit is in the FM-CTRL or FSMCTRL register. 
  */
  *ctrl |= PE;

  /* 
     compute upper and lower limits of the range to program 
  */
  lowerLimit = Offset - pageOffset;
  upperLimit = lowerLimit + Size;


  /*! 
    Step 5: Process all bytes from ref page and new data 
  */
 

  for ( i = 0 ; i < pageSize ; i++) {

    if ( i < lowerLimit || i >= upperLimit ) {
      /* out of range -> we take ref byte */
      curByte = *refPtr++;
    }
    else {
      /* in range-> we take new byte */
      curByte = *modPtr++;

	  /* Increment the Page pointer on the original Page to skip the modified bytes */
	  refPtr++;
    }
    /* odd bytes get orded with previous to build a word (lsb,msb)*/
    if ( i & 1 ) {
      work |= ((unsigned short) curByte << 8);

      /*!
	Wait for programming to complete 
      */
      while(*stat & FMFULL) {
        ;
      }

      /*!
	Write the constructed word 
      */
      *((FAR short *)address) = work;
      address += 2;            
      while(*stat & FMBUSY) {
        ;
      }     
      work = 0;    
    }
    /* even byte have to wait for the word complement on next iteration */
    else {
      work = curByte;
    }   
  } /* end of page */
 
  while(*stat & FMBUSY) {
    ;
  }
 
  if ( (*stat & PERR) == PERR ) {
    flashStatus = FLASH_PROG_ERR;
  }
  
  *ctrl &= ~PE;
  if (!alreadyEnableBit) { 	
    *wer &= ~(1 << writeEnableBit);
  }
  
  while(*stat & FMBUSY) {
    ;
  }
   
   if ( Type == FLASH_PROGRAM ) {
	/* Let's other device a chance at CPU resource */
	 __enable_interrupt();
   }

} /* end FlashProgram */	




/**********************************************************************************

	Function: FlashEraseAll
	
	Description
	-----------
        This function clears all of the specified portion of the flash, 
        and sets everything to 0xFF.
  
	Arguments
	---------
	Type: an indicator of the kind of flash we are earsing

	Return value
	------------
	None
	
**********************************************************************************/

void FlashEraseAll(memory_t Type )
{
  unsigned short pageIndex;              /* an interator index thru all the page of main block(s) */
  unsigned long  pageAddress;            /* address of the current page to erase */
  unsigned short pageSize;               /* size of a page of flash main block */ 
  unsigned short numberOfPages;          /* number of pages in a block */

  if ( Type == FLASH_PROGRAM ) {
    pageAddress = PROGRAM_FLASH_START_ADDR;
    numberOfPages = 256;
    pageSize = FLASH_PROGRAM_PAGE_SIZE;
  }
  else {
    pageAddress = DATA_FLASH_START_ADDR;
    numberOfPages = 16;
    pageSize = FLASH_DATA_PAGE_SIZE;
  }

  /*!
    Iterate thru the pages of the specified block and perform erase operatio on each one 
  */

  for(pageIndex = 0 ; pageIndex < numberOfPages; pageIndex++, pageAddress += pageSize) {
      FlashPageErase(Type, pageAddress);
  }
} /* end FlashEraseAll */ 


/******************************************************************************  
    Function: FlashPageErase

	Description
	-----------
         Main Block Page Erase
         A flash erase operation sets all of the bits in the erased region.
         Pages of a main block can be individually erased if their write enable bits are set. 
         Each page in Main Block 0 and 1 consists of 1024 bytes (512 words). 
         Each page in Main Block 2 consists of 512 bytes (256 words). 
      
	Arguments
	---------
	
        Type: 
             Type of the memory page to erase

	Address:
             address of the page to erase
  

	Return value
	------------

        None

        Pseudo Code:
        -----------

        To erase a page, the following steps are performed:
        Step 1:  Enable writing in the section corresponding to this page.
        Step 2:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
                 The FMBUSY bit is in the FMSTAT or FSMSTAT register.
        Step 3:  Prevent accesses to the flash memory while erasing is in progress.
        Step 4:  Set the Page Erase (PER) bit. The PER bit is in the FMCTRL or FSMCTRL register.
        Step 5:  Write to an address within the desired page.
        Step 6:  Wait until the FMBUSY bit becomes clear again.
        Step 7:  Check the Erase Error (EERR) bit to confirm successful erase of the page. 
	         The EERR bit is in the FMSTAT or FSMSTAT register.
        Step 8:  Repeat steps 4 through 6 to erase additional pages. 
                 ( not implemented in this fashion to be discussed in review)
        Step 9:  Clear the PER bit.
        Step 10: Disable writing in the section corresponding to this page 

        Notes: 
        -----
        
        Consideration of the write enable bits in wer registers:
        Block 0 and 1 contain 16 sections each of 8 page of 1k each
        Block 2 contains 16 section of 1 page of 512 bytes each
        the register wer ( fm0, fm1, and fsm 0 ) are used to keep track of the 
        sections that are protected. We remember that at reset all the pages are protected 
        and in order to access them the bit corresponding to the section to access needs to 
        be cleared before any operation can take place. 
        we deduce the section number from the address by dividing the address by the 
        size of a section. for block 0 and 1 a >>13 will divide by 2^13= 8K and give the 
        section number or for block 2 by 2^9= 512 for the same effect. 
        if the address lays in the main block 1 of program data the calculated bit has to 
        be taken modulo 16
        this section number is then used to address the corresponding bit in the appropriate 
        wer reg.
        if the corresponding bit is already cleared we assume someone is already manipulating 
        the section.
 
*******************************************************************************/

void FlashPageErase(memory_t Type, unsigned long Address)
{
  unsigned short  volatile FARIO *stat; /* status reg */
  unsigned short  volatile FARIO *ctrl; /* control reg */
  unsigned short  volatile FARIO *wer;  /* write enable reg */

  unsigned short  writeEnableBit;       /* a mask to enable/disable write to the corresponding section */
  boolean alreadyEnableBit = FALSE;	
  boolean alreadyPrograming = FALSE;
  
  /*! 
     Address appropriate register based on memory type 
  */
  if ( Type == FLASH_PROGRAM ) {
    stat = &FMSTAT;
    ctrl = &FMCTRL;
    wer  = &FM1WER;  /* assume block 1 until we know more about the address */
    /*!
      Compute which section bit to enable/disable
    */
    writeEnableBit = (Address - PROGRAM_FLASH_START_ADDR) >> 13; /* 8K */
    
    /* within the first block address the appropriate reg   */
    
    if ( Address <  FM1WER_START ) {
     wer = &FM0WER;    
    }
    else {
      /*! 
	for address in block 0 the bit offset is taken modulo 16 to be within the short 
      */
      writeEnableBit %= 16;
    }

  } 
  else {
    stat = &FSMSTAT;
    ctrl = &FSMCTRL;
    wer  = &FSM0WER;                               
    writeEnableBit = (Address - DATA_FLASH_START_ADDR) >> 9; /* 512 bytes */
  }

	if ( Type == FLASH_PROGRAM ) {
	  /*! 
		 Step 1:  Enable writing in the section corresponding to this page.
		 We are further checking if another module is already enabling this bit 
		 and if another module is already programing 
	  */
	  __disable_interrupt();
	}

  if (! (*wer & ( 1 << writeEnableBit)) ) {
    *wer |= (1 << writeEnableBit);
  }
  else {
    alreadyEnableBit = TRUE;
  }

  if ((*ctrl & PE) == PE ) {
    *ctrl &= ~PE;
    alreadyPrograming = TRUE;
  }

  /*! 
    Step 2:  Verify that the Flash Memory Busy (FMBUSY) bit is clear. 
    The FMBUSY bit is in the FMSTAT or FSMSTAT register. 
  */
  while(*stat & FMBUSY) ;

  /*!
    Step 3:  Prevent accesses to the flash memory while erasing is in progress. 
  */


  /*! 
     Step 4:  Set the Page Erase (PER) bit. The PER bit is in the FMCTRL or FSMCTRL register. 
  */
  *ctrl |= PER;

  /*! 
     Step 5:  Write to an address within the desired page.
  */
  *((FAR unsigned short *)Address) = 0xFFFF;    // Perform write - !! ERASES PAGE !!
                                        // Note: use __far as address can lie beyond
                                        // 64K boundary

  /*! 
     Step 6:  Wait until the FMBUSY bit becomes clear again. 
  */
  while(*stat & FMBUSY) ;

  /*! 
    Step 7:  Check the Erase Error (EERR) bit to confirm successful erase of the page. 
    The EERR bit is in the FMSTAT or FSMSTAT register. 
  */
  if ( (*stat & EERR) == EERR ) {
    flashStatus = FLASH_ERASE_ERR;
  }
  /*!
    Step 8:  Repeat steps 4 through 6 to erase additional pages.
    For now we handle only one page in this routine 
  */
 
  /*! 
    Step 9:  Clear the PER bit. 
  */
  *ctrl &= ~PER;
 
  /*!  
    Step 10: Disable writing in the section corresponding to this page 
  */
  if (!alreadyEnableBit) { 	
    *wer &= ~(1 << writeEnableBit);
  }
  if ( alreadyPrograming) {
    *ctrl |= PE;
  }

  if ( Type == FLASH_PROGRAM ) {
	__enable_interrupt();
  }
} /* end FlashPageErase */




