/************************************************************************
*	File Name:	gpio.c
*
*	Function:	This is the General Purpuse I/O (GPIO) driver.
*
************************************************************************/


#include "chipdef.h"
#include "drivers.h"

#ifdef PB_BASE_ADDR 
gpio_t * const Bport = (gpio_t *) PB_BASE_ADDR;
#endif

#ifdef PC_BASE_ADDR
gpio_t * const Cport = (gpio_t *) PC_BASE_ADDR;
#endif

#ifdef PE_BASE_ADDR
gpio_t * const Eport = (gpio_t *) PE_BASE_ADDR;
#endif

#ifdef PF_BASE_ADDR
gpio_t * const Fport = (gpio_t *) PF_BASE_ADDR;
#endif

#ifdef PG_BASE_ADDR
gpio_t * const Gport = (gpio_t *) PG_BASE_ADDR;
#endif

#ifdef PH_BASE_ADDR
gpio_t * const Hport = (gpio_t *) PH_BASE_ADDR;
#endif

#ifdef PI_BASE_ADDR
gpio_t * const Iport = (gpio_t *) PI_BASE_ADDR;
#endif

#ifdef PJ_BASE_ADDR
gpio_t * const Jport = (gpio_t *) PJ_BASE_ADDR;
#endif

static gpio_t * const port_tab[] =
{
#ifdef PA_BASE_ADDR 
	(gpio_t *) PA_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PB_BASE_ADDR 
	(gpio_t *) PB_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PC_BASE_ADDR 
	(gpio_t *) PC_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PD_BASE_ADDR 
	(gpio_t *) PD_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PE_BASE_ADDR 
	(gpio_t *) PE_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PF_BASE_ADDR 
	(gpio_t *) PF_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PG_BASE_ADDR 
	(gpio_t *) PG_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PH_BASE_ADDR 
	(gpio_t *) PH_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PI_BASE_ADDR 
	(gpio_t *) PI_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PJ_BASE_ADDR 
	(gpio_t *) PJ_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PK_BASE_ADDR 
	(gpio_t *) PK_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

#ifdef PL_BASE_ADDR 
	(gpio_t *) PL_BASE_ADDR,
#else
	(gpio_t *) NULL,
#endif

};

/******************************************************************************  
    Function: gpio_set_alt

	Description
	-----------
    This function configures the function of one or more GPIO pins in a port.
	It either sets the pins to serve their alternate functions (i.e., one of
	the peripherals) or to simply serve as GPIOs.

	Arguments
	---------
	port:
	An argument of type port_t which indicates the GPIO port to which the
	pins belong.

	mask:
	A byte mask which indicates the pins position (for each pin involved
	in this operation one bit is set in the corresponding position).

	mode:
	A boolean argument which indicates whether the pins will serve their
	alternate functions or as GPIOs. TRUE = alternate functions. FALSE = GPIOs.
*******************************************************************************/

void gpio_set_alt(port_t port, unsigned char mask, boolean mode)
{
	gpio_t *port_base_address = port_tab[(unsigned int)port];

	if (port_base_address == NULL)
		return;

	if (mode)
		port_base_address->pxalt |= mask;
	else
		port_base_address->pxalt &= (~mask);
		
}

/******************************************************************************  
    Function: gpio_set_alt_source

	Description
	-----------
    This function configures the alternate function source of one or more GPIO
    pins in a port. Some pins have two alternate functions, A and B and if
	the pin is set to serve its alternate function, then the alternate function
	source has to be set to A or B.
	It either sets the pins to serve their alternate functions (i.e., one of
	the peripherals) or to simply serve as GPIOs.

	Arguments
	---------
	port:
	An argument of type port_t which indicates the GPIO port to which the
	pins belong.

	mask:
	A byte mask which indicates the pins position (for each pin involved
	in this operation one bit is set in the corresponding position).

	source:
	An argument of type gpio_alt_src_t which determines the alternate source.
	GPIO_ALT_SRC_A = source A. GPIO_ALT_SRC_B = source B.
*******************************************************************************/

void gpio_set_alt_src(port_t port, unsigned char mask, gpio_alt_src_t source)
{
	gpio_t *port_base_address = port_tab[(unsigned int)port];

	if (port_base_address == NULL)
		return;

	if (source == GPIO_ALT_SRC_B)
		port_base_address->pxalts |= mask;
	else
		port_base_address->pxalts &= (~mask);
		
}

/******************************************************************************  
    Function: gpio_set_dir

	Description
	-----------
    This function sets the direction of one or more GPIO pins in a port.

	Arguments
	---------
	port:
	An argument of type port_t which indicates the GPIO port to which the
	pins belong.

	mask:
	A byte mask which indicates the pins position (for each pin involved
	in this operation one bit is set in the corresponding position).

	dir:
	An argument of type gpio_dir_t which indicates whether the pins direction
	should be input or output.
*******************************************************************************/
void gpio_set_dir(port_t port, unsigned char mask, gpio_dir_t dir)
{
	gpio_t *port_base_address = port_tab[(unsigned int)port];

	if (port_base_address == NULL)
		return;

	if (dir == GPIO_INPUT)
		port_base_address->pxdir &= (~mask);

	else
		port_base_address->pxdir |= mask;
}

/******************************************************************************  
    Function: gpio_write

	Description
	-----------
    This function writes to one or more GPIO pins in a port.

	Arguments
	---------
	port:
	An argument of type port_t which indicates the GPIO port to which the
	pins belong.

	mask:
	A byte mask which indicates the pins position (for each pin involved
	in this operation one bit is set in the corresponding position).

	data:
	An data byte containing the data to be written to the pins. For each pin
	involved in this operation the data bit will be in the corresponding
	position. Bit position for pins that are not involved (not set in the
	mask argument) will be ignored.
*******************************************************************************/
void gpio_write(port_t port, unsigned char mask, unsigned char data)
{
	gpio_t *port_base_address = port_tab[(unsigned int)port];

	if (port_base_address == NULL)
		return;

	/* Clear all bits in the data byte which do not correspond to pins 
	   involved in the operation.
	*/
	data &= mask;

	/* Now read the port clear pin bits and or with data. */
	port_base_address->pxdout = (port_base_address->pxdout & (~mask)) | data;
										 
}

/******************************************************************************  
    Function: gpio_read

	Description
	-----------
    This function reads from one or more GPIO pins in a port.

	Arguments
	---------
	port:
	An argument of type port_t which indicates the GPIO port to which the
	pins belong.

	mask:
	A byte mask which indicates the pins position (for each pin involved
	in this operation one bit is set in the corresponding position).

	Return Value
	------------
	An data byte containing the data hat was read from the pins. For each pin
	involved in this operation the data bit will be in the corresponding
	position. Bit position for pins that are not involved (not set in the
	mask argument) will be cleared and should be ignored anyway.
*******************************************************************************/
unsigned char gpio_read(port_t port, unsigned char mask)
{
gpio_t *port_base_address = port_tab[(unsigned int)port];

	if (port_base_address == NULL)
		return 0;

	return (port_base_address->pxdin & mask);
	
}



