/******************************************************************************
*  File:	mft.c
*  
*  This is Multi-Function Timer peripheral driver.
******************************************************************************/
#include <stdio.h>
#include "chipdef.h"

mft_t * const mft = (mft_t *)MFT_BASE_ADDR;

/* Flag to indicate to the calling program that the Interrupt has occured */
volatile unsigned int tick_flag_1;
volatile unsigned int tick_flag_2;
/* Variable to store the value of the time interval in milli seconds*/
volatile long TimerInterval_1;
volatile long TimerInterval_2;
volatile long TimerInterval_3;


/******************************************************************************  
    Function: mft_init

	Description
	-----------
    This function initializes the MFT hardware in the following modes:

    Mode 1 (single PWM and single system timer):
      - Timer I: generates a single 16-bit PWM signal on TA pin
      - Timer II: generate a clock tick every specified period of time

    Mode 3 (dual independant timers):
      - generates a clock tick every specified period of time for each timer
  
	Arguments
	---------
	Mode:
    - selects the mode of operation
  ClockTick1:
	  Mode 1:
      - the high/low pulse width
    Mode 3:
      - the required length of period between clock ticks in milliseconds for timer I
	ClockTick2:
	  Mode 1 and 3:
      - the required length of period between clock ticks in milliseconds for timer II
	ClockTick3:
    Mode 1:
      - the low/high pulse width
    Mode 3:
      - unused
      
      
*******************************************************************************/

void mft_init(const int mode, const int ClockTick1, const int ClockTick2, const int ClockTick3)
{
  if (mode == 1){
  	/* Clear the flag that indicate to the calling program that the Interrupt has occured */
  	tick_flag_2 = 0;
  
  	/* Calculate the value of the Timer Counter Value */
  	TimerInterval_1 = ((long)ClockTick1 * 32768) / 1000;
  	TimerInterval_2 = ((long)ClockTick2 * 32768) / 1000;
  	TimerInterval_3 = ((long)ClockTick3 * 32768) / 1000;
  
  	/* Enable the MFT module */
    TCTRL |= TMDSEL_MODE1;  
    /* Enable the TA PWM function on PG.7 - if required */
    MFTA_MUX_PORT->pxalt |= 0x80;    
    MFTA_MUX_PORT->pxalts &= 0x7F;    
    /* Enable the TA PWM output - if required */
    MFTA_MUX_PORT->pxdir |= 0x80;
  	/* Program the PWM Timer pulse widths to the Timer Interval */
  	TCRA = TimerInterval_1;
  	TCRB = TimerInterval_3;
  
  	/* Program the Timer ticks to the Timer II Interval */
  	TCNT2 = TimerInterval_2;

  	/* Clear the TA, TB and TD Interrupt Pending Bits */
  	TICLR = TACLR|TBCLR|TDCLR;

  	/* Disable the TA interrupt and enable the TD interrupt */
  	TICTL = TDIEN;
  
  	/* Select the slow clock as the clock source for both timers */
  	TCKC = 0x24;
  	/* Select the the prescaled system clock as the source for timer I (PWM) 
    and slow clock as the clock source for timer II*/
//  	TCKC = 0x21;
  
  	/* set pre-scalar to 1 */
  	TPRSC = 0x00;
  
  	/* Unmask the Timer II Interrupt*/
//  	icu_unmask(MFTA_IRQ);
  	icu_unmask(MFTB_IRQ);
 
   	/* Put Timers in PWM timer mode and enables the TA output*/
    TCTRL |= TAEN|TEN;  
  }

  if (mode == 3){
  	/* Clear the flags that indicate to the calling program that the Interrupt has occured */
  	tick_flag_1 = 0;
  	tick_flag_2 = 0;
  
  	/* Calculate the value of the Timer Counter Values */
  	TimerInterval_1 = ((long)ClockTick1 * 32768) / 1000;
  	TimerInterval_2 = ((long)ClockTick2 * 32768) / 1000;
  	
  	/* Clear the TA and TD Interrupt Pending Bits and enable the TA and TD interrupts */
  	TICLR = TACLR|TDCLR;
  	TICTL |= TAIEN|TDIEN;

  	/* Disable the outputs, just in case they were previously enabled in mode 1 */
    MFTA_MUX_PORT->pxalt &= 0x7F;    
    MFTA_MUX_PORT->pxalts &= 0x7F;    
    /* Disable the TA PWM output - if it was previously enabled */
    MFTA_MUX_PORT->pxdir &= 0x7F;
  	/* Select the clock source for both timers */
  	TCKC = 0x24; /* slow clock */
//  	TCKC = 0x09; /* system clock */
  
  	/* Set pre-scalar to 1 */
  	TPRSC = 0x00; /* prescaler = 1 */
//  	TPRSC = 0x1f; /* prescaler = 32 */
    
  	/* Program the Timer ticks to the Timer Interval */
  	TCNT1 = TimerInterval_1;
  	TCNT2 = TimerInterval_2;
  
  	/* Unmask the Timer Interrupts*/
  	icu_unmask(MFTA_IRQ);
  	icu_unmask(MFTB_IRQ);
  
  	/* Put Timers in dual independant timer mode*/
    TCTRL |= TMDSEL_MODE3|TEN;  
  }

}

void mft_delay(const int mode, const int Channel, unsigned short Dly)
{
  if (mode == 1){
    if (Channel == 2) {
      TCNT2  =   Dly;                            /* Load MFT counter with indicated delay counts         */
      TICLR  |=  TDCLR;                          /* Clr any pending underflow flag                       */
  
      while (!tick_flag_2); {                    /* Wait here for Timer 2 flag                           */
      }
  
      tick_flag_2 = 0;
    }
  }
  if (mode == 3){
    if (Channel == 1) {
      TCNT1  =   Dly;                            /* Load MFT counter with indicated delay counts         */
      TICLR  |=  TACLR;                          /* Clr any pending underflow flag                       */
  
      while (!tick_flag_1);                      /* Wait here for underflow                              */
  
      tick_flag_1 = 0;;
    }
    if (Channel == 2) {
      TCNT2  =   Dly;                            /* Load MFT counter with indicated delay counts         */
      TICLR  |=  TDCLR;                          /* Clr any pending underflow flag                       */
  
      while (!tick_flag_2); {                    /* Wait here for Timer 2 flag                           */
      }
  
      tick_flag_2 = 0;
    }
  }
}


/*Interrupt Handler*/
#ifdef __CR__
#pragma interrupt (mfta_isr)
void mfta_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+MFTA_IRQ)
__interrupt void mfta_isr(void)
#else
__interrupt[(16+MFTA_IRQ)*IDT_WIDTH] void mfta_isr(void)
#endif
#endif
{
  /*Reload the counter*/
  TCNT1 = TimerInterval_1;

  /* Set the flag to indicate to the calling program that the Interrupt has occured */
  tick_flag_1 = 1;
  
   /*Clear the Interrupt*/
  TICLR |= TACLR;
}

#ifdef __CR__
#pragma interrupt (mftb_isr)
void mftb_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+MFTB_IRQ)
__interrupt void mftb_isr(void)
#else
__interrupt[(16+MFTB_IRQ)*IDT_WIDTH] void mftb_isr(void)
#endif
#endif
{
  /*Reload the counter*/
  TCNT2 = TimerInterval_2;

  /* Set the flag to indicate to the calling program that the Interrupt has occured */
  tick_flag_2 = 1;
  
   /*Clear the Interrupt*/
  TICLR |= TDCLR;
}


