/******************************************************************************
*  File:	mw.c
*  
*  This is Microwire/SPI peripheral driver.
******************************************************************************/

#include "crcommon.h"
#include "chipdef.h"

#define BUF_SIZE 32

mw_t * const mw = (mw_t *)MW_BASE_ADDR;

static boolean master;
static unsigned char *tx_buf, *rx_buf;
static unsigned int tx_count, rx_count;
static volatile boolean tx_in_progress, rx_in_progress; 

static void (*mw_callback_p)(void);

/******************************************************************************  
    Function: mw_init

	Description
	-----------
    This function initializes the MW hardware and the MW driver. After
	returning from this function the MW is ready to transmit and receive
	data.

	Arguments
	---------
	master_mode:
	A boolean value that determines whether the MW should be configured
	to master mode or slave mode.

	divisor: 
	A number that specifies the clock rate used by the MW when in master
	mode. This number actually specifies the ratio between the system main clock
	and the clock that will be used by the MW. The larger the number the slower
	the MW clock. Only even numbers in the range 2-256 are acceptable. Any other
	value will be truncated. This argument is ignored in slave mode.

	callback_p:
	A pointer to a void function with no arguments that will be called when
	a transmit request is completed, receive request is completed or
	transmit/receive request is completed.
*******************************************************************************/
void mw_init(boolean master_mode, unsigned int divisor, void (*callback_p)(void))
{
	/* Save the master/slave flag. */
	master = master_mode;
  
  /* Enable alternate function of the Microwire interface pins. */
	MW_MUX_PORT->pxalt |= MW_MUX_MASK; 	

	/* Initialize the Microwire control register */
	if (master) {
		/* Set the divisor and the master bit*/
	   	if (divisor < 2)
			divisor = 2;
		divisor = (divisor >> 1) - 1;
		MWCTL1 = MNS | (divisor << 9);
	}

	/* Set the Enable Interrupts on Read bit and enable the Microwire. */
	MWCTL1 |= (MWEN | EIR);

	/* Unmask the Microwire interrupt */
	icu_unmask(MW_IRQ);

	/* Save the user callback function. */
	mw_callback_p = callback_p;
	
	tx_in_progress = FALSE;
	rx_in_progress = FALSE; 
}

/******************************************************************************  
    Function: mw_tx

	Description
	-----------
    This function transmits a specified amount (or less) of bytes
	starting from a specified memory location.

	Arguments
	---------
	buf:
	A pointer to a buffer that contains the data to be transmitted.

	size:
	The number of bytes to be transmitted.
	
	Return value
	------------
	None.
*******************************************************************************/
void mw_tx(unsigned char *buf, unsigned int size)
{
	/* Save the buffer pointer and the number of bytes. */
	tx_buf = buf;
	tx_count = size;

	tx_in_progress = TRUE;

	/* Enable interrupts for write (interrupts will be generated when the
	   microwire is ready to transmit the next byte).
	*/
	MWCTL1 |= EIW;

	/* If there is no callback function - wait until transmission is complete. */
	if (mw_callback_p == NULL)
		while(tx_in_progress);
}

/******************************************************************************  
    Function: mw_rx

	Description
	-----------
    This function reads a specified amount (or less) of received bytes
	into a specified memory location.

	Arguments
	---------
	buf:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be read.
	
	Return value
	------------
	None.
*******************************************************************************/
void mw_rx(unsigned char *buf, unsigned int size)
{
	/* Save the buffer pointer and the number of bytes. */
	rx_buf = buf;
	rx_count = size;

	rx_in_progress = TRUE;
	if (master)
		/* A master will have to transmit in order to receive. Therefore 
		   Enable interrupts for write (interrupts will be generated when the
	       microwire is ready to transmit the next byte).
		*/
    	MWCTL1 |= EIW;

	/* If there is no callback function - wait until reception is complete. */
	if (mw_callback_p == NULL)
		while(rx_in_progress);
}

/******************************************************************************  
    Function: mw_tx_rx

	Description
	-----------
    This function transmits a specified amount of bytes and receives the same
    amount of bytes at the same time.

	Arguments
	---------
	txbuf:
	A pointer to a buffer that contains the data to be transmitted.

	rxbuf:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be transmitted and received.
	
	Return value
	------------
	None.
*******************************************************************************/
void mw_tx_rx(unsigned char *rxbuf, unsigned char *txbuf, unsigned int size)
{
	/* Save the buffer pointers and the number of bytes. */
	tx_buf = txbuf;
	rx_buf = rxbuf;
	tx_count = rx_count = size;

	tx_in_progress = TRUE;
	rx_in_progress = TRUE;

	/* Enable interrupts for write (interrupts will be generated when the
	   microwire is ready to transmit the next byte).
	*/
	MWCTL1 |= EIW;

	/* If there is no callback function - wait until both transmission and
	   reception are complete.
	*/
	if (mw_callback_p == NULL)
		while(tx_in_progress || rx_in_progress);
}

/******************************************************************************  
    Function: mw_isr

	Description
	-----------
    This function is the Microwire/SPI interrupt service routine.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef __NSC_CRCC__
#pragma interrupt (mw_isr)
void mw_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+MW_IRQ)
__interrupt void mw_isr(void)
#else
__interrupt[(16+MW_IRQ)*IDT_WIDTH] void mw_isr(void)
#endif
#endif
{
	unsigned char cin, cout;
	boolean write_mwdat = FALSE;

	if (MWSTAT & RBF) {
		/* A byte is waiting in the receive buffer - read it*/
		cin = MWDAT;
		if (rx_in_progress) {
			/* This is a byte we are expecting - move to rx buffer
			   and update counter.
			*/
			*rx_buf++ = cin;
			rx_count--;
		}
		if (tx_in_progress && tx_count == 0) {
			/* If all bytes were transmitted before and tx_in_progress is
			   still true - we were just waiting for the last byte received
			   from the other side as a result of our last transmitted byte. 
			   At this point we can assume that transmission is over.
			*/
			tx_in_progress = FALSE;
		}
	}

	if (!(MWSTAT & BSY)) {
		/* Microwire is ready to transmit a byte */
		if (tx_in_progress) {
			/* If transmission is in progress - copy next byte from tx buffer
			   and update counter. Set a flag to indicate that we need to write
			   to the MWDAT register.
			*/
			cout = *tx_buf++;
			tx_count--;
			write_mwdat = TRUE;
		}
		else
			/* If tx is not in progress a byte will be transmitted anyway
			   at the same time we receive the next byte. Make it 0xff (??).
			*/
			cout = 0xff;

		if (rx_in_progress) {
			if (master) {
				/* If this is a master that receives it will have to write
				   to MWDAT to initiate a byte transaction. Set a flag to
				   indicate that.
				*/
				write_mwdat = TRUE;
        		if (rx_count == 1)
					/* If the master has one more byte to be received this next
					   write to MWDAT will take care of that and there is no
					   need to get more interrupts when the microwire is ready
					   to tramsmit - disable this option.
					*/
          			MWCTL1 &= ~EIW;
      		}
		}

		if (write_mwdat) {
			/* If this flag is set we have to write to MWDAT */
			MWDAT = cout;
		}
	}

	if (tx_in_progress && tx_count == 0)
		/* If all bytes were transmitted we don't need to get an interrupt
		   when the microwire is read to transmit the next byte. Disable this
		   option. But tx_in_progress remains true since we still need to
		   receive one more byte as a result of this byte.
		*/
		MWCTL1 &= ~EIW;

	if (rx_count == 0)
		/* If all expected bytes were received we can assume that reception
		   is over.
		*/
		rx_in_progress = FALSE;

	if(mw_callback_p != NULL && !tx_in_progress && !rx_in_progress)
      (*mw_callback_p)();
}
