/************************************************************************
*	File Name:	sibb.c
*
*	Function:	This drivers implements an SPI-like serial interface 
*				using bit banging on GPIO pins.
*
************************************************************************/

#include <stdlib.h>
#include "config.h"
#include "chipdef.h"
#include "drivers.h"

#define IOBUF_LEN 64


/******************************************************************************  
    Function: sibb_write

	Description
	-----------
    This function writes data over the serial interface.

	Arguments
	---------
	pindesc:
	A descriptor of the 3 interface pins. This descriptor includes the base
	address of GPIO port registers and masks of each one of the 3 interface
	pins: clock, data and enable. All 3 pins are currently assumed to belong
	to the same port.

	data:
	A pointer to a buffer which contains the data bits. The bit sequence
	order is from most significant to least significant, byte after byte.

	size:
	The size of the data in bits.
*******************************************************************************/

void sibb_write(pindesc_t *pindesc, unsigned char *data_p, unsigned int size)
{
	unsigned char *iobuf_p, *io_p;
	unsigned int i;
	unsigned char data_mask;
	unsigned char if_mask;

	/* Allocate memory for the I/O pattern buffer*/
	iobuf_p = malloc(IOBUF_LEN);
	if (iobuf_p == NULL)
		return;

	/* Initialize I/O pointer. */
	io_p = iobuf_p;

	/* Generate the I/O pattern. These are the values that will be written
	   to the port data out register. Start by generating the I/O pattern from
	   the data, one clock for every bit. Every clock is two elements in the
	   pattern (clock low, then clock high).
	*/
	data_mask = 0x80;
		
	for(i = 0; i < size; i++)
	{
		if ((*data_p) & data_mask)
		{
			/* data bit = 1 for one clock*/
			*(io_p++) = pindesc->datamsk;
			*(io_p++) = (pindesc->datamsk | pindesc->clkmsk);
		}
		else
		{
			/* data bit = 0 for one clock*/
			*(io_p++) = 0;
			*(io_p++) = pindesc->clkmsk;
		}

		/* Move to the next data bit. */
		if (data_mask == 0x01)
		{
			data_p++;
			data_mask = 0x80;
		}
		else
			data_mask >>= 1;
	}

	/* Finally generate a pattern where the enable bit is driven high for one
	   clock to latch the data.
	*/
	*(io_p++) = pindesc->enmsk;
	*(io_p++) = (pindesc->enmsk | pindesc->clkmsk);	

	/* Change the size from the data size to the final pattern size. */
	size = 2 * size + 2;

	/* Re-initialize the I/O pointer. */
	io_p = iobuf_p;

	/* Create the pin mask. */
	if_mask = pindesc->clkmsk | pindesc->datamsk | pindesc->enmsk;

	/* Set the pin direction to output. */
	gpio_set_dir(pindesc->port, if_mask, GPIO_OUTPUT);

	/* Now write the I/O pattern to the pins. */
	__disable_interrupt();
	for (i = 0; i < size; i ++)
	{
		gpio_write(pindesc->port, if_mask, *(io_p++));
	}
	__enable_interrupt();

	/* Reset the pin direction. */
	gpio_set_dir(pindesc->port, if_mask, GPIO_INPUT);

	/* Free the I/o buffer. */
	free(iobuf_p);
}
