/******************************************************************************
 *  File:	swreset.c
 *  
 *  This file is the driver  for the software reset 
 ******************************************************************************/

#include "chipdef.h"

/* Define the Operating Modes  as reflected in the MSTAT register*/
/* This should be moved in the cp3bt1/2x.h files */
#define PROG	0x4 
#define TEST	0x5
#define ISP	0x2
#define IRE	0x7
#define ERE	0x3
#define DEV	0x0
#define DEV_INT	0x1	

#if (CHIP_FAMILY == CP3BT2X)
static boolean IsCodeInFlash(uint16 prot);
#endif

/******************************************************************************
 *  Funtions:
 *  
 *		IspIsSwResetValid: 
 *		IspSwReset:
 *
 ******************************************************************************/


/******************************************************************************
 *  Function:	IsSwResetValid	
 *
 *	Description:
 *	-----------
 *		Check the conditions by which a SWRESET will reset in ISP successfully 
 *
 *	Arguments:
 *	---------
 *			None
 *		
 *	Returns:
 *	-------
 *			A boolean indication of the validity of the SwReset into ISP
 *
 *	Pseudo Code:
 *	-----------
 *		Step 1: verify that the device is in IRE or ERE mode
 *		Step 2: verify BOOTAREA is defined (has value other than 0xF) 
 *				in the flash protection word
 *		Step 3: ISPE is set in the flash protection word, 
 *				indicating that there is ISP code in the flash. 
 *		Step 4: Return a boolean result indicating that all three conditions '
 *				are met or not
 *		
 ******************************************************************************/

boolean IsSwResetValid(void) 
{
  boolean result = FALSE;

#if (CHIP_FAMILY == CP3BT2X) 

  /* Step 1: Are we in IRE or ERE mode ? */
  if (((MSTAT & IRE ) == IRE ) || ((MSTAT & ERE) == ERE)) {

    /* Step 2: Do we have a valid boot area ? */
    /* Boot area is 4 bits wide for the cp3bt2x */
    if ( ((FMAR1 & 0x000F ) != 0x000F)) {

      /* Step 3: Does Flash contains ISP code ? Majority vote required */
      if ( IsCodeInFlash(FMAR1) == TRUE) {
        result = TRUE;
      }
      else {
        //ResetStatus = -2;
      }
    }
    else {
      //ResetStatus = -3;
    }
  }
  else {
    //ResetStatus = -1;     
  } 
  
#endif
  return result;
} /* end IspIsSwResetValid */

/******************************************************************************
 *  Function:	SwReset	
 *
 *	Description:
 *	-----------
 *		
 *
 *	Arguments:
 *	---------
 *		
 *	Returns:
 *	-------
 *
 *	Pseudo Code:
 *	-----------
 *		
 ******************************************************************************/


void SwReset(swReset_t func) 
{
#if (CHIP_FAMILY == CP3BT2X)
  SWRESET = 0xE1;
  /* We have to ensure that the two writes to SWRESET 
     are not more than 127 clock cycles apart */
  switch (func) {

  case RESET_ISP:
    SWRESET	= 0x3E;
    break;
		
  case RESET_CLR:
    SWRESET	= 0x0E;
    break;
  }
#endif
} /* end SwReset */

/******************************************************************************
 *  Function:	IsCodeInFlash	
 *
 *	Description:
 *	-----------
 *	Determine if there is Isp Code in the internal flash part
 *
 *	Arguments:
 *	---------
 *	None
 *	
 *       Returns:
 *	-------
 *       A boolean indicating if code was found in flash or not
 *
 *	Pseudo Code:
 *	-----------
 *       TBD
 *		
 ******************************************************************************/
#if (CHIP_FAMILY == CP3BT2X)
static boolean IsCodeInFlash(uint16 prot)
{  
  uint8 curbit;
  uint8 nbBit=0;
  
  /* look at ISPE 3 bits only */
  /* Majortiy of bits set indicate we have ISP */
  /* Application code otherwise */
  prot >>= 7;
  prot &= 0x07;
  // Test to see if majority are '0'
  for (curbit = 0; curbit < 3; curbit++) {
      if (( prot & 0x0001) == 1) nbBit++;
      prot >>= 1;
  }
  if (nbBit <= 1)  return FALSE;
  return TRUE;  
}/* end CodeInFlash */
#endif




