/************************************************************************
	File Name : tcr.c

	Function:	This is a driver for the Triple Clock & Reset module.

************************************************************************/
#include "chipdef.h"

const struct {
	unsigned char prsfc_val;
	unsigned char pll_enable;
} clock_set_table[] = {
	{0x41, 0 },  /*  6 MHz */
	{0x40, 0 },  /* 12 MHz */
	{0x31, 1 },  /* 18 MHz */
	{0x41, 1 },  /* 24 MHz */
	{0x51, 1 },  /* 30 MHz */
	{0x30, 1 },  /* 36 MHz */
	{0x43, 1 },  /* 12 MHz with PLL */
	{0x42, 1 }   /* 16 MHz with PLL */
};

tcr_t * const tcr = (tcr_t *)TCR_BASE_ADDR;

void TCRPllEnable(void);
void TCRPllDisable(void);

/******************************************************************************  
    Function: TCRSetClock

	Description
	-----------
    This function sets the frequency of the system's main clock.

	Arguments
	---------
	freq:
	The required frequency. This is a value of type freq_t that has several
	possible values for different clock frequencies.

*******************************************************************************/
void TCRSetClock(freq_t freq)
{
	/* First of all diable PLL (in case it was enabled */
	TCRPllDisable();

	/* Set clcok divisor and PLL mode (in case PLL is required) */
	PRSFC = clock_set_table[freq].prsfc_val;

	/* If this frequency can only be generated through the PLL - enable PLL */
	if (clock_set_table[freq].pll_enable)
		TCRPllEnable();
}

/******************************************************************************  
    Function: TCRPllEnable

	Description
	-----------
    This function enables the on-chip PLL. It also selects the PLL output
	to be used as input for the high frequency clock generator (should this
	be separated from PLL 

	Arguments
	---------
	None

*******************************************************************************/
void TCRPllEnable(void)
{
	CRCTRL &= ~PLLPWD; /* Power up PLL */
	CRCTRL &= ~FCLK; /* Select PLL output as fast system clock source */
	while (CRCTRL & FCLK); /* Wait until FCLK is clear (PLL stabilized) */
}


/******************************************************************************  
    Function: TCRPllDisable

	Description
	-----------
    This function disables the on-chip PLL.

	Arguments
	---------
	None

*******************************************************************************/
void TCRPllDisable(void)
{
	CRCTRL |= FCLK; /* Select main clock as fast system clock source */
	while (!(CRCTRL & FCLK)); /* Wait until FCLK is set  */	
	CRCTRL |= PLLPWD; /* Power down PLL */
}
