/******************************************************************************
*  File:	timer.c
*  
*  This file provides system timer service. It currently uses the TWM
*  timer but it could work with any of the timing hardware blocks that are
*  available in the system.
*  In the heart of this driver is a wakeup call mechanism. A caller can
*  set up a wakeup call which allows a given function to be called after
*  a specified number of clock ticks either on a one time basis or on a
*  repetitive basis.
******************************************************************************/

#include "crcommon.h"
#include "drivers.h"
#ifdef UCOS
#  include "ucos.h"
#endif

#define WAKEUPLIST_SIZE 8

#ifndef UCOS
static void set_timeout(void);
static volatile boolean timeout;
#endif


/* The following structure defines an entry in the wakeup list */ 

typedef struct {
	boolean occupied;         // Indicates if the entry is occupied/valid.
	void (* func)(void);      // Pointer to function to be called.
	unsigned int clock_ticks; // Number of clock ticks before time expires.
	unsigned int counter;     // Counts umber of clock ticks elapsed so far.
	boolean repeat;           // Indicates of repetitive or one time request.
}wakeup_t;

static wakeup_t wakeuplist[WAKEUPLIST_SIZE];

/******************************************************************************  
    Function: timer_init

	Description
	-----------
    This function initializes the TWM hardware and the timer driver,
    specifically the wakeup list.

	Arguments
	---------
	None
*******************************************************************************/
void timer_init(void)
{
	int i;

	/* Initialize the TWM */
#ifdef UCOS
	twm_init(10, OSTimeTick);
#else
	twm_init(10, timer_wakeup_check);
#endif

	/* Initialize the wakeup list. */
	for (i = 0; i < WAKEUPLIST_SIZE; i++)
		wakeuplist[i].occupied =  FALSE;
}

/******************************************************************************  
    Function: timer_set_wakeup

	Description
	-----------
    This function sets up a wakeup call. It does this by inserting an entry to
    the wakeup list.

	Arguments
	---------
	func:
	A pointer to a function that will be called when the time expires.

	clock_ticks:
	The number of clocks ticks to elapse before the time expires.

	repeat:
	A flag that specifies where this is a reptitive wakeup call or a one time
	wakeup call. If FALSE there will be a one time wakeup call and then the
	wakeup entry will be invalidated. If TRUE the time count will restart
	automatically when the time expires.

	Returns
	-------
	A handle to be used for this wakeup call for further references (e.g.,
	if the wakeup call is to be cancelled later). In practice this is the
	index of the wakeup call in the wakeup list.
*******************************************************************************/

int timer_set_wakeup(void (* func)(void), unsigned int clock_ticks, boolean repeat)
{
	int i;

#ifdef UCOS
	OS_ENTER_CRITICAL();
#endif
	/* search for an available entry */ 
	for (i = 0; i < WAKEUPLIST_SIZE; i++) {
		if (wakeuplist[i].occupied ==  FALSE) break;
	}

	if (i != WAKEUPLIST_SIZE)
		wakeuplist[i].occupied = TRUE;

#ifdef UCOS
	OS_EXIT_CRITICAL();
#endif

	if (i == WAKEUPLIST_SIZE)
		return -1; /* No entry is available */

	/* Fill the entry with wakeup request data. */
	wakeuplist[i].func = func;
	wakeuplist[i].clock_ticks = clock_ticks;
	wakeuplist[i].counter = 0;
	wakeuplist[i].repeat = repeat;

	return i;
}

/******************************************************************************  
    Function: timer_clear_wakeup

	Description
	-----------
    This function deactivate a given wakeup call.

	Arguments
	---------
	The wakeup call handle (index).
*******************************************************************************/

void timer_clear_wakeup(int index)
{
	/* All we do is invalidate the wakeup entry. This entry will not be occupied
	   anymore.
	*/
	wakeuplist[index].occupied = FALSE;
}

/******************************************************************************  
    Function: timer_wakeup_check

	Description
	-----------
    This function is the heart if the timer wakeup mechanism. When called
	it scans the wakeup list, increments time counts and if for a specific 
	wakeup entry the time has expired it calls the wakeup function. This
	function should be called every time tick.

	Arguments
	---------
	None.
*******************************************************************************/

void timer_wakeup_check(void)
{
	int i;

	/* Scan the wakeup list */ 
	for (i = 0; i < WAKEUPLIST_SIZE; i++) {
		/* If not an occupied (valid) entry - skip this entry. */
		if (wakeuplist[i].occupied == FALSE) continue;
		/* Increment time count for this entry. */
		wakeuplist[i].counter++;
		/* Check if time has expired. */
		if (wakeuplist[i].counter == wakeuplist[i].clock_ticks) {
			/* If time has expired call wakeup function. */
			(*wakeuplist[i].func)();
			if (wakeuplist[i].repeat)
				/* If this is a repetitive request reset counter. */
				wakeuplist[i].counter = 0;
			else
				/* Otherwise - clear (invalidate) the wakeup entry because
				   this was a one time request.
				*/
				wakeuplist[i].occupied = FALSE;
		}
	}

}

/******************************************************************************  
    Function: timer_wait

	Description
	-----------
    This function provides a wait (delay) service. It delays the execution 
	of the program for a give number of ticks.
	Note: The implementation is different between the OS and OS-less versions.
	In the OS version we simply delay the current task for the required number
	of clock ticks (but other tasks may run in the mean time).
	In the OS-less version there is a busy wait	which is implemented by polling
	a flag that will become TRUE after the required number of clock ticks.

	Arguments
	---------
	ticks:
	The number of clock ticks to wait/delay.

	Returns
	-------
	A boolean flag indicating if the wait was completed successfully.
*******************************************************************************/
boolean timer_wait(unsigned int ticks)
{
#ifdef UCOS
	OSTimeDly(ticks);
#else
	int idx;

    /* Initialize timeout flag. */
	timeout = FALSE;
	/* Set a wakeup call with the required number of ticks. */
	idx = timer_set_wakeup(set_timeout, ticks, FALSE);
	if (idx < 0)
		return FALSE;
	/* Wait for timeout flag to become TRUE. */
	while(!timeout);
#endif
	return TRUE;
}

/******************************************************************************  
    Function: set_timeout

	Description
	-----------
    This function is the wakeup function for the wait service. It should be
	called when the wait/delay time expires. All it does is set the timeout
	flag to TRUE.

	Arguments
	---------
	ticks:
	The number of clock ticks to wait/delay.
******************************************************************************/
#ifndef UCOS
static void set_timeout(void)
{
	timeout = TRUE;
}
#endif
