/******************************************************************************
*  File:	twm.c
*  
*  This Timing and Watchdog Module (TWM) driver.
******************************************************************************/

#include "crcommon.h"
#include "chipdef.h"
#ifdef UCOS
#include "ucos.h"
#endif

twm_t * const twm = (twm_t *)TWM_BASE_ADDR;

static void (* tick_cb_p)();

/******************************************************************************  
    Function: twm_init

	Description
	-----------
    This function initializes the TWM peripheral and the TWM driver.

	Arguments
	---------
	tick_period:
	The length of the clock tick generated by the timing module in milliseconds.

	tick_callback_p:
	A pointer to a function that will be called on every clock tick.
*******************************************************************************/
void twm_init(unsigned int tick_period, void (*tick_callback_p)(void))
{
  
	/* Save the callback function pointer */
  	tick_cb_p = tick_callback_p;

  	/* select 32kHz oscillator as the source of slow clock */
  	//CRCTRL |= SCLK;

  	/* set pre-scalar to 1 */
  	TWCP = 0;
  
  	/* Set TWMT0 down counter according to the requited tick period */
  	TWMT0 = (unsigned short)(((unsigned long)tick_period * 32768) / 1000);

  	/* Restart the timer and enable interrupts */
  	T0CSR |= (T0INTE | RST);

  	/* unmask the TWM interrupt */
  	icu_unmask(TWM_IRQ);
}

/******************************************************************************  
    Function: wd_init

	Description
	-----------
    This function initializes Watchdog.

	Arguments
	---------
	ticks:
	This is the number of clock ticks the Watchdg counter will be loaded with.
	The Watchdog will count down and unless wd_restart is called early enough,
	it will reach zero and reset the CPU.
*******************************************************************************/
void wd_init(unsigned char ticks)
{
	/* Set the number of ticks */
	WDCNT = ticks;
	/* Lock the WDCNT register (to avoid firther writes) and enable Watchdog
	   restart through WDSDM register (see the wd_resrat function)
	*/
	TWCFG |= (LWDCNT | WDSDME); 
}

/******************************************************************************  
    Function: wd_restart

	Description
	-----------
    This function restarts the Watchdog. This causes the Watchdog counter to
	be reloaded with the original value that was assigned to it and start
	the countdown again. If this is not done early enough the the Watchdog
	Error signal will be trigered and that will reset the CPU.

	Arguments
	---------
	None.
*******************************************************************************/
void wd_restart()
{
	/* restart the Watchdog by writing the magic number of 0x5c to WDSDM. */
	WDSDM = 0x5c;
}

/******************************************************************************  
    Function: twm_isr

	Description
	-----------
    This is the TWM interrupt service routine.

	Arguments
	---------
	None.
*******************************************************************************/
#ifdef __NSC_CRCC__
#pragma interrupt (twm_isr)
void twm_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+TWM_IRQ)
__interrupt void twm_isr(void)
#else
__interrupt[(16+TWM_IRQ)*IDT_WIDTH] void twm_isr(void)
#endif
#endif
{

#ifdef UCOS
	SAVE_CONTEXT();
	//OSIntEnter();
#endif

	/* Clear TWM interrupt, yes this actually reads the register! */
  	T0CSR;

	/* Call the tick callback function */
	if (tick_cb_p != NULL)
  		(*tick_cb_p)();

#ifdef UCOS
	OSIntExit();
	RESTORE_CONTEXT();
#endif
}

