/************************************************************************
	File Name : usart.c

	Function:	This is a driver for the USART peripheral.

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "crcommon.h"
#include "chipdef.h"
#include "dl.h"
#ifdef UCOS
#include "ucos.h"
#endif

#define TX_BUF_SIZE 256
#define RX_BUF_SIZE 256

/* USART base address table */
static usart_t * const usart_tab[] = {
	USART_BASE_ADDRESSES
};

/* Table of base addresses of ports muxed with each individual USART. */
static gpio_t * const usart_mux_port_tab[] = {
	USART_MUX_PORTS
};

/* Table of bit masks of pins of ports that are masked with the pins of each
   individual USART.
*/
static unsigned char const usart_mux_mask_tab[] = {
	USART_MUX_MASK
};

/* Table of transmit interrupt IRQs for each of the individual USARTs */
static unsigned char const usart_tx_irqs[] = {
	USARTTX_IRQS
};

/* Table of receive interrupt IRQs for each of the individual USARTs */
static unsigned char const usart_rx_irqs[] = {
	USARTRX_IRQS
};

/* This table contains initialization value for 3 of the USART registers
   according to target bit rate. This table assumes input clock of 12MHz.
   It should be later upgraded to support higher frequencies as well.
 */ 
const unsigned char bitrate_reg_table[][3] = 
{
  {0x0a, 0x08, 0x7c},	/*   9600 BPS */
  {0x00, 0x10, 0x0c},	/*  38400 BPS */
  {0x0d, 0x08, 0x07},	/* 115200 BPS */
  {0x0d, 0x08, 0x03},	/* 230400 BPS */
  {0x0d, 0x08, 0x01},	/* 460800 BPS */
  {0x0d, 0x08, 0x00}	/* 921600 BPS */
};

static unsigned char *tx_ptr[N_USARTS], *rx_ptr[N_USARTS],
					*client_tx_ptr[N_USARTS], *client_rx_ptr[N_USARTS];
static dl_t tx_dl_s[N_USARTS], rx_dl_s[N_USARTS];
static dl_t *rx_dl[N_USARTS], *tx_dl[N_USARTS];
static flow_control_t fc[N_USARTS];

#ifdef UCOS
static OS_EVENT *txsem[N_USARTS];
#else
static boolean tx_busy[N_USARTS];
#endif

static void usart_tx_isr(unsigned int);
static void usart_rx_isr(unsigned int);

/******************************************************************************  
    Function: usart_init

	Description
	-----------
    This function initializes the USART hardware and the USART driver. After
	returning from this function the USART is ready to transmit and receive
	data.

	Arguments
	---------
	usart_num:
	The number of the USART to be initialized.

	bitrate:
	The required communication bit rate. Check the bitrate_t type for supported
	bit rates.

 	flow_control:
	Indicates what type of flow control is to be used (if any). Check the
	flow_control_t type for flow control types. Note: software flow control is
	not yet supported.

	rx_callback: 
	An optional pointer. This function will automatically be called when data is
	received unless the pointer is NULL.This function should be as short as
	possible as it is called from the USART receive interrupt handler.
*******************************************************************************/

void usart_init(unsigned int usart_num, bitrate_t bitrate, flow_control_t flow_control,
                void (*rx_callback_p)(void))
{
  usart_t * usart;
  unsigned char *tx_buf, *rx_buf;

  /* Set the pointer to the base address of the slected USART. */
  usart = usart_tab[usart_num];

  /* Allocate buffer memory */
  tx_buf = malloc(TX_BUF_SIZE);
  rx_buf = malloc(RX_BUF_SIZE);
  if (tx_buf == NULL || rx_buf == NULL)
	/* Temporary solution - abort. */
	panic();

  /* Initialize data links (circular buffers) and data link pointers */
  tx_dl[usart_num] = &tx_dl_s[usart_num];
  rx_dl[usart_num] = &rx_dl_s[usart_num];       
  dl_init(tx_dl[usart_num], tx_buf, TX_BUF_SIZE);
  dl_init(rx_dl[usart_num], rx_buf, RX_BUF_SIZE);
  dl_init_ptr(rx_dl[usart_num], rx_ptr[usart_num]);
  dl_init_ptr(rx_dl[usart_num], client_rx_ptr[usart_num]);
  dl_init_ptr(tx_dl[usart_num], tx_ptr[usart_num]);
  dl_init_ptr(tx_dl[usart_num], client_tx_ptr[usart_num]);

  //dl_p_set_action(tx_dl, usart_unmask_tx);
  dl_p_set_action(rx_dl[usart_num], rx_callback_p);

#ifdef UCOS
  /* Create a tx semaphore */
  txsem[usart_num] = OSSemCreate(1);
#endif


  /* Save flow control method. */
  fc[usart_num] = flow_control;

  /* If h/w flow control required - turn on h/w flow control. */
  if (fc[usart_num] == HW_FLOW_CONTROL) {
	UMDSL1 |= UFCE;
	UMDSL1 |= URTS;
  }

  /* Enable alternate function those general purpose I/O pins that are
     multiplexed with the USART pins.
  */
  (usart_mux_port_tab[usart_num])->pxalt |= usart_mux_mask_tab[usart_num];



  /* Set the bit rate registers according to the required bit rate */	
  UOVR = bitrate_reg_table[bitrate][0];
  UPSR = bitrate_reg_table[bitrate][1];	       
  UBAUD = bitrate_reg_table[bitrate][2];

  UICTRL = (UETI|UERI);	//enable Transmit and Receive interrupts.

  /* Unmask USART RX interrupt. The TX interrupt will be unmasked on
     a need-basis i.e. when data has to be transmitted */
  icu_unmask(usart_rx_irqs[usart_num]);
 
}

/******************************************************************************  
    Function: usart_shutdown

	Description
	-----------
    This function shuts down the USART and releases resources.

	Arguments
	---------
	usart_num:
	The number of the USART to be shut down.
*******************************************************************************/
void usart_shutdown(unsigned int usart_num)
{
#ifdef UCOS
	unsigned char err;
#endif

	/* Mask interrupt. */
	icu_mask(usart_rx_irqs[usart_num]);
	icu_mask(usart_tx_irqs[usart_num]);

#ifdef UCOS
	/* Delete semaphore. */
	OSSemDel(txsem[usart_num], OS_DEL_ALWAYS, &err);
#endif

	/* Free data link buffers. */
	free(dl_get_buf(tx_dl[usart_num]));
	free(dl_get_buf(rx_dl[usart_num]));
}

/******************************************************************************  
    Function: usart_tx

	Description
	-----------
    This function transmits a specified amount (or less) of bytes
	starting from a specified memory location.

	Arguments
	---------
	usart_num:
	The number of the USART to perform this operation on.

	data:
	A pointer to the start address of the data to be transmitted.

	size:
	The number of bytes to be transmitted.
	
	Return value
	------------
	The number of bytes that were actually transmitted. Could be less than
	the requested number.
*******************************************************************************/
unsigned int usart_tx(unsigned int usart_num, unsigned char *data, unsigned int size)
{
  unsigned int i, avail;

#ifdef UCOS
  unsigned char errp;

  OSSemPend (txsem[usart_num], 0, &errp);
#else
  if (tx_busy[usart_num])
	return 0;
  tx_busy[usart_num] = TRUE;
#endif

  do {
  	avail = dl_p_space_available(tx_dl[usart_num]);
  	if (size > avail)
#ifdef UCOS
		OSTimeDly(1);
#else
		size = avail;
#endif
  } while (size > avail);

  for (i = 0; i < size; i++) {
	*client_tx_ptr[usart_num] = *data;
	dl_advance(tx_dl[usart_num], client_tx_ptr[usart_num]);
	data++;
  }

  dl_p_commit(tx_dl[usart_num], size);
  icu_unmask(usart_tx_irqs[usart_num]);

#ifdef UCOS
  OSSemPost (txsem[usart_num]);
#else
  tx_busy[usart_num] = FALSE;
#endif

  return size;
}

/******************************************************************************  
    Function: usart_rx

	Description
	-----------
    This function reads a specified amount (or less) of received bytes
	into a specified memory location.

	Arguments
	---------
	usart_num:
	The number of the USART to perform this operation on.

	data:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be read.
	
	Return value
	------------
	The number of bytes that were actually read. Could be less than
	the requested number.
*******************************************************************************/
unsigned int usart_rx(unsigned int usart_num, unsigned char *outbuf, unsigned int size)
{
  usart_t * usart;
  unsigned int i, avail;


  /* Set the pointer to the base address of the slected USART. */
  usart = usart_tab[usart_num];

  avail = dl_cnt(rx_dl[usart_num]);
  if (size > avail)
	size = avail;

  for (i = 0; i < size; i++) {
	*outbuf = *client_rx_ptr[usart_num];
	dl_advance(rx_dl[usart_num], client_rx_ptr[usart_num]);
	outbuf++;
  }

  dl_c_commit(rx_dl[usart_num], size);

  
  if (fc[usart_num] == HW_FLOW_CONTROL && (UMDSL1 & URTS) == 0 &&
  		dl_p_space_available(rx_dl[usart_num]) > 16)
	UMDSL1 |= URTS;

  return size;
}

/******************************************************************************  
    Function: usart_rx_count

	Description
	-----------
    This function returns the number of bytes that can be currently read from
	the USART driver

	Arguments
	---------
	usart_num:
	The number of the USART to perform this operation on.

	
	Return value
	------------
	The number of bytes that can be currently read from the USART driver.
*******************************************************************************/
unsigned int usart_rx_count(unsigned int usart_num)
{
  return(dl_cnt(rx_dl[usart_num]));	
}


/******************************************************************************  
    Function: usart_tx_isr

	Description
	-----------
    This function is the core of the USART transmit interrupt handler. It reads
    a byte from the transmit data link and copies it to the USART transmit buffer.
    The function is called from the transmit interrupt service routine of each of
    the individual USARTs.

	Arguments
	---------
	usart_num:
	The number of the USART to perform this operation on.
	
	Return value
	------------
	None.
*******************************************************************************/
static void usart_tx_isr(unsigned int usart_num)
{
  usart_t * usart;

  /* Set the pointer to the base address of the slected USART. */
  usart = usart_tab[usart_num];

  UTBUF = *tx_ptr[usart_num];
  dl_c_advance_commit(tx_dl[usart_num], tx_ptr[usart_num], 1);
  if (dl_empty(tx_dl[usart_num]))
    icu_mask(usart_tx_irqs[usart_num]);
}


/******************************************************************************  
    Function: usart_rx_isr

	Description
	-----------
    This function is the core of the USART receive interrupt handler. It reads
    a byte from the USART receive buffer and copies it to the receive data link.
    The function is called from the receive interrupt service routine of each
    of the individual USARTs.

	Arguments
	---------
	usart_num:
	The number of the USART to perform this operation on.
	
	Return value
	------------
	None.
*******************************************************************************/
static void usart_rx_isr(unsigned int usart_num)
{
  usart_t * usart;

  /* Set the pointer to the base address of the slected USART. */
  usart = usart_tab[usart_num];
	
  /* Read character from receive buffer into data link buffer */
  *rx_ptr[usart_num] = URBUF;
 
  /* sync on data link */	
  dl_p_advance_commit(rx_dl[usart_num], rx_ptr[usart_num], 1);

  /* If we are running out of space in the rx buffer de-assert RTS. */
  if (dl_p_space_available(rx_dl[usart_num]) <= 16 && fc[usart_num] == HW_FLOW_CONTROL)
	UMDSL1 &= ~URTS;
}

/******************************************************************************  
    Function: usart0_tx_isr

	Description
	-----------
    This function is the USART0 transmit interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART0TX_IRQ 
#ifdef __NSC_CRCC__
#pragma interrupt(usart0_tx_isr)
void usart0_tx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART0TX_IRQ)
__interrupt void usart0_tx_isr(void)
#endif
{
	usart_tx_isr(0);
}
#endif

/******************************************************************************  
    Function: usart1_tx_isr

	Description
	-----------
    This function is the USART1 transmit interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART1TX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart1_tx_isr)
void usart1_tx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART1TX_IRQ)
__interrupt void usart1_tx_isr(void)
#endif
{
	usart_tx_isr(1);
}
#endif

/******************************************************************************  
    Function: usart2_tx_isr

	Description
	-----------
    This function is the USART2 transmit interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART2TX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart2_tx_isr)
void usart2_tx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART2TX_IRQ)
__interrupt void usart2_tx_isr(void)
#endif
{
	usart_tx_isr(2);
}
#endif

/******************************************************************************  
    Function: usart3_tx_isr

	Description
	-----------
    This function is the USART3 transmit interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART3TX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart3_tx_isr)
void usart3_tx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART3TX_IRQ)
__interrupt void usart3_tx_isr(void)
#endif
{
	usart_tx_isr(3);
}
#endif

/******************************************************************************  
    Function: usart0_rx_isr

	Description
	-----------
    This function is the USART0 receive interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART0RX_IRQ 
#ifdef __NSC_CRCC__
#pragma interrupt(usart0_rx_isr)
void usart0_rx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART0RX_IRQ)
__interrupt void usart0_rx_isr(void)
#endif
{
	usart_rx_isr(0);
}
#endif

/******************************************************************************  
    Function: usart1_rx_isr

	Description
	-----------
    This function is the USART1 receive interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART1RX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart1_rx_isr)
void usart1_rx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART1RX_IRQ)
__interrupt void usart1_rx_isr(void)
#endif
{
	usart_rx_isr(1);
}
#endif

/******************************************************************************  
    Function: usart2_rx_isr

	Description
	-----------
    This function is the USART2 receive interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART2RX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart2_rx_isr)
void usart2_rx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART2RX_IRQ)
__interrupt void usart2_rx_isr(void)
#endif
{
	usart_rx_isr(2);
}
#endif

/******************************************************************************  
    Function: usart3_rx_isr

	Description
	-----------
    This function is the USART3 receive interrupt handler.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifdef USART3RX_IRQ
#ifdef __NSC_CRCC__
#pragma interrupt(usart3_rx_isr)
void usart3_rx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#pragma vector=(16+USART3RX_IRQ)
__interrupt void usart3_rx_isr(void)
#endif
{
	usart_rx_isr(3);
}
#endif

