/************************************************************************
	File Name : usart_dma.c

	This version of the USART driver uses DMA transfers for transmit and
	receive.

************************************************************************/

#include <stdio.h>
#include "crcommon.h"
#include "chipdef.h"
#include "dl.h"
#ifdef UCOS
#include "ucos.h"
#endif

#define TX_BUF_SIZE 256
#define RX_BUF_SIZE 256 // Must be a power of 2.

static unsigned char tx_buf[TX_BUF_SIZE];
static unsigned char rx_buf[RX_BUF_SIZE];

usart_t * const usart = (usart_t *)USART0_BASE_ADDR;
dma_t * const dma = (dma_t *)DMA_BASE_ADDR;

/* Table of base addresses of ports muxed with each individual USART. */
static gpio_t * const usart_mux_port_tab[] = {
	USART_MUX_PORTS
};

/* Table of bit masks of pins of ports that are masked with the pins of each
   individual USART.
*/
static unsigned char const usart_mux_mask_tab[] = {
	USART_MUX_MASK
};

static void usart_dma_ch1_enable(void);
static void usart_dma_timeout(void);


/* This table contains initialization value for 3 of the USART registers
   according to target bit rate. This table assumes input clock of 12MHz.
   It should be later upgraded to support higher frequencies as well.
 */ 
const unsigned char bitrate_reg_table[][3] = 
{
  {0x0a, 0x08, 0x7c},	/*   9600 BPS */
  {0x00, 0x10, 0x0c},	/*  38400 BPS */
  {0x0d, 0x08, 0x07},	/* 115200 BPS */
  {0x0d, 0x08, 0x03},	/* 230400 BPS */
  {0x0d, 0x08, 0x01},	/* 460800 BPS */
  {0x0d, 0x08, 0x00}	/* 921600 BPS */
};

static unsigned char *tx_ptr, *rx_ptr,*client_tx_ptr, *client_rx_ptr;
static dl_t tx_dl_s, rx_dl_s;
static dl_t *rx_dl, *tx_dl;
static flow_control_t fc;

static volatile unsigned int rx_clock;
static unsigned int rx_bytes_left;
static unsigned int actual_rx_xfer_size, tx_xfer_size, timeout;

#ifdef UCOS
static OS_EVENT *txsem;
#else
static boolean tx_busy = FALSE;
#endif

/******************************************************************************  
    Function: usart_dma_init

	Description
	-----------
    This function initializes the USART according to the required bit rate.
	It initializes DMA channels 0 for USART rx and channel 1 for USART tx.
	Note: rx_xfersz (the rx transfer size) must be an integral part of the rx
	data link buffer size (the buffer size should be divisible by rx_zfersz
	with no remainder).

    **Note**
	The function timer_init MUST be called prior to calling usart_dma_init.

	Arguments
	---------
	bitrate:
	The required communication bit rate. Check the bitrate_t type for supported
	bit rates.

 	flow_control:
	Indicates what type of flow control is to be used. Check the flow_control_t
	type for flow control types. Note: software flow control is not yet
	supported.

	rx_callback: 
	An optional pointer. This function will automatically be called when data is
	received unless the pointer is NULL. This function should be as short as
	possible as it is called from an interrupt handler.
*******************************************************************************/

void usart_dma_init(bitrate_t bitrate, flow_control_t flow_control, 
                    void (*rx_callback_p)(void))
{
  /* Initialize data links (circular buffers) and data link pointers */
  tx_dl = &tx_dl_s;
  rx_dl = &rx_dl_s;       
  dl_init(tx_dl, tx_buf, TX_BUF_SIZE);
  dl_init(rx_dl, rx_buf, RX_BUF_SIZE);
  dl_init_ptr(rx_dl, rx_ptr);
  dl_init_ptr(rx_dl, client_rx_ptr);
  dl_init_ptr(tx_dl, tx_ptr);
  dl_init_ptr(tx_dl, client_tx_ptr);
  dl_p_set_action(tx_dl, usart_dma_ch1_enable);
  dl_p_set_action(rx_dl, rx_callback_p);

#ifdef UCOS
  /* Create a tx semaphore */
  txsem = OSSemCreate(1);
#endif

  actual_rx_xfer_size = 32;

  /* Initialize timeout. Make it a <= 10ms timeout for now. */
  timeout = 1; 

  /* Save flow control method. */
  fc = flow_control;

  /* If h/w flow control required - turn on h/w flow control. */
  if (fc == HW_FLOW_CONTROL)
	UMDSL1 |= (UFCE | URTS);


/* Enable alternate function those general purpose I/O pins that are
     multiplexed with the USART pins.
  */
  (usart_mux_port_tab[0])->pxalt |= usart_mux_mask_tab[0];

  /* Set the bit rate registers according to the required bit rate */	
  UOVR = bitrate_reg_table[bitrate][0];
  UPSR = bitrate_reg_table[bitrate][1];	       
  UBAUD = bitrate_reg_table[bitrate][2];

  /* Configure USART to use trigger DMA requests for tx and rx */
  UMDSL1 |= (UETD |  UERD);

  /* Enable Rx error interrupt - required for debugging */
  //UICTRL |= UEEI;

  /* Initialize DMA channel 0 registers for rx operations */
  rx_bytes_left = actual_rx_xfer_size;
  ADCA0 = (void FAR *)(&URBUF); //DMA device A address: USART rx buffer address
  ADCB0 = (void FAR *)rx_ptr;   //DMA device B address: data link buffer pointer
  BLTC0 = actual_rx_xfer_size;            //DMA counter 
  DMACNTL0 = ETC | IND | ADB; 		//Interrupt on terminal count, indirect
  									//xfer, auto-increment on device B address,
									//direction: device A -> device B.

  DMASTAT0 |= VLD;					//Xfer parameters valid.
  DMACNTL0 |= CHEN;					//Channel enable.

  /* InitializeDMA registers for double buffer operation */
  dl_advance_multiple(rx_dl, rx_ptr, actual_rx_xfer_size); 
  ADRA0 = (void FAR *)(&URBUF); //DMA device A address: USART rx buffer addres
  ADRB0 = (void FAR *)rx_ptr;   //DMA device B address: data link buffer pointer
  BLTR0 = actual_rx_xfer_size;

  /* Initialize rx clock and Set up a timeout function. */
  rx_clock = 0;

  timer_set_wakeup(usart_dma_timeout, timeout, TRUE);

  /* Initialize DMA channel 1 registers for tx operations.
     Note: channel 1 will not be enabled until there is
	 data in the data link buffer. Channel enable will be triggered
	 by the data link producer action.
   */
  ADCA1 = (void FAR *)tx_ptr;    //DMA device A address: data link buffer pointer
  ADCB1 = (void FAR *)(&UTBUF);  //DMA device B address: USART tx buffer address
  DMACNTL1 = ETC | IND | ADA;       //Interrupt on terminal count, indirect xfer, auto-increment on device A address
  DMASTAT1 |= VLD;                  //Xfer parameters valid.  

  /* Unmask DMA channel 0 and 1 interrupts */
  icu_unmask(DMA_CHANNEL0_IRQ);
  icu_unmask(DMA_CHANNEL1_IRQ);
 
}

/******************************************************************************  
    Function: usart_dma_tx

	Description
	-----------
    This function transmits a specified amount (or less) of bytes
	starting from a specified memory location.

	Arguments
	---------
	data:
	A pointer to the start address of the data to be transmitted.

	size:
	The number of bytes to be transmitted.
	
	Return value
	------------
	The number of bytes that were actually transmitted. Could be less than
	the requested number.
*******************************************************************************/
unsigned int usart_dma_tx(unsigned char *data, unsigned int size)
{				  
  unsigned int i, avail;

#ifdef UCOS
  unsigned char errp;

  OSSemPend(txsem, 0, &errp);
#else
  if (tx_busy)
	return 0;
  tx_busy = TRUE;
#endif

  do {
  	avail = dl_p_space_available(tx_dl);
  	if (size > avail)
#ifdef UCOS
		OSTimeDly(1);
#else
		size = avail;
#endif
  } while (size > avail);

  for (i = 0; i < size; i++) {
	*client_tx_ptr = *data;
	dl_advance(tx_dl, client_tx_ptr);
	data++;
  }

  dl_p_commit(tx_dl, size);

#ifdef UCOS
  OSSemPost(txsem);
#else
  tx_busy = FALSE;
#endif

  return size;
}

/******************************************************************************  
    Function: usart_dma_rx

	Description
	-----------
    This function reads a specified amount (or less) of received bytes
	into a specified memory location.

	Arguments
	---------
	data:
	A pointer to the start address of the buffer to receive the data.

	size:
	The number of bytes to be read.
	
	Return value
	------------
	The number of bytes that were actually read. Could be less than
	the requested number.
*******************************************************************************/
unsigned int usart_dma_rx(unsigned char *outbuf, unsigned int size)
{
  unsigned int i, avail;

  avail = dl_cnt(rx_dl);
  if (size > avail)
	size = avail;

  for (i = 0; i < size; i++) {
	*outbuf = *client_rx_ptr;
	dl_advance(rx_dl, client_rx_ptr);
	outbuf++;
  }

  dl_c_commit(rx_dl, size);

  
  /* If RTS has been de-asserted and rx buffer now has enough space for
     the next transfer - re-assert RTS.
  */
  if (fc == HW_FLOW_CONTROL && (UMDSL1 & URTS) == 0 && 
      dl_p_space_available(rx_dl) > 2* actual_rx_xfer_size)
	UMDSL1 |= URTS;

  return size;
}

/******************************************************************************  
    Function: usart_dma_rx_count

	Description
	-----------
    This function returns the number of bytes that can be currently read from
	the USART-DMA driver

	Arguments
	---------
	None.
	
	Return value
	------------
	The number of bytes that can be currently read from the USART-DMA driver.
*******************************************************************************/
unsigned int usart_dma_rx_count(void)
{
  return(dl_cnt(rx_dl));	
}

/******************************************************************************  
    Function: dma_channel0_isr

	Description
	-----------
    This function is DMA channel 0 interrupt handler. The interrupt is triggered
    when the RX DMA transfer is complete. The handler releases (commits) on the
    bytes in the data link that were not released so far (thereby making them
    available to the consumer) . Them it re-programs the DMA for the next RX
    transfer.

	Arguments
	---------
	None.
*******************************************************************************/

#ifdef __NSC_CRCC__
#pragma interrupt(dma_channel0_isr)
void dma_channel0_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+DMA_CHANNEL0_IRQ)
__interrupt void dma_channel0_isr(void)
#else
__interrupt[(16+DMA_CHANNEL0_IRQ)*IDT_WIDTH] void dma_channel0_isr(void)
#endif
#endif
{	
   
  /* Advance data link pointer and commit data to consumer */	
  dl_p_commit(rx_dl, rx_bytes_left);

  /* If we are running out of space in the rx buffer de-assert RTS. */
  if (dl_p_space_available(rx_dl) <= 2*actual_rx_xfer_size && fc == HW_FLOW_CONTROL)
	UMDSL1 &= ~URTS;

  /* Re-program DMA channel 0 registers for the next transfer. */
  rx_bytes_left = actual_rx_xfer_size;
  dl_advance_multiple(rx_dl, rx_ptr, actual_rx_xfer_size);
  DMASTAT0 |= TC;
  ADRA0 = (void FAR *)(&URBUF); //DMA device A address: USART rx buffer addr
  ADRB0 = (void FAR *)rx_ptr;
  BLTR0 = actual_rx_xfer_size;

  rx_clock = 0;
}

/******************************************************************************  
    Function: dma_channel1_isr

	Description
	-----------
    This function is DMA channel 1 interrupt handler. The interrupt is triggered
    when the TX DMA transfer is complete. The handler commits on the bytes being
    consumed (thereby freeing space in the data link buffer).Them it re-programs
    the DMA for the next TX transfer, but re-enables it only if there are pending
    bytes for transmission.

	Arguments
	---------
	None.
*******************************************************************************/

#ifdef __NSC_CRCC__
#pragma interrupt(dma_channel1_isr)
void dma_channel1_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+DMA_CHANNEL1_IRQ)
__interrupt void dma_channel1_isr(void)
#else
__interrupt[(16+DMA_CHANNEL1_IRQ)*IDT_WIDTH] void dma_channel1_isr(void)
#endif
#endif
{
	unsigned int contig;

  	/* Advance data link pointer and commit on data consumption. */ 
  	dl_c_advance_commit(tx_dl, tx_ptr, tx_xfer_size);

  	/* Re-program DMA channel 1 registers for the next transfer.
       Note: Channel 1 will be re-enabled only if there is data 
	   in the data link buffer. The tx transfer size depends on
	   the number of contiguous bytes available in the data link
	   buffer since the DMA controller cannot wrap around when it
	   reaches the end of the buffer.
   	*/
  	DMACNTL1 &= ~CHEN;
  	ADCA1 = (void FAR *)tx_ptr;
  	DMASTAT1 |= VLD;
  	if ((contig = dl_get_max_linear_data_size(tx_dl, tx_ptr)) > 0) {
		/* Set DMA counter according to number of contiguous data bytes
		   in the data link buffer.
		*/
		BLTC1 = tx_xfer_size = contig; //must be less than 2^16 !!
		/* Enable DMA channel 1*/
		DMACNTL1 |= CHEN;
	}
}


/******************************************************************************  
    Function: usart_dma_ch1_enable

	Description
	-----------
    This function enables DMA channel1 and sets the DMA counter if data
	is available in the tx data link buffer for transmission. The tx transfer
	size depends on the number of contiguous bytes available in the data link
	buffer since the DMA controller cannot wrap around when it reaches the end
	of the buffer.

	Arguments
	---------
	None.
*******************************************************************************/
static void usart_dma_ch1_enable(void)
{
	unsigned int contig;

	if (!(DMASTAT1 & (CHAC | TC))) {
		/* DMA channel 1 not active and no TC interrupt is pending. It's
		   OK to start a new DMA transfer.
		*/
		if ((contig = dl_get_max_linear_data_size(tx_dl, tx_ptr)) > 0) {
			/* Set DMA counter according to number of contiguous data bytes
			   in the data link buffer.
			*/
			BLTC1 = tx_xfer_size = contig;
			/* Enable DMA channel 1*/
			DMACNTL1 |= CHEN;
		}
	}
}

/******************************************************************************  
    Function: usart_dma_timeout

	Description
	-----------
    This function checks if the DMA receive clock has passed the timeout limit
	If it did, this means that a certain amout of time has passed since the
	last DMA rx transfer has been launched (or since the previous timeout)
	but the DMA counter has not reached the terminal count of zero. In this
	case we want to see if some new bytes have been transferred by the DMA
	and if so we want to immediately report them to the consumer (we don't
	want them to sit there forever).

	Arguments
	---------
	None.
*******************************************************************************/

static void usart_dma_timeout(void)
{
	unsigned int bltc0, new_bytes;

	/* Increment rx clock */
	rx_clock++;
	if (rx_clock > timeout) {
		/* Rx clock passed the timeout limit. Check for new bytes */
		__disable_interrupt();  
		bltc0 = BLTC0;
		if (bltc0 < rx_bytes_left) {
			new_bytes = rx_bytes_left - bltc0;
			rx_bytes_left = bltc0;
			/* New bytes available. Commit to consumer. */
			dl_p_commit(rx_dl, new_bytes);
	
			/* Reset Rx clock */
			rx_clock = 0;
		}
		__enable_interrupt();	
	}
}

/******************************************************************************  
    Function: usart_rx_isr

	Description
	-----------
    This function is the USART receive interrupt handler. Within this driver
    we use it only to be notified about errors. Data receive is handled by
	the DMA and not by the USART receive interrupt.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
/*
#ifdef __NSC_CRCC__
#pragma interrupt(usart_rx_isr)
void usart_rx_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+USARTRX_IRQ)
__interrupt void usart_rx_isr(void)
#else
__interrupt[(16+USARTRX_IRQ)*IDT_WIDTH] void usart_rx_isr(void)
#endif
#endif
{
  panic();
}
*/
