/************************************************************************
	  $File: //CR/cr16bt/drivers/usb.c $
	  $Date: 2003/07/09 $
	  $Revision: #29 $

	Function:	This is a driver for the USB peripheral. This file 
	contains routines more involved in the USB protocol itself. See 
	usbhw.c for the stuff closer to the HW.

	Public API provided (as exposed by drivers.h):
		usb_init()
		node_resume()
		get_node_status()
		usb_shutdown()

	API for interaction with usbhw.c 
		handle_std_usb_request()

************************************************************************/

#include "drivers.h"

#include <stdio.h>
#include <string.h>
#include "crcommon.h"
#include "chipdef.h"
#include "boarddef.h"


#ifdef UCOS
#include "ucos.h"
#endif

/* USB Global Variables */
usb_node_stat_t USBStatusInfo; // Holds current module status information
uint16 usISOSynchFrame[6]; // Holds synch frame for endpoints if ISO is used

// Defined and stored in usbhw.c not the neatest TODO - tidy
extern usb_init_info_t USBInitInfo;	// Initialisation information from user

/*** External Stuff defined in descrpt.h - unique to users app ***************/
extern const uint8 ucNROF_STRING_DESCRIPTORS;
extern const UsbDeviceDescriptorType DeviceDescriptor;
extern const t_TotalDescriptorType TotalDescriptor[];
extern const uint8 USBLangId[];
extern const uint8 * const lang1StringDescriptors[];
extern const t_EndpointSetup ** EPConfiguration[];

// Holds number of available interfaces for each configuration
extern const uint8 ucNoOfInterfaces[];	

/******************************************************************************  
    Function: usb_init

	Description
	-----------

	Arguments
	---------
	
*******************************************************************************/
extern boolean usb_init( usb_init_info_t *usbinitinfo )
{
	boolean bTest;

	/* Set the device address, config and interface to zero */
	BZERO(USBStatusInfo);

	set_device_state( NOT_ATTACHED );
	
	bTest = usb_hwinit( usbinitinfo );

	/* Unmask USB interrupt */
	icu_unmask(USB_IRQ);

	node_attach();

	// Wait until dialogue with host gets us to configured state 
	// if in blocking mode
	if ( !USBInitInfo.evt_clbk )
		while( USBStatusInfo.ucDeviceState != CONFIGURED );

	return bTest;
}

/******************************************************************************  
    Function: usb_shutdown

	Description
	-----------
	Detaches node from USB bus and shuts down the block and driver

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
void usb_shutdown( void )
{
#ifndef USB_COMPACT
	node_detach();

	/* Unmask USB interrupt */
	icu_mask(USB_IRQ);
	
    // Halt the block and disappear from the bus
    MCNTRL &= ~USBEN;
#endif
}

/******************************************************************************  
    Function: node_resume

	Description
	-----------
	Signals the resume on the USB bus, if in suspend state

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

boolean node_resume (uint8 ucState)
{
#ifndef USB_COMPACT
	uint8 ucRamALTEV;

	// First check if remote wake up is an enabled feature or not
	if (USBStatusInfo.ucWakeUpEnable == TRUE)
	{
		if (ucState == SIGNAL_RESUME)
		{
			if (USBStatusInfo.rem_wake_ok)
			{
				NFSR = RSM_ST;
				return TRUE;
			}
			else
				return FALSE;
		}
		else if (ucState == GO_OPERATIONAL)
		{
			// Recover all registers
			resume_endpoint();
			
			// TODO: Check that this is a satisfactory method
			// for checking the EOP
			// TODO: Perhaps get user to check status of EOP bit
			// through a function call?
			// Get RAM copy of ALTEV as it is clear on read
			// Wait for the EOP to be set, then we know
			// the resume is complete
			do
			{
				ucRamALTEV = ALTEV;
			} while (!(ucRamALTEV & EOP));
			
			// Check if EOP is set to tell us everything went ok!
			if (ucRamALTEV & EOP)
				return TRUE;
			else
				return FALSE;
		}
	}

#endif
	return FALSE;
}



/******************************************************************************  
    Function: get_node_status

	Description
	-----------
	Returns the current USB state of the node.

	Arguments
	---------
	None.
	
	Return value
	------------
	The current USB state of the node.

*******************************************************************************/

usb_node_stat_t get_node_status (void)
{
	return USBStatusInfo;
}

/******************************************************************************  
    Function: clear_stall_ep

	Description
	-----------
	Clears an endpoint stall as commanded by the host

	Arguments
	---------
	wIndex from USB command packet.
	
	Return value
	------------
	None.

*******************************************************************************/
#ifndef USB_COMPACT
static void clear_stall_ep( uint wIndex )
{
	int ep;

	// The Index holds the required endpoint
	ep = (2*(wIndex & 0x0F) - ((wIndex & 0x80) >> 4));

	if (ep < 1)
		EPC0 &= ~STALL;
	else if ((ep > 0) && (ep <7))
		EPC(ep) &= ~STALL;
}
#endif

/******************************************************************************  
    Function: clear_feature

	Description
	-----------
	Clears a usb feature through control endpoint zero. (USB 2.0 Sec 9.4.1)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void clear_feature (usb_request_t USBRequest)
{
#ifndef USB_COMPACT
	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
		// Handle Device requests
		if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE) &&
			((uint8) USBRequest.field.wValue == DEVICE_REMOTE_WAKEUP))
		{
			USBStatusInfo.ucWakeUpEnable = FALSE;
		}
		// Handle Endpoint requests -> only valid for EP0 in ADDRESS state
		// In ADDRESS mode can only clear EP0
		else 
			stall_ep0();
		break;
		// End of case ADDRESS

	case CONFIGURED:
		// Handle Device requests
		if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE) &&
			((uint8) USBRequest.field.wValue == DEVICE_REMOTE_WAKEUP))
		{
			USBStatusInfo.ucWakeUpEnable = FALSE;
		}
		// Handle Endpoint requests -> clears STALL on selected endpoint
		else if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPENDPOINT) &&
				((uint8) USBRequest.field.wValue == ENDPOINT_HALT))
			clear_stall_ep( USBRequest.field.wIndex );
		// If we get here, stall EP0 as command is not valid
		else
			stall_ep0();

		break;
		// End of case CONFIGURED

	default:
		/* STALL EP0 if the request is not valid */
		stall_ep0();
	}

	// Setup EP0 to send zero length packet - Note: STALL will still be sent
	// if STALL in EP0 was set
	sendEP0data (0, 0, NULL);
#endif
}

/******************************************************************************  
    Function: get_configuration

	Description
	-----------
	Returns the current configuration state. (USB 2.0 Sec 9.4.2)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void get_configuration (usb_request_t USBRequest)
{
	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
	case CONFIGURED:
		sendEP0data (1, 1, &USBStatusInfo.ucUSBConfiguration);
		break;
	
	default:
		/* STALL EP0 if the request is not valid */
		stall_ep0();
	}
}

/******************************************************************************  
    Function: get_descriptor

	Description
	-----------
	Starts the return of the requested USB descriptor. (USB 2.0 Sec 9.4.3)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
static void get_descriptor (usb_request_t USBRequest)
{
	uint16 usDescrptLen;
	uint8 ucDescrptType, ucDescrptIndx;
  
	// Get required values from usb_request_t
	usDescrptLen = USBRequest.field.wLength;
	ucDescrptIndx = (uint8) USBRequest.field.wValue;
	USBRequest.field.wValue >>= 8;
	ucDescrptType = (uint8) USBRequest.field.wValue;

	// TODO: Check to see if the first request could ever be less than eight bytes
	//		 or if a descriptor could ever be shorter than eight bytes

	// Determine descriptor type
	switch (ucDescrptType)
	{
	case DEVICE:
		// Send requested amount of requested descriptor 
		// (USB 2.0 9.6.2 & Table 9-8)
		sendEP0data (usDescrptLen,(uint16) DeviceDescriptor.bLength,
					 (uint8 *) &DeviceDescriptor);
  	break;

	case CONFIGURATION:
		// wValue's index supports multiple configurations
		sendEP0data (usDescrptLen, 
					 TotalDescriptor[ucDescrptIndx].usLengthOfDesc,
					 TotalDescriptor[ucDescrptIndx].ucpConfigDesc);
		break;

	case STRING:
		// Work out which string is being requested from list
		if (ucDescrptIndx < ucNROF_STRING_DESCRIPTORS)
		{
			// Find out if the request was for the LANGID

			if (ucDescrptIndx == 0)
				sendEP0data (usDescrptLen, (uint16)USBLangId[0],
							 (uint8 *) &USBLangId);
			else // Get a pointer to the string descriptor pointed by the index
				sendEP0data (usDescrptLen, 
					(uint16)lang1StringDescriptors[ucDescrptIndx-1][0],
					(uint8 *) lang1StringDescriptors[ucDescrptIndx-1]);
		}
		else
			stall_ep0_send00();
		break;

	default:
		stall_ep0_send00();
	}
}

/******************************************************************************  
    Function: get_interface

	Description
	-----------
	Returns the selected alternate setting. (USB 2.0 Sec 9.4.4)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void get_interface (usb_request_t USBRequest)
{
	uint8 ucInterface;
	
	// Get the Interface number
//	USBRequest.field.wIndex >>= 8;
	ucInterface = (uint8) USBRequest.field.wIndex;

	// Perform request according to device state
	if ((USBStatusInfo.ucDeviceState == CONFIGURED) &&
		(ucInterface == USBStatusInfo.ucUSBInterface))
			// Send the alternate setting value
			sendEP0data (1, 1, &USBStatusInfo.ucUSBAlternate);
	else
		stall_ep0_send00();
}

/******************************************************************************  
    Function: stall_requested_ep

	Description
	-----------
	Takes a USB request field (wIndex) and stalls the endpoint referred to.
	If a non existent endpoint is referred to then stall EP0.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
static void stall_requested_ep( uint wIndex )
{
#ifndef USB_COMPACT
	int ep;
	uint stat;

	// The Index holds the required endpoint
	ep = (2*(wIndex & 0x0F) - ((wIndex & 0x80) >> 4));

	if (ep < 1)
	{
		stall_ep0();
		// Send the endpoint status setting value
		if (EPC0 & STALL)
			stat = 0x0001;

		sendEP0data (2, 2, (uint8 *)&stat);
	}
	else if ((ep > 0) && (ep <7))
		{
			stall_ep((ep_id_t)ep);
			// Send the endpoint status setting value
			if (EPC(ep) & STALL)
				stat = 0x0001;

			sendEP0data (2, 2, (uint8 *)&stat);
		}
		else
			stall_ep0_send00();
#endif
}

/******************************************************************************  
    Function: get_status

	Description
	-----------
	Returns status of specified recipient (USB 2.0 Sec 9.4.5)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void get_status (usb_request_t USBRequest)
{
	uint16 usStatus = 0x0000;

	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
		// We can only respond to a device or endpoint zero requests
		// First device requests
		if ((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE)
		{
			// Return current Remote Wakeup and Self Powered status USB 2.0 Fig 9-4
			if (USBStatusInfo.ucWakeUpEnable)
				usStatus |= 0x0002;

			if (USBStatusInfo.ucSelfPowered)
				usStatus |= 0x0001;

			// Send the device status setting value
			sendEP0data (2, 2, (uint8 *)&usStatus);
		}
		// Next Control Endpoint Zero only requests
		else if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPENDPOINT) &&
				   USBRequest.field.wIndex == 0x0000)
		{
			if (EPC0 & STALL)
				usStatus = 0x0001;

			// Send the device status setting value
			sendEP0data (2, 2, (uint8 *)&usStatus);

		}
		else
			stall_ep0_send00();
		break;

	case CONFIGURED:
		// As long as the requested interface exists, respond appropriately
		if ((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE)
		{
			// Return current Remote Wakeup and Self Powered status USB 2.0 Fig 9-4
			if (USBStatusInfo.ucWakeUpEnable)
				usStatus |= 0x0002;

			if (USBStatusInfo.ucSelfPowered)
				usStatus |= 0x0001;

			// Send the device status setting value
			sendEP0data (2, 2, (uint8 *)&usStatus);
		}
		// Interface requests always return 0x0000
		else if ((USBRequest.field.bmRequestType & RECIPIENT) == RECIPINTERFACE)
			// Send the interface status setting value
			sendEP0data (2, 2, (uint8 *)&usStatus);

		// Now for Endpoint requests
		else if ((USBRequest.field.bmRequestType & RECIPIENT) == RECIPENDPOINT)
			stall_requested_ep( USBRequest.field.wIndex );

		break;

	default:
		stall_ep0_send00();
	}
}

/******************************************************************************  
    Function: set_address

	Description
	-----------
	Sets the device address. (USB 2.0 Sec 9.4.6)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void set_address (usb_request_t USBRequest)
{
	// Get our address from the wValue
	switch (USBStatusInfo.ucDeviceState)
	{
	// The processing for the DEFAULT and ADDRESS states is basically the same
	// but a change from ADDRESS to DEFAULT state occurs if the address given
	// is 0x00 and we are currently in the ADDRESS state. However the actual
	// change of state should occur once the complete SET_ADDRESS request is
	// complete, which only occurs on the reception of the final ACK in the
	// EP0 ISR.
	case DEFAULT:
	case ADDRESS:
		// Get a copy of the address and set Address Enable bit
		//USBStatusInfo.ucUSBAddress = (uint8) USBRequest.field.wValue | 0x80;
		USBStatusInfo.ucUSBAddress = (uint8) USBRequest.field.wValue;
		// Use DEF bit functionality to setup new address, but respond to 
		// default address for the next IN token
		EPC0 |= DEF;
		FARREG = USBStatusInfo.ucUSBAddress | 0x80;
		// And now send a zero length packet
		sendEP0data (0, 0, NULL);
		// Set our status to the ADDRESS state
		set_device_state( ADDRESS );
		break;
	default:
		// Stall EP0
		stall_ep0_send00();
	}

}

/******************************************************************************  
    Function: set_configuration

	Description
	-----------
	Sets a new device configuration. (USB 2.0 Sec 9.4.7)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void set_configuration (usb_request_t USBRequest)
{
	uint8 ucConfiguration;

	// Note chosen configuration
	ucConfiguration = (uint8) USBRequest.field.wValue;

	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
	case CONFIGURED:
		// If the Configuration Value is 0x00,
		// and we are in the ADDRESS state then stay in the
		// ADDRESS state. (NOTE: Upper byte of wValue is reserved)
		// If Config value is 0x00 and we are in CONFIGURATION state,
		// move back into the ADDRESS state
		// Else take on the requested configuration if it is available
		// and if in ADDRESS state move into CONFIGURATION state.
		if (ucConfiguration <= DeviceDescriptor.bNumConfigurations)
		{
			reset_ep_pids();

			// If the configuration requested is non-zero, this
			// will be updated in the next 'if' statement.

			// Change our state to ADDRESS
			set_device_state( ADDRESS );
			
			// Set the current configuration value to zero
			USBStatusInfo.ucUSBConfiguration = 0x00;

			if (ucConfiguration != 0x00)
			{
				// Decrement value by one to suite array
				//ucConfiguration -= 1;

				// Setup the endpoints to Interface 0, Alt Setting 0
				setup_endpoints (ucConfiguration, 0,0);

				// Now change our state to CONFIGURED
				set_device_state( CONFIGURED );

				// and also note the configuration value
				//USBStatusInfo.ucUSBConfiguration = ucConfiguration + 1;
				USBStatusInfo.ucUSBConfiguration = ucConfiguration;

				if (USBInitInfo.evt_clbk)
	    			USBInitInfo.evt_clbk( USB_CONFIGURED, RXTXEP0 );
			}
		}
		// Otherwise issue a Request Error
		else
			stall_ep0();

		break;

	default:
		stall_ep0();
	}

	// Setup EP0 to send zero length packet
	// Note: STALL will still be sent if STALL in EP0 was set
	sendEP0data (0, 0, NULL);
}

/******************************************************************************  
    Function: set_descriptor

	Description
	-----------
	Update existing or add new descriptors (USB 2.0 Sec 9.4.8)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
static void set_descriptor (usb_request_t USBRequest)
{

	// Implemented as a user callback function
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
	case CONFIGURED:
		if (USBInitInfo.set_desc_clbk)
			USBInitInfo.set_desc_clbk(USBRequest);
		else
			stall_ep0();
		break;
	default:
		stall_ep0();
	}

	sendEP0data (0, 0, NULL);

}
#endif

/******************************************************************************  
    Function: set_feature

	Description
	-----------
	Sets a specific feature. (USB 2.0 Sec 9.4.9)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
static void set_feature (usb_request_t USBRequest)
{
	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
		// Handle Device requests
		if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE) &&
			 ((uint8) USBRequest.field.wValue == DEVICE_REMOTE_WAKEUP))
			USBStatusInfo.ucWakeUpEnable = TRUE;

		// Handle Endpoint requests -> only valid for EP0 in ADDRESS state
		// In ADDRESS mode can only clear EP0
		else if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPENDPOINT)
		         && ((USBRequest.field.wIndex & 0x0080) 
		         |  (USBRequest.field.wIndex & 0x0000)) 
		         &&	((uint8) USBRequest.field.wValue == ENDPOINT_HALT))
			stall_ep0();
		else // If we get here, stall EP0 as command is not valid
			stall_ep0();

		break;
		// End of case ADDRESS

	case CONFIGURED:
		// Handle Device requests
		if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPDEVICE) &&
			 ((uint8) USBRequest.field.wValue == DEVICE_REMOTE_WAKEUP))
			USBStatusInfo.ucWakeUpEnable = TRUE;
		// Handle Endpoint requests -> sets STALL on selected endpoint
		else if (((USBRequest.field.bmRequestType & RECIPIENT) == RECIPENDPOINT)
		        && ((uint8) USBRequest.field.wValue == ENDPOINT_HALT))
			stall_requested_ep( USBRequest.field.wIndex );
		break;
		// End of case CONFIGURED

	default:
		/* STALL EP0 if the request is not valid */
		stall_ep0();
	}

	// Setup EP0 to send zero length packet - Note: STALL will still be sent
	// if STALL in EP0 was set
	sendEP0data (0, 0, NULL);
}
#endif

/******************************************************************************  
    Function: set_interface

	Description
	-----------
	Allows selection of alternative setting for specified interface.
	(USB 2.0 Sec 9.4.10)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
static void set_interface (usb_request_t USBRequest)
{
	uint Interface;
	uint AltSetting;
	uint PossConfig;

	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
		stall_ep0();
		break;
	case CONFIGURED:
		// Change to alternate interface
		// Retrieve requested interface from request
		Interface = USBRequest.field.wIndex;
		// Retrieve requested alternate setting from request
		AltSetting = USBRequest.field.wValue;
		// Check to see if interface exists
		
		// Temp value just to tidy things a but
		PossConfig = USBStatusInfo.ucUSBConfiguration-1;
		
		if (Interface < ucNoOfInterfaces[PossConfig])
		{
			// Check to see if alternate setting exists for this
			// configuration, interface and alternate setting
			if (EPConfiguration[PossConfig][Interface][AltSetting].bActiveEndpoint)
			{
				// If array position was not 0x00, we know this 
				// Configuration, Interface and Alt Setting exists
				// Now set this to be the current interface
				setup_endpoints(USBStatusInfo.ucUSBConfiguration, 
									Interface, AltSetting);

				// Now note new settings
				USBStatusInfo.ucUSBInterface = Interface;
				USBStatusInfo.ucUSBAlternate = AltSetting;
			}
			else
				stall_ep0();
		}
		else
			stall_ep0();
		break;
	default:
		stall_ep0();
	}
	// Setup EP0 to send zero length packet - Note: STALL will still be sent
	// if STALL in EP0 was set
	sendEP0data (0, 0, NULL);
}
#endif

/******************************************************************************  
    Function: synch_frame

	Description
	-----------
	Sets and reports an endpoints sync frame. (USB 2.0 Sec 9.4.11)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void synch_frame (usb_request_t USBRequest)
{
	uint16 usFrameNumber;
	uint  ucEndpoint;

	// Perform request according to device state
	switch (USBStatusInfo.ucDeviceState)
	{
	case ADDRESS:
		stall_ep0();
		break;

	case CONFIGURED:
		if (USBRequest.field.wValue == 0x0000 
		    && USBRequest.field.wLength == 0x0002)
		{
			// Work out which endpoint (1 to 3) is being requested
			ucEndpoint = (uint8) USBRequest.field.wIndex & 0x000F;
			
			// Make ucEndpoint value 2, 4 or 6
			ucEndpoint *= 2;
			
			// Look at direction bit to make 1, 3 or 5 if required
			if (USBRequest.field.wIndex & 0x0080)
				ucEndpoint -= 1;

			// Get agreed synch frame number for enpoint in wIndex
			usFrameNumber = usISOSynchFrame[ucEndpoint];

			// Send data through EP0
			sendEP0data (2, 2, (uint8 *) &usFrameNumber);
		}
		else
			stall_ep0();
		break;

	default:
		stall_ep0();
	}
}

/******************************************************************************  
    Function: handle_std_usb_request

	Description
	-----------
    Select the relevant routine to handle standard USB requests received. 
    Called from usb.c

	Arguments
	---------
	
	Return value
	------------
	None.

*******************************************************************************/
void handle_std_usb_request( usb_request_t USBRequest )
{
	switch (USBRequest.field.bRequest)
	{
	case CLEAR_FEATURE:  
		clear_feature(USBRequest);
		break;

	case GET_CONFIGURATION:   
		get_configuration(USBRequest);
		break;

	case GET_DESCRIPTOR:
		get_descriptor(USBRequest);
		break;

	case GET_INTERFACE: 
		get_interface(USBRequest);
		break;

	case GET_STATUS: 
		get_status(USBRequest);
		break;

	case SET_ADDRESS:	
		set_address(USBRequest);
		break;

	case SET_CONFIGURATION: 
		set_configuration(USBRequest);
		break;
#ifndef USB_COMPACT
	case SET_DESCRIPTOR:
		set_descriptor(USBRequest);
		break;

	case SET_FEATURE:  
		set_feature(USBRequest);
		break;

	case SET_INTERFACE: 
		set_interface(USBRequest);
		break;
#endif
	case SYNCH_FRAME:
		synch_frame(USBRequest);
		break;

	// Unsupported Request -> Stall the EP0
	default:
		stall_ep0_send00();
	}
}

/******************************************************************************  
    Function: set_device_state

	Description
	-----------
    Set the current USB device state. Only for use by usb.c and usbhw.c

	Arguments
	---------
	The new device state.
	
	Return value
	------------
	None.

*******************************************************************************/
void set_device_state( usb_dev_state_t new_dev_state )
{
	USBStatusInfo.ucDeviceState = new_dev_state;
}

/******************************************************************************  
    Function: set_remote_wake_ok

	Description
	-----------
    Set the current remote wake enable. Only for use by usb.c and usbhw.c

	Arguments
	---------
	The new wake enable state.
	
	Return value
	------------
	None.

*******************************************************************************/
void set_remote_wake_ok( boolean wake_ok )
{
	USBStatusInfo.rem_wake_ok = wake_ok;
}

/******************************************************************************  
    Function: indicate_suspended

	Description
	-----------
    Set the current suspend state. Only for use by usb.c and usbhw.c

	Arguments
	---------
	The new device suspend state.
	
	Return value
	------------
	None.

*******************************************************************************/
void indicate_suspended( boolean indic_susp )
{
	USBStatusInfo.suspended = indic_susp;
}
