/************************************************************************
	$File: //CR/cr16bt/drivers/usbhw.c $
	$Date: 2003/07/08 $
	$Revision: #7 $

	Function:	This is the HW driver for the USB peripheral. usb.c
		calls the api provided by this module.
			
************************************************************************/

#include "drivers.h"

#include <stdio.h>
#include <string.h>
#include "crcommon.h"
#include "chipdef.h"
#include "boarddef.h"

#ifdef UCOS
#include "ucos.h"
#endif

// Variables accessed by usb.c - should hide these at some point
usb_init_info_t USBInitInfo;	// Initialisation information from user
usb_t * const usb = (usb_t *)USB_BASE_ADDR; // Our device register block

// Local variables - well hidden from the world
static usb_ep_stat_t USBEPSupport[N_ENDPOINTS];	// EP status information 

#ifndef USB_COMPACT
static uint8 ucSavedEPCStatus[N_ENDPOINTS]; // Status of EPC reg's during suspend
static uint8 ucSavedTXCStatus; // Status of TX_EN and TX_LAST for TXC 
							   // reg's during suspend
static uint8 ucSavedRXCStatus;	// Status of RX_EN for RXC reg's during suspend
#endif
static uint8 ucEndpointPID;	// Endpoint DATA PID Status in bit format
static uint16 usDescriptorRemaining; // Holds how many bytes of requested 
								     // descriptor still need to be transmitted
static uint8 *pucDescriptorPointer;	// Pointer to requested descriptor
static uint16 usBuffRemainBytes[N_ENDPOINTS]; // Bytes remaining in each user EP buffer
static uint8 *pucCurrentBuffAddr[N_ENDPOINTS];	// Current pointer to address of next 
										// user EP buffer storage location

static boolean bZeroLengthPacket;	// Used to tell EP0 that it should send a  
					//zero length packet at end of modulo 8 byte length of data
static uint ep_copy_cnt[N_ENDPOINTS]; // Records amount of data copied into 
								// endpoints buffer that has not yet been ACKed

/*** External Stuff defined in descrpt.h - unique to users app ***************/
extern const t_EndpointSetup ** EPConfiguration[];

// Internal funcions called by the public API that need forward declarations
static void copy_to_tx_buff( uint Endpoint, uint txep );
static void toggle_txep_pid( uint ep, uint txep );

/******************************************************************************  
    Function: usb_tx_data

	Description
	-----------
    Set up a data transmit/IN through chosen EPs (1,3,5)

	Arguments
	---------
	ep: indicate one of the tx eps 1,3,or 5
	data_addr: pointer to users buffer of data
	len: Number of bytes to transfer (0 < len < 65)
	
	Return value
	------------
	If there's no event callback defined then the number of bytes actually 
	transferred is returned, otherwise zero.
*******************************************************************************/

uint usb_tx_data (ep_id_t ep, uint8 * data_addr, uint len )
{
	// Check to see that ep is a TX ep (odd numbered ep)
	if ( ep & 0x01 ) 
	{
		uint txep = txep_from_ep(ep);

		// No bytes transferred yet
		USBEPSupport[ep].xfer_cnt = 0;

		// Find out how many bytes we must send
		usBuffRemainBytes[ep] = len;

		// Retain ep address pointer to storage location
		pucCurrentBuffAddr[ep] = data_addr;

		copy_to_tx_buff( ep, txep );

		// Change status to transferring
		USBEPSupport[ep].ucTransferStatus = TRNSFRING;		

		// Enable TX'ing ready for next IN
		TXC(txep) |= TX_EN;

		// If no callback block until transfer complete
		if (!USBInitInfo.evt_clbk)
		{
			while( USBEPSupport[ep].ucTransferStatus == TRNSFRING );		

			return USBEPSupport[ep].xfer_cnt;
		}
	}

	return 0;
}

/******************************************************************************  
    Function: usb_rx_data

	Description
	-----------
    Set up a data receive/OUT through chosen EPs (2,4,6)

	Arguments
	---------
	ep: indicate one of the tx endpoints 2,4, or 6
	data_addr: pointer to users buffer of data
	len: Number of bytes to transfer (0 < usBytes < 65)
	
	Return value
	------------
	If there's no event callback defined then the number of bytes actually 
	transferred is returned, otherwise zero.
*******************************************************************************/

uint usb_rx_data (ep_id_t ep, uint8 *data_addr, uint len )
{
	// Check to see that ep is an RX ep (even numbered ep)
	if ( !(ep & 0x01) ) 
	{
		uint rxep = rxep_from_ep(ep);

		USBEPSupport[ep].xfer_cnt = 0;

		// Find out how many bytes we can receive
		usBuffRemainBytes[ep] = len;

		// Reset EP address pointer to storage location
		pucCurrentBuffAddr[ep] = data_addr;

		// Change status to transferring
		USBEPSupport[ep].ucTransferStatus = TRNSFRING;

		// Now set up endpoint to receive on next OUT token
		RXC(rxep) |= RX_EN;

		// If no callback, block until transfer complete
		if (!USBInitInfo.evt_clbk)
		{
			while( USBEPSupport[ep].ucTransferStatus == TRNSFRING );	

			return USBEPSupport[ep].xfer_cnt;
		}
	}
	return 0;
}

/******************************************************************************  
    Function: get_ep_stat

	Description
	-----------
    Return the current state of an endpoint and the number of bytes transferred 
	through that endpoint since the last rx/tx_data() call for the endpont.

	Arguments
	---------
	ep: Select which endpoint to get data for.
	curr_stat: returns which transfer state the endpoint is currently in
	xfer_cnt: The number of bytes tranferred through this endpoint since the
	    last rx/tx_data() call for this endpoint.

	Return value
	------------
	None.

*******************************************************************************/
void get_ep_stat( ep_id_t ep, ep_xfer_stat_t *curr_stat, uint *xfer_cnt )
{
#ifndef USB_COMPACT
	if (ep < 7)
	{
		*curr_stat = USBEPSupport[ep].ucTransferStatus;
		*xfer_cnt = USBEPSupport[ep].xfer_cnt;
	}
#endif
}


/******************************************************************************  
    Function: sendEP0data

	Description
	-----------
	Fills EP0 Tx FIFO and enables the EP to transmit when an IN token is 
	recieved. This allows all the USB Device Requests to easily send response
	data, zero length packets or prime the EP0 FIFO with descriptor
	information.

	Arguments
	---------
	req_len: Length of data requested by host
	avail_len: Length of data available on device
	data_addr: Pointer to data
	
	Return value
	------------
	None.
*******************************************************************************/
void sendEP0data (uint req_len, uint avail_len, uint8 *data_addr)
{
	uint8 ucEP0BytesToSend, ucCount;

	// First flush FIFO
	TXC0 |= TX_FLUSH;

	// Clear the global variable holding the number of remaing bytes to send 
	// through EP0, as used by us and the txep0_handler
	usDescriptorRemaining = 0;
	
	// Copy the pointer to the data to our global pointer used by us
	// and the txep0_handler ISR so we are synched
	pucDescriptorPointer = data_addr;

	// Check to see if more bytes were requested than actually available
	// Choose the lower value
	req_len = (avail_len < req_len) ? avail_len : req_len;

	// Note how many bytes remain to be sent in the global variable, as
	// also used by txep0_handler ISR
	usDescriptorRemaining = req_len;

	// We can only send a max of 8 bytes through EP0, so see
	// how many we should send
	ucEP0BytesToSend = ((req_len < 8) ? (uint8) req_len : 8);

	// If ucNoOfBytes is between 1 and 8, copy bytes to FIFO
	if (ucEP0BytesToSend > 0 && ucEP0BytesToSend < 9)
	{
		// Use a copy of ucEP0BytesToSend here as we need to use it later
		ucCount = ucEP0BytesToSend;
		while (ucCount != 0)
		{
			TXD0 = *pucDescriptorPointer++;
			ucCount--;
			usDescriptorRemaining--;
		}
	}
	// If ucNoOfBytes is zero, we will send a zero length packet by default

	// Set the TOGGLE bit correctly
	toggle_txep_pid(0,0);

	// Now enable the endpoint and watch that data go with the next IN token!
	TXC0 |= TX_EN;

	// Now tidy up the global variables so the txep0_handler ISR knows what
	// to send when the interrupt occurs
	// First check to see if an EP0 length amount of data was requested
	// i.e. 8 bytes, as we need to follow this with a zero-length packet
	if (req_len == 8)
		bZeroLengthPacket = TRUE;
}

/******************************************************************************  
    Function: stall_ep0

	Description
	-----------
	Create stall condition on endpoint 0

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
void stall_ep0() 
{
	EPC0 |= STALL;
}

/******************************************************************************  
    Function: stall_ep

	Description
	-----------
	Create stall condition for the given endpoint eps 1-6 (NOT 0)

	Arguments
	---------
	ep: selected endpoint.
	
	Return value
	------------
	None.
*******************************************************************************/
void stall_ep(ep_id_t ep) 
{
	EPC(ep) |= STALL;
}

/******************************************************************************  
    Function: usb_hwinit

	Description
	-----------
    This function initialises the USB hardware. This function
	returns a value of TRUE unless the USB module cannot be detected, or if the
	PLL has not been enabled. 
	After returning from this function the user must call the usb_attach() 
	function to attach the node to the USB bus.

	Arguments
	---------
	ucFrameHandler : 
	Defines if user wants to use the FRAME interrupt to perform various checks
	and data handling instead of manual calls to their associated 'update'
	functions.

*******************************************************************************/

boolean usb_hwinit( usb_init_info_t *usbinitinfo )
{

#ifndef USB_COMPACT
	boolean bError = FALSE;


	/* Check to see if the PLL is on */
	if (CRCTRL & PLLPWD)
		return bError;
#endif
	/* Setup transceiver power mode */
	MCFG |= USB_ENABLE;

	/* Enable the USB module */
	MCNTRL |= USBEN;

	/* Set up USB interrupt masks */
	/* Switch off NAK event masks  */
	NAKMSK |= (NAK_I0 | NAK_O0);

	/* Enable EP0 TX complete mask */
	TXMSK |= TXFIFO0;

	/* Enable EP0 RX complete mask */
	RXMSK |= RXFIFO0;

	/* Set up alternate event mask */
	ALTMSK |= (SD3 | SD5 | RESET_A | RESUME_A);

	/* Set up main mask */
	MAMSK |= (WARN | ALT | TX_EV | NAK | ULD | RX_EV | INTR_E);

	// Clear the zero length packet variable
	bZeroLengthPacket = FALSE;

	/* Get the usb initialisation information */
	if (usbinitinfo)
		USBInitInfo = *usbinitinfo;

	return TRUE;
}

/******************************************************************************  
    Function: node_attach

	Description
	-----------
	Attaches node to USB bus

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

void node_attach (void)
{
	// Set node state to operational
	// NOTE: Must make node state operational as all endpoints are disabled
	// when NFSR = NodeReset.
	NFSR |= OPR_ST;

	// Flush then enable RX EP0
	RXC0 |= RX_FLUSH;
	RXC0 |= RX_EN;

	// Set EPC0 to respond to default address
	FARREG = 0x80;
	EPC0 |= DEF;

	// Attach node
	MCNTRL |= NAT;

	// Change device state
	set_device_state( POWERED );
}

/******************************************************************************  
    Function: node_detach

	Description
	-----------
	Detaches node from USB bus

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
void node_detach (void)
{
	// Set node functional state to reset
	NFSR |= RST_ST;

	// Flush then enable RX EP0
	RXC0 |= RX_FLUSH;

	// Set EPC0 to respond to default address
	FARREG = 0x80;
	EPC0 |= DEF;

	// Detach node
	MCNTRL &= ~NAT;

	set_device_state( NOT_ATTACHED );
}
#endif
/******************************************************************************  
    Function: reset_ep_pids

	Description
	-----------
	Ensure that all Endpoint (except EP0) TOGGLE's (for Data0/1 packets)are 
	cleared to zero.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
void reset_ep_pids( void )
{
	ucEndpointPID &= 0x01;
}


/******************************************************************************  
    Function: clear_ep_stall

	Description
	-----------
	Clear stall condition for the given endpoint eps 1-6 (NOT 0)

	Arguments
	---------
	ep: selected endpoint.
	
	Return value
	------------
	None.
*******************************************************************************/
void clear_ep_stall(ep_id_t ep) 
{
	EPC(ep) &= ~STALL;
}


/******************************************************************************  
    Function: resume_endpoint

	Description
	-----------
	Performs recovery of registers when return to Node Operational state
	(seperate function as it is used by RESUME signalling, and remote
	wakeup support)

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

void resume_endpoint (void)
{
#ifndef USB_COMPACT
	uint ep, bit_flag;
#endif

	// Set the endpoint to operational
	NFSR = OPR_ST;

#ifndef USB_COMPACT
	EPC0 = ucSavedEPCStatus[0];
	// Set EPC status registers to values prior to suspend
	for( ep = 1; ep < N_ENDPOINTS; ep++ )
		EPC(ep) = ucSavedEPCStatus[ep];

	bit_flag = 1;
	// Retrieve all R/TX_EN and LAST status for R/TXC registers
	for( ep = 0; ep < 4; ep++ )
	{
		if (ucSavedTXCStatus & bit_flag)
			TXC(ep) |= TX_EN;
		if (ucSavedRXCStatus & bit_flag)
			RXC(ep) |= RX_EN;
		
		bit_flag <<= 1;

		if (ucSavedTXCStatus & bit_flag)
			TXC(ep) |= TX_LAST;

		bit_flag <<= 1;
	}

	// Note that the node is no longer suspended
	indicate_suspended( FALSE );

	// Note that remote wake up is no longer supported
	set_remote_wake_ok( FALSE );

	// Mask on the SD3 and SD5 bits again, so we respond to new SUSPEND
	// signalling
	ALTMSK |= (SD3 | SD5);
	
   	if (USBInitInfo.evt_clbk)
    	USBInitInfo.evt_clbk( USB_RESUME, RXTXEP0 );
#endif
}


/******************************************************************************  
    Function: setup_endpoints

	Description
	-----------
	Sets up the endpoints as per predefined settings passed to function
	NOTE: Interface and Alt Interface passed to function MUST exist

	Arguments
	---------
	ucConfiguration: selected configuration
	ucInterface: selected interface
	ucAltSetting: alternate setting
	
	Return value
	------------
	None.
*******************************************************************************/
void setup_endpoints (uint8 ucConfiguration, uint8 ucInterface, 
						uint8 ucAltSetting)
{
	uint rxtxep;
	uint en_flag;
	uint ep;
	const t_EndpointSetup *ep_cfg_ptr;
	
	// Turn off all enabled TX endpoints
	TXC(1) = 0x00;
	TXC(2) = 0x00;
	TXC(3) = 0x00;

	// Clear previous configuration settings
	for( ep = 1; ep < N_ENDPOINTS; ep++)
		EPC(ep) = 0x00;

	TXMSK = TXFIFO0;
	RXMSK = RXFIFO0;

	reset_ep_pids();
	
	// Subtract 1 from ucConfiguration to fit EPConfiguration array
	// NOTE: Configuration value is always => 1
	ucConfiguration--;

	ep_cfg_ptr = &EPConfiguration[ucConfiguration][ucInterface][ucAltSetting];

	// Set EP_EN and address for each active endpoint against
	// current configuration, interface and alternate setting chosen
	ep = 1;
	en_flag = 1;
	for( rxtxep = 1; rxtxep < 4; rxtxep++ )
	{
		if (ep_cfg_ptr->bActiveEndpoint	& EP_BIT(ep))
		{
			EPC(ep) |= (rxtxep | EP_EN);
			TXC(rxtxep) |= TX_FLUSH;
			TXMSK |= EP_BIT(rxtxep);
		}

		ep++;

		if (ep_cfg_ptr->bActiveEndpoint & EP_BIT(ep))
		{
			EPC(ep) |= (rxtxep | EP_EN);
			RXC(rxtxep) |= RX_FLUSH | IGN_SETUP | RX_EN;
			RXMSK |= EP_BIT(rxtxep);
		}

		ep++;
		en_flag <<= 1;
	}

	USBEPSupport[0].ucTransferStatus = TRNSFRSTOPPED;

	for( ep = 1; ep < N_ENDPOINTS; ep++)
		if (ep_cfg_ptr->bIsoEndpoint & EP_BIT(ep))
			USBEPSupport[ep].ucTransferStatus = TRNSFRSTOPPED;
}

/******************************************************************************  
    Function: toggle_txep_pid

	Description
	-----------
	Toggle the PID in HW for a TX endpoint.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
static void toggle_txep_pid( uint ep, uint txep )
{
	if (ucEndpointPID & EP_BIT(ep))
		TXC(txep) |= TX_TOGL;
	else
		TXC(txep) &= ~TX_TOGL;
}


/******************************************************************************  
    Function: stall_ep0_send00

	Description
	-----------
    Combines the common combination of stalling ep0 then sending 0 len packet
	to stall ep0.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
void stall_ep0_send00( void )
{
	stall_ep0();
	sendEP0data(0, 0, NULL);
}

/******************************************************************************  
    Function: rxep0_handler

	Description
	-----------
    This function handles the RX EP0 endpoint

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
static void rxep0_handler (void)
{
	usb_request_t USBRequest;
	uint8 ucLoop;
	
	/* Put RXS0 into RAM (clear on read register) */
	uint8 RamRXS0 = RXS0;

	/* See if a setup packet was received and that it was ACKed*/
	if ((RamRXS0 & SETUP_R) && (RamRXS0 & RX_LAST))
	{
		/* Check to see if the PID is 0 */
		if ((RamRXS0 & RX_TOGL) == 0)
		{
			/* Check to see if we have eight bytes of data */
			if ((RamRXS0 & RCOUNT) == 0x08)
			{
				/* Clear any STALL condition */
				EPC0 &= ~(STALL);
				
				// Set PID to DATA1
				ucEndpointPID |= EP_BIT(0);

				// Copy request into request structure
				for (ucLoop = 0; ucLoop < 8 ; ucLoop++ )
					USBRequest.byte[ucLoop] = RXD0;
                
				switch (USBRequest.field.bmRequestType & TYPE)
				{
				// Handle 'Standard' requests for all recipients
				// Note: May pass request to external function
				// if it is not handled by us

				// TODO: Should we check the RequestType direction
				// bit????
				case TYPESTANDARD:
					handle_std_usb_request( USBRequest );
					break;

#ifndef USB_COMPACT
				// Handle 'Class' requests through callback function or STALL
				// if unavailable
				case TYPECLASS:
					if (USBInitInfo.class_clbk)
						USBInitInfo.class_clbk(USBRequest);
					else
						stall_ep0_send00();
					break;
#endif

				// Handle 'Vendor' requests through callback function or STALL
				// if unavailable	
				case TYPEVENDOR:
					if (USBInitInfo.vendor_clbk)
						USBInitInfo.vendor_clbk(USBRequest);
					else
						stall_ep0_send00();
					break;

				// Default - causes a STALL to occur
				default:
					stall_ep0_send00();
				}

			}
			/* If not, we stall the endpoint until the next SETUP packet */
			/* USB1.1 - Sec. 8.5.2 */
			else
				stall_ep0_send00();
		}		
	}
	// If we didn't get a SETUP packet, this might be a zero-length 
	// status packet! As per USB 2.0 Sec. 8.5.3
	else if ((RamRXS0 & RX_LAST) && ((RamRXS0 & RCOUNT) == 0))
	{
		// Also check that it was DATA1 PID!!!
		if (RamRXS0 & RX_TOGL)
		{
			// Flush any data out of TXEP0 FIFO
			if (TXC0 & TX_EN)
			{
				TXC0 &= ~TX_EN;
				TXC0 |= TX_FLUSH;
			}
			// State that any descriptors being sent are complete
			// by zeroing usDescriptorRemaining
			usDescriptorRemaining = 0;
		}
	}
}

/******************************************************************************  
    Function: txep0_handler

	Description
	-----------
    This function handles the TX EP0 endpoint

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void txep0_handler (void)
{
	/* Put TXS0 into RAM (clear on read register) */
	uint8 RamTXS0 = TXS0;
	uint8 ucEP0BytesToSend, ucCount;

	// Make sure that the TX_DONE bit is set
	if (RamTXS0 & TX_DONE)
	{
		// Also check that ACK_STAT is set (packet was acknowledged)
		if (RamTXS0 & ACK_STAT)
		{
			// Toggle the EP0 PID
			if (ucEndpointPID & EP_BIT(0))
				ucEndpointPID &= ~EP_BIT(0);
			else
				ucEndpointPID |= EP_BIT(0);

			// If there are some more bytes to send, send them
			if (usDescriptorRemaining)
			{
				// If the remaining descriptor = 8 then we will need to 
				// send a zero length packet next time to complete the
				// transaction
				if (usDescriptorRemaining == 8)
					bZeroLengthPacket = TRUE;

				// Find out how many bytes to send (if usDescriptorRemaining
				// < 8 as EP0 is only 8 bytes deep)
				ucEP0BytesToSend = ((usDescriptorRemaining < 8) 
								? (uint8) usDescriptorRemaining : 8);
				// If ucNoOfBytes is between 1 and 8, copy bytes to FIFO
				if (ucEP0BytesToSend > 0 && ucEP0BytesToSend < 9)
				{
					// Copy ucEP0BytesToSend here as we need to use it later
					ucCount = ucEP0BytesToSend;
					while (ucCount != 0)
					{
						TXD0 = *pucDescriptorPointer++;
						ucCount--;
						usDescriptorRemaining--;
					}
				}
				// If ucNoOfBytes is zero, 
				// we will send a zero length packet by default

				// Set the TOGGLE bit correctly
				toggle_txep_pid(0,0);

				// Now enable the endpoint and watch that data go with 
				// the next IN token!
				TXC0 |= TX_EN;
			}
			else if (bZeroLengthPacket)
			{
				// Fill endpoint with a zero length packet, in case
				// we sent a modulo 8 packet previously and need to 
				// indicate the end of a transfer
				sendEP0data (0, 0, NULL);

				// Clear Zero Length flag
				bZeroLengthPacket = FALSE;
			}
			else
			{
				// Flush RX EP0 and enable RX EP0
				RXC0 |= RX_FLUSH;
				RXC0 |= RX_EN;
			}
		}
		else
		{
			// If ACK_STAT was not set, we'd better set the STALL bit
			// and get this error cleared
			stall_ep0();
		}
	}
}

/******************************************************************************  
    Function: copy_to_tx_buff

	Description
	-----------
    Copies data from users buffer into USB FIFO
    
	Arguments
	---------
	Both forms of endpoint index.
	
	Return value
	------------
	None.
*******************************************************************************/
static void copy_to_tx_buff( uint Endpoint, uint txep )
{
	uint orig_byte_cnt = usBuffRemainBytes[Endpoint];

	// Fill endpoint FIFO
	while (((TXS(txep) & TCOUNT) != 0) && (usBuffRemainBytes[Endpoint] != 0))
	{
		// Copy a byte of users data into FIFO, incrementing local pointer
		TXD(txep) = * pucCurrentBuffAddr[Endpoint]++;
			// Decrement number of bytes remaining in buffer
		usBuffRemainBytes[Endpoint]--;
	}

	ep_copy_cnt[Endpoint] = (orig_byte_cnt - usBuffRemainBytes[Endpoint]);
	
	// Set or clear TOGGLE as appropriate
	toggle_txep_pid( Endpoint, txep );
	
	// Now set up endpoint to transmit on next IN token
	TXC(txep) |= TX_LAST | TX_EN;
}

/******************************************************************************  
    Function: txep_handler

	Description
	-----------
    This function handles the TX EPn endpoints > 0

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void txep_handler (uint txep)
{
	/* Put TXS1 into RAM (clear on read register) */
	uint8 RamTXS = TXS(txep);
	uint ep = ep_from_txep(txep); // Which endpoint is this TX endpoint
	
	// Check TX_DONE is set
	if (RamTXS & TX_DONE)
	{
		// Check that last packet was ACK'ed
		// TODO: Add ISO support
		if (RamTXS & ACK_STAT)
		{
			// Now we know that the last run of data copied into FIFO has been
			// transferred properly and acked, add it to the total xfer count
			USBEPSupport[ep].xfer_cnt += ep_copy_cnt[ep];

			// Toggle PID
			ucEndpointPID ^= EP_BIT(ep);

			// If the last packet was completely full then another packet needs
			// to be sent even if its just a zero length one
			if (ep_copy_cnt[ep] == USB_MAX_PKT_LEN)
				copy_to_tx_buff( ep, txep );
			// if there was no data to send, then flag TRNSFRCOMPLETE
			else
			{
				USBEPSupport[ep].ucTransferStatus = TRNSFRCOMPLETE;
#ifndef USB_COMPACT
				if (USBInitInfo.evt_clbk)
					USBInitInfo.evt_clbk( USB_EP_EVT, (ep_id_t)ep );
#endif
			}
		}
		// If we didn't get an ACK, refill FIFO
		// TODO: ISO Support
		else
			// As long as tx'ing has not been stopped
			if (USBEPSupport[ep].ucTransferStatus == TRNSFRING)
			{
				// Repeat last transaction - Refill FIFO
				TXC(txep) |= RFF;
	
				// Now set up endpoint to transmit on next IN token
				TXC(txep) |= TX_LAST | TX_EN;
			}
	}
}

/******************************************************************************  
    Function: copy_from_rx_buffer

	Description
	-----------
    Copies data from USB rx FIFO into users buffer.
    
	Arguments
	---------
	Both forms of endpoint index.
	
	Return value
	------------
	None.
*******************************************************************************/
static void copy_from_rx_buffer( uint ep, uint rxep )
{
	uint orig_byte_cnt = usBuffRemainBytes[ep];
	uint curr_xfer;

	// Copy data from FIFO to user buffer, until endpoint is empty
	while (((RXS(rxep) & RCOUNT) != 0) && (usBuffRemainBytes[ep] != 0))
	{
		// Copy a byte of users data into FIFO, incrementing pointer
		*pucCurrentBuffAddr[ep] = RXD(rxep);
		pucCurrentBuffAddr[ep]++;

		usBuffRemainBytes[ep]--;
	}

	curr_xfer = (orig_byte_cnt - usBuffRemainBytes[ep]);

	// Note number of total bytes transferred
	USBEPSupport[ep].xfer_cnt += curr_xfer;

	// State that transfer is complete if app buffer is full or 
	// partial packet received
	if ((usBuffRemainBytes[ep] == 0) || (curr_xfer < USB_MAX_PKT_LEN))
	{
		USBEPSupport[ep].ucTransferStatus = TRNSFRCOMPLETE;
#ifndef USB_COMPACT
		if (USBInitInfo.evt_clbk)
			USBInitInfo.evt_clbk( USB_EP_EVT, (ep_id_t)ep );
#endif
	}
	else
	    // Get rest of transfer
	    RXC(rxep) |= RX_EN;
}

/******************************************************************************  
    Function: rxep_handler

	Description
	-----------
    This function handles the RX EPn endpoints > 0

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void rxep_handler (uint rxep)
{
	/* Put RXS1 into RAM (clear on read register) */
	uint8 RamRXS = RXS(rxep);
	uint ep = ep_from_rxep(rxep);
			
	// Check that packet was received without error
	if (RamRXS & RX_ERR)
	{
		// If packet contained error, flush ep
		RXC(rxep) |= RX_FLUSH;
		RXC(rxep) |= RX_EN;
	}
	// Otherwise check RX_DONE is set
	else if (RamRXS & RX_LAST)
	{
		// Check PID
		// TODO: Add ISO support
		if (RamRXS & RX_TOGL)
		{
			// If it matches our expected value, toggle our
			// internal value for next time
			if (ucEndpointPID & EP_BIT(ep))
			{
				ucEndpointPID &= ~EP_BIT(ep);
			}
			// Otherwise flush the EP ready for the next packet
			else
			{
				RXC(rxep) |= RX_FLUSH;
				RXC(rxep) |= RX_EN;
				return;
			}
		}
		// Else if the TOGGLE bit is not set
		else
		{
			// If it matches our expected value, toggle our
			// internal value for next time
			if (!(ucEndpointPID & EP_BIT(ep)))
			{
				ucEndpointPID |= EP_BIT(ep);
			}
			// Otherwise flush the EP ready for the next packet
			else
			{
				RXC(rxep) |= RX_FLUSH;
				RXC(rxep) |= RX_EN;
				return;
			}
		}

		// If we get here, there is valid data to unload into the
		// users buffer
		if (USBEPSupport[ep].ucTransferStatus == TRNSFRING)
			copy_from_rx_buffer( ep, rxep );
	}

}


/******************************************************************************  
    Function: warn_event

	Description
	-----------
    Handles warning events.
    ToDo, setup warn events and tidal marks etc...

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
static void warn_event (void)
{
}
#endif
/******************************************************************************  
    Function: alt_event

	Description
	-----------
    Handles alternate events

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void alt_event (void)
{
	uint8 RamALTEV = ALTEV;

	// Mask off bits that aren't enabled to cause interrupts
	RamALTEV &= ALTMSK;

	if (RamALTEV & EOP)
	{
		// If EOP signalling is received, and the EOP flag in the ALTMSK is
		// enable, then everything ok, re-enable all other interrupt bits again
		if (ALTMSK & EOP)
		{
			/* Set up alternate event mask */
			ALTMSK |= (SD3 | SD5 | RESET_A | RESUME_A);

			/* Set up main mask */
			MAMSK |= (WARN | ALT | TX_EV | NAK | ULD | RX_EV | INTR_E);
		}
	}
	else if (RamALTEV & SD3)	// 3ms of SUSPEND detected - suspend the endpoint
	{
#ifndef USB_COMPACT
		uint ep, bit_flag;

		// Store EP control registers for restoration after resume
		ucSavedEPCStatus[0] = EPC0;
		for( ep = 1; ep < N_ENDPOINTS; ep++ )
			ucSavedEPCStatus[ep] = EPC(ep);

		// Store R/TX_EN flags for restoration after resume
		ucSavedTXCStatus = 0;
		ucSavedRXCStatus = 0;
		bit_flag = 1;
		for( ep = 0; ep < 4; ep++ )
		{
			if (TXC(ep) & TX_EN)
				ucSavedTXCStatus |= bit_flag;
			if (RXC(ep) & RX_EN)
				ucSavedRXCStatus |= bit_flag;

			bit_flag <<= 1;

			if (TXC(ep) & TX_LAST)
				ucSavedTXCStatus |= bit_flag;

			bit_flag <<= 1;
		}
		ucSavedTXCStatus &= ~0x02; // Special case: Actually there is no 
		                           // TX_LAST for EP0 just ensure the value 
		                           // isn't restored on resume
		// Suspend the endpoint
		NFSR = SUS_ST;

		// Note that the node is suspended
		indicate_suspended( TRUE );

		// Mask off the SD3 bit, as we don't want to respond again to this
		ALTMSK &= ~SD3;

    	if (USBInitInfo.evt_clbk)
	    	USBInitInfo.evt_clbk( USB_SD3_SUSPEND, RXTXEP0 );
#endif
	}
	else if (RamALTEV & SD5) // 5ms SUSPEND detected, remote wake up now allowed
	{
		// This indicates that we can legally perform a remote wakeup
		set_remote_wake_ok( TRUE );

		// Mask off the SD5 bit, as we don't want to respond again to this
		ALTMSK &= ~SD5;
#ifndef USB_COMPACT
		// TODO: Users callback function to say this feature is available
    	if (USBInitInfo.evt_clbk)
	    	USBInitInfo.evt_clbk( USB_SD5_SUSPEND, RXTXEP0 );
#endif
	} 
	else if (RamALTEV & RESET_A)
	{
		// Set NodeFunctionalState Reg to RESET
		NFSR = RST_ST;

		// Set node address to 0x00, and clear address enable bit
		FARREG = 0x00;

		// Clear EP0 control reg
		EPC0 = 0x00;

		// Flush TX EP0
		TXC0 |= TX_FLUSH;

		// Flush RX EP0
		RXC0 |= RX_FLUSH;

		// Disable reception of IN tokens ?
//		RXC0 &= RX_EN;

		// Set internal state to DEFAULT
		set_device_state( DEFAULT );

		// Set address enable bit
		FARREG |= AD_EN;
	
		// Mask on the SD3 and SD5 bits again, so we respond to new SUSPEND
		// signalling
		ALTMSK |= (SD3 | SD5);

		// Set node functional state back to operational
		NFSR = OPR_ST;
	}
	else if (RamALTEV & RESUME_A) // Resume seen on bus
	{ // re-instate all the registers and change our state
		resume_endpoint();
	}
}

/******************************************************************************  
    Function: frame_event

	Description
	-----------
    Handles frame events

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/
#ifndef USB_COMPACT
static void frame_event (void)
{
   	if (USBInitInfo.evt_clbk)
    	USBInitInfo.evt_clbk( USB_FRAME_EVT, RXTXEP0 );

}
#endif
/******************************************************************************  
    Function: nak_event

	Description
	-----------
    Handles NAK events

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

static void nak_event (void)
{
	if (NAKEV & NAK_O0)
	{
		// Stop sending any descriptor data and flush EP
		usDescriptorRemaining = 0;
		TXC0 |= TX_FLUSH;

		// Now enable RX EP0
		RXC0 |= RX_EN;
	}
}

/******************************************************************************  
    Function: usb_isr

	Description
	-----------
    This function is the USB interrupt handler. It handles all the enabled 
	interrupts in the USB interrupt tree.

	Arguments
	---------
	None.
	
	Return value
	------------
	None.
*******************************************************************************/

#ifdef __NSC_CRCC__
#pragma interrupt(usb_isr)
void usb_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+USB_IRQ)
__interrupt void usb_isr(void)
#else
__interrupt[(16+USB_IRQ)*IDT_WIDTH] void usb_isr(void)
#endif
#endif
{	
	uint8 RamMAEV = MAEV;
	uint8 RamTXEV;
	uint8 RamRXEV;

	// Mask off bits that aren't enabled to cause interrupts
	RamMAEV &= MAMSK;

	// Check for Tx Events
	if (RamMAEV & TX_EV)
    {
		RamTXEV = TXEV & TXMSK;

        if (RamTXEV & TXFIFO0)
            txep0_handler();
		else if (RamTXEV & TXFIFO1)
			txep_handler(1);
		else if (RamTXEV & TXFIFO2)
			txep_handler(2);
		else if (RamTXEV & TXFIFO3)
			txep_handler(3);
	}
	else if (RamMAEV & RX_EV) // Check for Rx Events
	{
		RamRXEV = RXEV & RXMSK;

		if (RamRXEV & RXFIFO0)
			rxep0_handler();
		else if (RamRXEV & RXFIFO1)
			rxep_handler(1);
		else if (RamRXEV & RXFIFO2)
			rxep_handler(2);
		else if (RamRXEV & RXFIFO3)
			rxep_handler(3);
	}
    else if (RamMAEV & ALT)   // Alternate Event (includes Reset Event)
        alt_event();
#ifndef USB_COMPACT
    else if (RamMAEV & WARN)  // Warning Event
		warn_event();
    else if (RamMAEV & FRAME) // Frame Event
		frame_event();
#endif
    else if (RamMAEV & NAK)   // NAK events should be handled with lowest priority
        nak_event();
}
