/******************************************************************************
*  File:	vtu.c
*  
*  This is Versatile Timer Unit peripheral driver.
******************************************************************************/
#include <stdio.h>
#include "chipdef.h"

vtu_t * const vtu = (vtu_t *)VTU_BASE_ADDR;

/* Variables to store the value of the time interval in milli seconds*/
volatile long Duty;
volatile long Period;

/******************************************************************************  
    Function: vtu_init

	Description
	-----------
    This function initializes a single VTU channel. This function can also be called to 
    update a VTU configuration (ie. modulating the PWM values). 
    
    The following modes are supported:

    Mode 0 (low power):
      - all VTU timers are stopped
    Mode 1 (dual 8-bit PWM):
      - sets the duty and period registers for 2 8-bit PWM timers in a single VTU channel
    Mode 2 (single 16-bit PWM)
      - sets the duty and period registers for 1 16-bit PWM timer in a single VTU channel
      
	Arguments
	---------
  Channel:
    - selects the VTU channel being initialized
	Mode:
    - selects the mode of operation
  Count:
    - initializes the COUNT register in the VTU channel
  Duty:
	  Mode 1:
      - Duty[7:0] corresponds to the duty cycle value for the 8-bit PWM timer 1 in the VTU channel
      - Duty[15:8] corresponds to the duty cycle value for the 8-bit PWM timer 2 in the VTU channel
    Mode 2:
      - Duty[15:0] corresponds to the duty cycle value for the 16-bit PWM timer
	Period:
	  Mode 1:
      - Period[7:0] corresponds to the period value for the 8-bit PWM timer 1 in the VTU channel
      - Period[15:8] corresponds to the period value for the 8-bit PWM timer 2 in the VTU channel
    Mode 2:
      - Period[15:0] corresponds to the period value for the 16-bit PWM timer
      
*******************************************************************************/

void vtu_init(const int Channel, const int VTUmode, const int Count, const int Duty, const int Period, const int Prescaler)
{
  if (Channel == 1){
    if (VTUmode == 0){
      /* stop the VTU channel */
      MODE &= 0xfff3;/* clear bits [3:2]*/
    }
    if (VTUmode == 1){
      /* stop the VTU channel */
      MODE &= 0xfffc;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD1_DUAL_8PWM;
      /* Enable alternate function of the VTU interface pins. */
	    VTU_MUX_PORT->pxalt |= 0x03;     
      VTU_MUX_PORT->pxalts |= 0x03;    
      VTU_MUX_PORT->pxdir |= 0x03;
      /* initialize the COUNT register */
      COUNT1 = Count;
      /* initialize the DTYCAP register */
      DTYCAP1 = Duty;
      /* initialize the period */
      PERCAP1 = Period;
      /* initialize the clock prescaler */
      CLK1PS &= 0xFF00;
      CLK1PS |= (Prescaler & 0x000F);
      /* start the VTU channel */
      MODE |= T1RUN|T2RUN;
    }
    if (VTUmode == 2){
      /* stop the VTU channel */
      MODE &= 0xfffc;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD1_16PWM;
      VTU_MUX_PORT->pxalt &= 0xFC;    
      VTU_MUX_PORT->pxalt |= 0x01;    
      VTU_MUX_PORT->pxalts &= 0xFC;    
      VTU_MUX_PORT->pxalts |= 0x01;    
      VTU_MUX_PORT->pxdir &= 0xFC;
      VTU_MUX_PORT->pxdir |= 0x01;
      /* initialize the COUNT register */
      COUNT1 = Count;
      /* initialize the duty cycle */
      DTYCAP1 = Duty;
      /* initialize the period */
      PERCAP1 = Period;
      /* initialize the clock prescaler */
      CLK1PS &= 0xFF00;
      CLK1PS |= (Prescaler & 0x000F);
      /* start the VTU channel */
      MODE |= T1RUN;
    }
  }
  if (Channel == 2){
    if (VTUmode == 0){
      /* stop the VTU channel */
      MODE &= 0xff3f;/* clear bits [7:6]*/
    }
    if (VTUmode == 1){
      /* stop the VTU channel */
      MODE &= 0xffcf;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD2_DUAL_8PWM;
      VTU_MUX_PORT->pxalt |= 0x0C;    
      VTU_MUX_PORT->pxalts |= 0x0C;    
      VTU_MUX_PORT->pxdir |= 0x0C;
      /* initialize the COUNT register */
      COUNT2 = Count;
      /* initialize the duty cycle */
      DTYCAP2 = Duty;
      /* initialize the period */
      PERCAP2 = Period;
      /* initialize the clock prescaler */
      CLK1PS &= 0x00FF;
      CLK1PS |= (Prescaler & 0x000F)<<4;
      /* start the VTU channel */
      MODE |= T3RUN|T4RUN;
    }
    if (VTUmode == 2){
      /* stop the VTU channel */
      MODE &= 0xffcf;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD2_16PWM;
      VTU_MUX_PORT->pxalt &= 0xF3;    
      VTU_MUX_PORT->pxalt |= 0x04;    
      VTU_MUX_PORT->pxalts &= 0xF3;    
      VTU_MUX_PORT->pxalts |= 0x04;    
      VTU_MUX_PORT->pxdir &= 0xF3;
      VTU_MUX_PORT->pxdir |= 0x04;
      /* initialize the COUNT register */
      COUNT2 = Count;
      /* initialize the duty cycle */
      DTYCAP2 = Duty;
      /* initialize the period */
      PERCAP2 = Period;
      /* initialize the clock prescaler */
      CLK1PS &= 0x00FF;
      CLK1PS |= (Prescaler & 0x000F)<<4;
      /* start the VTU channel */
      MODE |= T3RUN;
    }
  }
  if (Channel == 3){
    if (VTUmode == 0){
      MODE &= 0xf3ff;/* clear bits [11:10]*/
    }
    if (VTUmode == 1){
      /* stop the VTU channel */
      MODE &= 0xfcff;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD3_DUAL_8PWM;
      VTU_MUX_PORT->pxalt |= 0x30;    
      VTU_MUX_PORT->pxalts |= 0x30;    
      VTU_MUX_PORT->pxdir |= 0x30;
      /* initialize the COUNT register */
      COUNT3 = Count;
      /* initialize the duty cycle */
      DTYCAP3 = Duty;
      /* initialize the period */
      PERCAP3 = Period;
      /* initialize the clock prescaler */
      CLK2PS &= 0xFF00;
      CLK2PS |= (Prescaler & 0x000F);
      /* start the VTU channel */
      MODE |= T5RUN|T6RUN;
    }
    if (VTUmode == 2){
      /* stop the VTU channel */
      MODE &= 0xfcff;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD3_16PWM;
      VTU_MUX_PORT->pxalt &= 0xCF;    
      VTU_MUX_PORT->pxalt |= 0x10;    
      VTU_MUX_PORT->pxalts &= 0xCF;    
      VTU_MUX_PORT->pxalts |= 0x10;    
      VTU_MUX_PORT->pxdir &= 0xCF;
      VTU_MUX_PORT->pxdir |= 0x10;
      /* initialize the COUNT register */
      COUNT3 = Count;
      /* initialize the duty cycle */
      DTYCAP3 = Duty;
      /* initialize the period */
      PERCAP3 = Period;
      /* initialize the clock prescaler */
      CLK2PS &= 0xFF00;
      CLK2PS |= (Prescaler & 0x000F);
      /* start the VTU channel */
      MODE |= T5RUN;
    }
  }
  if (Channel == 4){
    if (VTUmode == 0){
      MODE &= 0x3fff;/* clear bits [15:14]*/
    }
    if (VTUmode == 1){
      /* stop the VTU channel */
      MODE &= 0xcfff;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD4_DUAL_8PWM;
      VTU_MUX_PORT->pxalt |= 0xC0;    
      VTU_MUX_PORT->pxalts |= 0xC0;    
      VTU_MUX_PORT->pxdir |= 0xC0;
      /* initialize the COUNT register */
      COUNT4 = Count;
      /* initialize the duty cycle */
      DTYCAP4 = Duty;
      /* initialize the period */
      PERCAP4 = Period;
      /* initialize the clock prescaler */
      CLK2PS &= 0x00FF;
      CLK2PS |= (Prescaler & 0x000F)<<4;
      /* start the VTU channel */
      MODE |= T7RUN|T8RUN;
    }
    if (VTUmode == 2){
      /* stop the VTU channel */
      MODE &= 0xcfff;
      /* enable the VTU channel (if it hasn't been enabled already*/
      MODE |= TMOD4_16PWM;
      VTU_MUX_PORT->pxalt &= 0x3F;    
      VTU_MUX_PORT->pxalt |= 0x40;    
      VTU_MUX_PORT->pxalts &= 0x3F;    
      VTU_MUX_PORT->pxalts |= 0x40;    
      VTU_MUX_PORT->pxdir &= 0x3F;
      VTU_MUX_PORT->pxdir |= 0x40;
      /* initialize the COUNT register */
      COUNT4 = Count;
      /* initialize the duty cycle */
      DTYCAP4 = Duty;
      /* initialize the period */
      PERCAP4 = Period;
      /* initialize the clock prescaler */
      CLK2PS &= 0x00FF;
      CLK2PS |= (Prescaler & 0x000F)<<4;
      /* start the VTU channel */
      MODE |= T7RUN;
    }
  }
}

/*Interrupt Handler*/
#ifdef __CR__
#pragma interrupt (vtua_isr)
void vtua_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+VTUA_IRQ)
__interrupt void vtua_isr(void)
#else
__interrupt[(16+VTUA_IRQ)*IDT_WIDTH] void vtua_isr(void)
#endif
#endif
{
  /*clear the source of the interrupt*/
//  INTPND |= I1APD; /* uncomment the appropriate line */
//  INTPND |= I1BPD;
//  INTPND |= I1CPD;
//  INTPND |= I1DPD;
}

#ifdef __CR__
#pragma interrupt (vtub_isr)
void vtub_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+VTUB_IRQ)
__interrupt void vtub_isr(void)
#else
__interrupt[(16+VTUB_IRQ)*IDT_WIDTH] void vtub_isr(void)
#endif
#endif
{
  /*clear the source of the interrupt*/
//  INTPND |= I2APD; /* uncomment the appropriate line */
//  INTPND |= I2BPD;
//  INTPND |= I2CPD;
//  INTPND |= I2DPD;
}

#ifdef __CR__
#pragma interrupt (vtuc_isr)
void vtuc_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+VTUC_IRQ)
__interrupt void vtuc_isr(void)
#else
__interrupt[(16+VTUC_IRQ)*IDT_WIDTH] void vtuc_isr(void)
#endif
#endif
{
  /*clear the source of the interrupt*/
//  INTPND |= I3APD; /* uncomment the appropriate line */
//  INTPND |= I3BPD;
//  INTPND |= I3CPD;
//  INTPND |= I3DPD;
}

#ifdef __CR__
#pragma interrupt (vtud_isr)
void vtud_isr(void)
#endif
#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#pragma vector=(16+VTUD_IRQ)
__interrupt void vtud_isr(void)
#else
__interrupt[(16+VTUD_IRQ)*IDT_WIDTH] void vtud_isr(void)
#endif
#endif
{
  /*clear the source of the interrupt*/
//  INTPND |= I4APD; /* uncomment the appropriate line */
//  INTPND |= I4BPD;
//  INTPND |= I4CPD;
//  INTPND |= I4DPD;
}


