/******************************************************************************
*  File:	CommIntp.c
*  
*  Command interpreter module for the CPDemo.
*  The ProcessHostCommands receives the data packet in a COMMBLOCK_CP 
*  processes it accordingly and sends the appropriate response back to
*  the host.
*
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>		  
#include <string.h>
#include <ctype.h>
#include "crcommon.h"
#include "config.h"
#include "boarddrv.h"

#include "protocol.h"   
#include "commintp.h"

COMMBLOCK Response;
static unsigned char DataBlock[256];

static unsigned char DevProp[]=
{
	CHIP_FAMILY,
	DEV_BOARD,
	NVS_MODE,
	LAST_PROPERTY
};

/******************************************************************************  
*    Function: ProcessHostCommands
*
*	Description
*	-----------
*		This function is called when a new command arrives from the Host.
*		The command specified in the COMMBLOCK_CP structure is processed and
*		A response is sent back to the host.
*
******************************************************************************/
void ProcessHostCommands (COMMBLOCK *pHostInput)
{
	unsigned short ByteOffset;
	int i, iCount;

	//
	// Check if the block is consistent
	// nParams has the number of parameters sent, 
	// so, Param[nParams] must be EOB for this block to be consistent.
	// 
	if (pHostInput->Param[pHostInput->nParams] == EOB)
	{
		switch (pHostInput->CommandID)
		{
			case GET_DEV_PROP:
			{ 
				for (i=0; DevProp[i]!=LAST_PROPERTY; ++i)
				{
					Response.Param[i+1] = DevProp[i];
				}
				
				SendResponse (pHostInput, LAST_PROP);        
				break;     
			}

			case LED_SET_ALL:
			{
				led_set_all(pHostInput->Param[0]);
				SendResponse (pHostInput, NO_PARAMS);
				break;        
			}

			case EEPROM_READ:
			{
				ByteOffset = pHostInput->Param[1] | (((unsigned short) pHostInput->Param[2]) << 8);

				eep_buf_read(ByteOffset, DataBlock, pHostInput->Param[0]);

				for (i=0; i<pHostInput->Param[0]; i++)
				   Response.Param[i+1] = DataBlock[i];

				SendResponse (pHostInput, pHostInput->Param[0]);
				break;
			}
        
			case EEPROM_WRITE:
			{
				ByteOffset = pHostInput->Param[1] | (((unsigned short) pHostInput->Param[2]) << 8);

				for (i=0; i<pHostInput->Param[0]; ++i)
				DataBlock[i] = pHostInput->Param[i+3];

				eep_buf_write(ByteOffset, DataBlock, pHostInput->Param[0]);

				SendResponse (pHostInput, NO_PARAMS);
				break;      
			}
        
			case SWITCH_ON:
			{
				if ( switch_on((int) pHostInput->Param[0]) )
				   Response.Param[1] = 1;
				else
				   Response.Param[1]= 0;
   
				SendResponse (pHostInput, 1);
				break;
			}

			case START_RECORD:
			{
				Record();
				SendResponse (pHostInput, NO_PARAMS);        
				break;
			}
        
			case START_PLAYBACK:
			{
				Playback();
				SendResponse (pHostInput, NO_PARAMS);        
				break;
			}

			case PLAY_TONE:
			{
				TonePlay ((unsigned int) pHostInput->Param[0], 40, NULL);
				SendResponse (pHostInput, NO_PARAMS);        
				break;
			}

			case SET_VOLUME:
			{
				codec_set_volume ((unsigned int) pHostInput->Param[0]);
				SendResponse (pHostInput, NO_PARAMS);        
				break;
			}

			case START_COM:
			{
				SendResponse (pHostInput, NO_PARAMS);        
				break;
			}
            
			case STOP_COM:
				break;
      
			default:
			{
				//
				//  Purge rx buffer.
				//
				do
				{
				  iCount = usart_dma_rx_count();
				} while (iCount!=0);

				SendNACK (pHostInput);
				break;
			}
		}
	}
	else
	{  
		//
		//  Purge rx buffer.
		//
		do
		{
		  iCount = usart_dma_rx_count();
		} while (iCount!=0);
    
		SendNACK (pHostInput);
	}
}

/******************************************************************************  
*
*    Function: SendResponse
*
*	Description
*	-----------
*		Sends the response back to the host.
*
*	Parameter
*	---------
*      
*       unsigned char nParam - Number of parameters to be sent back.
* 
*******************************************************************************/ 
void SendResponse (COMMBLOCK *pHostInput, unsigned char nParams)
{
  Response.CommandID = pHostInput->CommandID; 
  
  Response.nParams = nParams+1;    //   nParams + ACK
  Response.Param[0] = SER_ACK;         
  Response.Param[Response.nParams] = EOB;
  usart_dma_tx ((unsigned char *) &Response, Response.nParams + COMMAND_SIZE);
}

/******************************************************************************  
*
*    Function: SendResponse
*
*	Description
*	-----------
*		Sends a no-acknoledge message back to the host.
*
*******************************************************************************/  
void SendNACK (COMMBLOCK *pHostInput)
{
	Response.CommandID                    = pHostInput->CommandID;
	Response.nParams                      = 1;            // Size
	Response.Param[0]                     = SER_NACK;        
	Response.Param[Response.nParams + 1]  = EOB;

	usart_dma_tx ((unsigned char *) &Response, Response.nParams + COMMAND_SIZE);
}
