#ifndef DL_H
#define DL_H

/************************************************************************
* Data Link
*
* Data Link is a simple concept that used to synchronize to software
* entities called Producer and Consumer. A producer continuously generates
* data that is later read by the consumer. They are not synchronized 
* except through the data link mechanism. The core of the data link
* is a circular buffer that is used for the data transfer and data link
* counter which tells how many bytes have been produced by the producer
* and not consumed by the consumer. This counter can be anywhere between
* zero and the size of the circular buffer in bytes. The consumer and
* the producer each maintain a pointer to the circular buffer. The producer
* pointer points to the location in the buffer to which the producer 
* should write the next byte. The consumer pointer point to the location
* in the buffer from which the consumer will read the next byte. The 
* producer can write to the data link buffer as long as the buffer is
* not full (count equal to buffer size). The consumer can read from the
* data link buffer as long as the buffer is not empty (count equal to zero).

* The data link structure (type: dl_t) contains the information that's
* required for the producer and consumer to synchronize
************************************************************************/

#include <stdio.h>
#ifdef __NSC_CRCC__
#include <asm.h>
#define __disable_interrupt() _di_()
#define __enable_interrupt() _ei_()
#endif

#ifdef __IAR_SYSTEMS_ICC__
#ifdef __ICCCR16C__
#include <intrinsics.h>
#else
#include <incr16.h>
#endif
#endif


#ifdef __CR16C__
#  define PTR_CAST unsigned long
#else
#  define PTR_CAST unsigned short
#endif

/* Data Link structure */
typedef struct {
  unsigned char *buf_addr; /* Circular buffer start address */
  unsigned char *buf_end;  /* Next memory address after buffer last address */
  unsigned int buf_size;   /* Circular buffer size */
  volatile unsigned int count; /* Number of produced bytes that were not consumed yet */
  void (*p_action)(void); /* Producer action after commit - if required */
  void (*c_action)(void); /* Consumer action after commit - if required */
} dl_t;


/* dl_init
  Initializes a data link structure given a circular buffer address and
  size in bytes.
*/
#define dl_init(dl, bufaddr, bufsize) \
  dl->buf_addr = bufaddr; \
  dl->buf_size = bufsize;  \
  dl->buf_end = dl->buf_addr + dl->buf_size; \
  dl->count = 0; \
  dl->p_action = dl->c_action = NULL

/* dl_init_ptr
   Initializes a given pointer (ptr) to a start address of a data link
   buffer. 
*/
#define dl_init_ptr(dl, ptr) ptr = dl->buf_addr

/* dl_set_buf_size
   Modifies the data link buffer size (in some cases we need to do it
   after the data link has been initialized.
*/
#define dl_set_buf_size(dl, bufsize) dl->buf_size = bufsize; \
		dl->buf_end = dl->buf_addr + dl->buf_size

/* dl_get_buf
   provides the data link buffer start address
*/
#define dl_get_buf(dl) (dl->buf_addr)

/* dl_get_buf_size
   provides the data link buffer size
*/
#define dl_get_buf_size(dl) (dl->buf_size)

/* dl_cnt
   provides the current number of produced and unconsumed bytes
*/
#define dl_cnt(dl) (dl->count)

/* dl_advance
   Advances a pointer by one position in the circular buffer. Performs
   a wrap around if required.
*/
#define dl_advance(dl, ptr) ptr++; if (ptr==dl->buf_end) ptr=dl->buf_addr

/* dl_advanc_multiple
   Advances a pointer by multiple positions in the circular buffer. Performs
   a wrap around if required.
*/
#define dl_advance_multiple(dl, ptr, size) \
        if ((PTR_CAST)dl->buf_end-(PTR_CAST)ptr <= size) \
            ptr-=(dl->buf_size-size);\
            	else\
			ptr+=size;

/* dl_regress
   Regresses a pointer by one position in the circular buffer. Performs
   a wrap around if required.
*/
#define dl_regress(dl, ptr) if (ptr==dl->buf_addr) ptr=dl->buf_end; ptr--

/* dl_p_commit
   Producer commit. The prducer releases a given number of bytes to the
   consumer. The bytes have already been written and now they are officially
   released by incrementing the data link counter. A producer action is
   taken if one exists.
*/
#define dl_p_commit(dl, size) __disable_interrupt(); dl->count+=size; __enable_interrupt(); \
        if (dl->p_action != NULL) (*(dl->p_action))()

/* dl_c_commit
   Consumer commit. The consumer reports the consumption of a certain amount
   of bytes. The bytes have already been read and now they are officially
   declared consumed by decrementing the data link counter. The memory space
   previously occupied by these bytes is now available to the producer. A
   consumer action is taken if one exists.
*/
#define dl_c_commit(dl, size) __disable_interrupt(); dl->count-=size; __enable_interrupt(); \
        if (dl->c_action != NULL) (*(dl->c_action))()

/* dl_p_advance_commit
   Producer pointer advance and commit. A combination of two actions already
   described above.
*/
#define dl_p_advance_commit(dl, ptr, size) dl_advance_multiple(dl, ptr,size); dl_p_commit(dl,size)


/* dl_c_advance_commit
   Consumer pointer advance and commit. A combination of two actions already
   described above.
*/
#define dl_c_advance_commit(dl, ptr, size) dl_advance_multiple(dl, ptr,size); dl_c_commit(dl,size)

/* dl_p_set_action
   Set producer action function. This is a pointer to a function that will
   be called upon a producer commit.
*/
#define dl_p_set_action(dl, action) dl->p_action = action

/* dl_c_set_action
   Set consumer action function. This is a pointer to a function that will
   be called upon a consumer commit.
*/
#define dl_c_set_action(dl, action) dl->c_action = action

/* dl_empty
   A predicate telling whether the data link buffer is empty or not
*/
#define dl_empty(dl) (dl->count == 0)

/* dl_full
   A predicate telling whether the data link buffer is full or not
*/
#define dl_full(dl) (dl->count == dl->buf_size)

/* dl_p_space_available
   A predicate telling the producer how much space is available in the
   buffer for writing new data.
*/
#define dl_p_space_available(dl) (dl->buf_size - dl->count)

/* dl_get_max_linear_data_size
   This macro returns the maximal size of data in the data link buffer
   that can be read without a wrap around from a given a pointer in the
   buffer.
*/
#define dl_get_max_linear_data_size(dl, ptr) \
        ((PTR_CAST)(dl->buf_end) - (PTR_CAST)ptr < dl->count ?\
		(PTR_CAST)(dl->buf_end) - (PTR_CAST)ptr:\
		dl->count)


/* dl_reset_count
  Re-initializes a data link structure, by clearing the count.
*/
#define dl_reset_count(dl) (dl->count = 0)

#endif /* DL_H */

