
/************************************************************************************************************
*                                               uC/OS-II
*                                         CR16C-Specific Kernel macros
* File : CR16C_C.H
* By   : Jeffrey Wright
*************************************************************************************************************/
#ifndef CR16CCH
#define CR16CCH

#ifdef  OS_CPU_GLOBALS
#define OS_CPU_EXT
#else
#define OS_CPU_EXT  extern
#endif
                                                    /* Task's Stack frames may come in these flavors...     */
#define FRAME_INITIAL   0                           /* Initial frame format                                 */
#define FRAME_TASK      1                           /* After task calls OS service                          */
#define FRAME_INT       2                           /* Task is interrupted                                  */

#ifdef __ICCCR16C__
#include  <intrinsics.h>
#else
#include <asm.h>

#endif

/*
*********************************************************************************************************
*                                          FUNCTION PROTOTYPES
*                                  (Compiler Specific ISR prototypes)
*********************************************************************************************************
*/

#ifdef __ICCCR16C__

//#pragma vector=5
__trap void  OSCtxSw(void);

#else

void  OSCtxSw(void);

#endif

void    OSStartHighRdy();
void    OSIntCtxSw();

/***********************************************************************************************************
*                                              DATA TYPES
*                                         (Compiler Specific)
************************************************************************************************************/

typedef unsigned char  BOOLEAN;
typedef unsigned char  INT8U;                       /* Unsigned  8 bit quantity                             */
typedef signed   char  INT8S;                       /* Signed    8 bit quantity                             */
typedef unsigned short INT16U;                      /* Unsigned 16 bit quantity                             */
typedef signed short   INT16S;                      /* Signed   16 bit quantity                             */
typedef unsigned long  INT32U;                      /* Unsigned 32 bit quantity                             */
typedef signed   long  INT32S;                      /* Signed   32 bit quantity                             */
typedef float          FP32;                        /* Single precision floating point                      */
typedef double         FP64;                        /* Double precision floating point                      */

/***********************************************************************************************************
*                               Legacy DATA TYPES
*                           (Included for legacy code)
************************************************************************************************************/
#define BYTE        INT8S                           /* Define data types for backward compatibility ...     */
#define UBYTE       INT8U                           /* ... to uC/OS V1.xx.  Not actually needed for ...     */
#define WORD        INT16S                          /* ... uC/OS-II.                                        */
#define UWORD       INT16U
#define LONG        INT32S
#define ULONG       INT32U
#define UCHAR       char
typedef volatile INT8U  VUBYTE;                     /* Volatile Unsigned  8 bit quantity                    */
typedef volatile INT16U VUWORD;                     /* Volatile Unsigned 16 bit quantity                    */
typedef volatile INT32U VULONG;                     /* Volatile Unsigned 32 bit quantity                    */

#define OS_STK_GROWTH   1                           /* Stack grows from HIGH to LOW memory on CR16B         */

typedef INT16U          OS_STK;                     /* Each stack entry is 16-bit wide                      */

OS_CPU_EXT  ULONG   OSStkPtr;                       /* Used in SUPERVISOR MODE on the CR16C                 */

/***********************************************************************************************************
*                                uC/OS-II allows two methods of Disabling/Enabling interrupts:
*
* Method #1:  Disable/Enable interrupts using simple instructions.  After critical section, interrupts
*             will be enabled even if they were disabled before entering the critical section.  You MUST
*             change the constant in OS_CPU_A.ASM, function OSIntCtxSw() from 10 to 8.
*
* Method #2:  Disable/Enable interrupts by preserving the state of interrupts.  In other words, if 
*             interrupts were disabled before entering the critical section, they will be disabled when
*             leaving the critical section.  You MUST change the constant in OS_CPU_A.ASM, function 
*             OSIntCtxSw() from 8 to 10.
*************************************************************************************************************/

#ifdef __ICCCR16C__                                 /* IAR                                                  */

    #define OS_ENTER_CRITICAL()     __disable_interrupt()
    #define OS_EXIT_CRITICAL()      __enable_interrupt()
    #define ENABLE_INTERRUPTS()     __set_PSR_I_bit()
    #define DISABLE_INTERRUPTS()    __clear_PSR_I_bit()
    #define OS_TASK_SW()            __raise_exception(5)  // This uses the svc trap as a system call
    #define __ASS__                 asm

#else                                               /* NSC                                                  */

    #define OS_ENTER_CRITICAL()     _di_()          /* Disable interrupts                                   */
    #define OS_EXIT_CRITICAL()      _ei_()          /* Enable  interrupts                                   */
    #define ENABLE_INTERRUPTS()     set_i_bit()
    #define DISABLE_INTERRUPTS()    clear_i_bit()
    #define OS_TASK_SW()            _excp_(5)       /* This uses the svc trap as a system call              */
    #define __ASS__                 __asm__

#endif

/*
*************************************************************************************************************
*                               CONTEXT SAVE/RESTORE MACROS
*   The following macros allow use of either USER or SUPERVISOR MODEs of operations
*************************************************************************************************************
*/

#ifdef USER_MODE
#define CTX_SAVE_METHOD_STORM
#else

/****   SUPERVISOR MODE USES PUSH INSTRUCTIONS TO SAVE CONTEXT  ****/

#define CTX_SAVE_METHOD_PUSH

    #ifdef __ICCCR16C__

        #define  USE_SYSTEM_STK()                   \
                                                    \
        OSTCBCur    =   OSTCBCur;                   \
        OSStkPtr    =   OSStkPtr;                   \
        do                                          \
        {                                           \
            __ASS__ ("loadd   OSTCBCur,(r1,r0)");   \
            __ASS__ ("stord   (sp),0(r1,r0)");      \
            __ASS__ ("loadd   OSStkPtr,(sp)");      \
        }                                           \
        while(0);

        #define  USE_TASK_STK()                     \
                                                    \
        OSTCBCur    =   OSTCBCur;                   \
        OSStkPtr    =   OSStkPtr;                   \
                                                    \
        do                                          \
        {                                           \
            __ASS__ ("stord   (sp),OSStkPtr");      \
            __ASS__ ("loadd   OSTCBCur,(r1,r0)");   \
            __ASS__ ("loadd   0(r1,r0),(sp)");      \
        }                                           \
        while(0);

    #else

        #define  USE_SYSTEM_STK()                   \
                                                    \
        do                                          \
        {                                           \
            __ASS__ ("loadd   _OSTCBCur,(r1,r0)");  \
            __ASS__ ("stord   (sp),0(r1,r0)");      \
            __ASS__ ("loadd   _OSStkPtr,(sp)");     \
        }                                           \
        while(0);

        #define  USE_TASK_STK()                     \
                                                    \
        do                                          \
        {                                           \
            __ASS__ ("stord   (sp),_OSStkPtr");   \
            __ASS__ ("loadd   _OSTCBCur,(r1,r0)");\
            __ASS__ ("loadd   0(r1,r0),(sp)");      \
        }                                           \
        while(0);

    #endif

#endif

/************************************************************************************************************
;
;   The SAVE/RESTORE_CTX macros define two methods for saving user's context.  The first method is used
;   during USER mode and utilizes a storm/loadm sequence. It avoids having to switch back to USER mode that
;   utilizing push/pop to save context would requre. The second method is used in SUPERVISOR (default) mode
;   and utilizes a push/pop sequence which are faster
;
;*************************************************************************************************************/

#define PUSH_SAFE()         do                              \
                            {                               \
                                __ASS__ ("push  $1,r7");    \
                                __ASS__ ("push  $8,r8");    \
                            }                               \
                            while(0);

#define POP_SAFE()          do                              \
                            {                               \
                                __ASS__ ("pop   $8,r8");    \
                                __ASS__ ("pop   $1,r7");    \
                            }                               \
                            while(0);

#ifdef USER_MODE

    #define  SAVE_CTX()                                                 \
                                                                        \
        do                                                              \
        {                                                               \
/* 4  Save user's R6,7          */  __ASS__ ("push  $2,r6");            \
/* 1  Get user's TOS            */  __ASS__ ("sprd  usp,(r7,r6)");      \
/* 1  Adjust ptr to new TOS     */  __ASS__ ("addd  $-36,(r7,r6)");     \
/* 3  Load new user TOS         */  __ASS__ ("lprd  (r7,r6),usp");      \
/* 9  Save R2,3,4,5,8,9,10,11   */  __ASS__ ("stormp    $8");           \
/* 1  Copy R12,13,14,0,1        */  __ASS__ ("movd  (r12),(r3,r2)");    \
/* 1  to permit storing...      */  __ASS__ ("movd  (r13),(r5,r4)");    \
/* 1                            */  __ASS__ ("movd  (ra),(r9,r8)");     \
/* 1                            */  __ASS__ ("movd  (r1,r0),(r11,r10)");\
/* 9  Save R12,13,14,0,1        */  __ASS__ ("stormp    $8");           \
/* 4  Restore user's R6,7       */  __ASS__ ("pop   $2,r2");            \
/* 3  Save R6,7                 */  __ASS__ ("stormp    $2");           \
        }                                                               \
        while(0);
/* 38 cycles                    */

    #define  RESTORE_CTX()                                              \
                                                                        \
        do                                                              \
        {                                                               \
/* 1  Get user's current TOS    */  __ASS__ ("sprd  usp,(r1,r0)");      \
/* 1  Adjust ptr for R6,7       */  __ASS__ ("addd  $32,(r1,r0)");      \
/* 4  Load R6 and R7            */  __ASS__ ("loadmp    $2");           \
/* 1  Copy R6,R7                */  __ASS__ ("movd  (r3,r2),(r7,r6)");  \
/* 3  Load new user TOS         */  __ASS__ ("lprd  (r1,r0),usp");      \
/* 1  Adjust ptr for nxt grp    */  __ASS__ ("addd  $-20,(r1,r0)");     \
/* 10 Load R12,13,14,0,1        */  __ASS__ ("loadmp    $8");           \
/* 1  Copy remaining regs to    */  __ASS__ ("movd  (r3,r2),(r12)");    \
/* 1  default for saving        */  __ASS__ ("movd  (r5,r4),(r13)");    \
/* 1                            */  __ASS__ ("movd  (r9,r8),(ra)");     \
/* 3  Save R1,R0                */  __ASS__ ("push  $2,r10");           \
/* 1  Adjust ptr for nxt grp    */  __ASS__ ("addd  $-32,(r1,r0)");     \
/* 10 Load R2,3,4,5,8,9,10,11   */  __ASS__ ("loadmp    $8");           \
/* 4  Get R1,R0                 */  __ASS__ ("pop   $2,r0");            \
        }                                                               \
        while(0);
/* 42 cycles                    */

#else                           /* CTX_SAVE_METHOD_PUSH                 */

    #define  SAVE_CTX()     PUSH_SAFE()
    #define  RESTORE_CTX()  POP_SAFE()

#endif

/************************************************************************************************************/

#ifdef __ICCCR16C__

    #define  INC_NEST_CTR()                     \
                                                \
    do                                          \
    {                                           \
        __ASS__ ("loadb OSIntNesting,r0");    \
        __ASS__ ("addb  $1,r0");                \
        __ASS__ ("storb r0,OSIntNesting");    \
    }                                           \
    while(0);

#else
    
    #define  INC_NEST_CTR()                     \
                                                \
    do                                          \
    {                                           \
        __ASS__ ("loadb _OSIntNesting,r0");   \
        __ASS__ ("addb  $1,r0");                \
        __ASS__ ("storb r0,_OSIntNesting");   \
    }                                           \
    while(0);

#endif

#ifdef OS_NEST_EN
            
    #ifdef __ICCCR16C__

        #ifdef USER_MODE

            #define SAVE_CONTEXT()                                      \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Used by INC_NEST_CTR  */  __ASS__ ("push  $1,r0");                \
/* 5  Increment nest counter*/  INC_NEST_CTR()                          \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $1,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .+36");                 \
/* 3                        */  __ASS__ ("pop   $1,r0");                \
/* 37 Save context...       */  SAVE_CTX()                              \
/* 1  Skip the redundant POP*/  __ASS__ ("br    .+4");                  \
/* 3  EXIT                  */  __ASS__ ("pop   $1,r0");                \
                }                                                       \
                while(0);
/*------------------------------*/
/* 52 cycles                    */

            #define RESTORE_CONTEXT()                                   \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Load nesting counter  */  __ASS__ ("loadb OSIntNesting,r0");    \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $0,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .+40");                 \
/*                          */  RESTORE_CTX()                           \
                }                                                       \
                while(0);

        #else

            #define SAVE_CONTEXT()                                      \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Used by INC_NEST_CTR  */  __ASS__ ("push  $1,r0");                \
/* 5  Increment nest counter*/  INC_NEST_CTR()                          \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $1,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .+10");                 \
/* 3                        */  __ASS__ ("pop   $1,r0");                \
/* 37 Save context...       */  PUSH_SAFE()                             \
/* 1  Skip the redundant POP*/  __ASS__ ("br    .+4");                  \
/* 3  EXIT                  */  __ASS__ ("pop   $1,r0");                \
                }                                                       \
                while(0);
/*------------------------------*/
/* 52 cycles                    */

            #define RESTORE_CONTEXT()                                   \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Load nesting counter  */  __ASS__ ("loadb OSIntNesting,r0");    \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $0,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .+4");                  \
/*                          */  POP_SAFE()                              \
                }                                                       \
                while(0);

        #endif

    #else   //NSC

        #define SAVE_CONTEXT()                                          \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Used by INC_NEST_CTR  */  __ASS__ ("push  $1,r0");                \
/* 5  Increment nest counter*/  INC_NEST_CTR()                          \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $1,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .SCTX1");               \
/* 3                        */  __ASS__ ("pop   $1,r0");                \
/* 37 Save context...       */  SAVE_CTX()                              \
/* 1  Skip the redundant POP*/  __ASS__ ("br    .+4");                  \
/*                          */  __ASS__ (".SCTX1:");                    \
/* 3  EXIT                  */  __ASS__ ("pop   $1,r0");                \
                }                                                       \
                while(0);
/*------------------------------*/
/* 52 cycles                    */

        #define RESTORE_CONTEXT()                                       \
                                                                        \
                do                                                      \
                {                                                       \
/* 2  Load nesting counter  */  __ASS__ ("loadb _OSIntNesting,r0");   \
/* 1  Nested interrupt??    */  __ASS__ ("cmpb  $0,r0");                \
/* 1  YES - goto EXIT       */  __ASS__ ("bne   .RCTX1");               \
/*                          */  RESTORE_CTX()                           \
/*                          */  __ASS__ (".RCTX1:");                    \
                }                                                       \
                while(0);

    #endif

#else   // No nesting...

    #define SAVE_CONTEXT()                                              \
                                                                        \
            do                                                          \
            {                                                           \
/* 37 Save context...           */  SAVE_CTX()                          \
/* 5  Increment nest counter    */  ++OSIntNesting;                     \
            }                                                           \
            while(0);
/*------------------------------*/
/* 52 cycles                    */

        #define RESTORE_CONTEXT()                                       \
            do                                                          \
            {                                                           \
                                    RESTORE_CTX()                       \
            }                                                           \
            while(0);
/*------------------------------*/
/* 52 cycles                    */


#endif
#endif

