/******************************************************************************
*  File:	tone.c
*  
*  This is a tone generation utility module.
******************************************************************************/

#include "boarddrv.h"
#include "tone.h"

#define BUF_SIZE 64

/* The following structure type includes a pointer to a tone sample data
   and and a number that indicates the number of samples in each tone cycle.
*/
typedef struct {
	const short * CycleData;
	unsigned short CycleSize;
} ToneCycleType;


static void ToneEnd(void);
static void ToneFillBuf(void);

static dl_t mem2audio_dl_s;
static dl_t *mem2audio_dl;
static unsigned char mem2audio_buf[BUF_SIZE]; 
static unsigned char *mem2audio_ptr;
static const ToneCycleType *ToneP;
static void (*ToneCBP)(void);


/* Each of the following arrays represents one tone. */
const short Tone0[] = { 4095, -4095 };
const short Tone1[] = { 3547, 0, -3547 };
const short Tone2[] = { 2896, 2896, -2896, -2896 };
const short Tone3[] = { 2407, 3895, 0, -3895, -2407 };
const short Tone4[] = { 2048, 4095, 2048, -2047, -4095, -2048 };
const short Tone5[] = { 1777, 3993, 3202, 0, -3202, -3993, -1777 };
const short Tone6[] = { 1567, 3784, 3784, 1567, -1567, -3784, -3784, -1567 };
const short Tone7[] = { 1400, 3547, 4033, 2632, 0, -2632, -4033, -3547, -1400 };
const short Tone8[] = { 1265, 3313, 4095, 3313, 1265, -1265, -3313, -4095, -3313, -1265 };
const short Tone9[] = { 1153, 3095, 4054, 3725, 2214, 0, -2214, -3725, -4054, -3095, -1153 };
const short Tone10[] = { 1060, 2896, 3956, 3956, 2896, 1060, -1060, -2896, -3956, -3956, -2896, -1060 };
const short Tone11[] = { 980, 2716, 3829, 4066, 3370, 1903, 0, -1903, -3370, -4066, -3829, -2716, -980 };
const short Tone12[] = { 911, 2553, 3690, 4095, 3690, 2553, 911, -911, -2553, -3690, -4095, -3690, -2553, -911 };
const short Tone13[] = { 851, 2407, 3547, 4073, 3895, 3043, 1665, 0, -1665, -3043, -3895, -4073, -3547, -2407, -851 };
const short Tone14[] = { 799, 2275, 3405, 4017, 4017, 3405, 2275, 799, -799, -2275, -3405, -4017, -4017, -3405, -2275, -799 };
const short Tone15[] = { 752, 2156, 3268, 3939, 4078, 3666, 2759, 1479, 0, -1479, -2759, -3666, -4078, -3939, -3268, -2156, -752 };


/* The following array is a list of tones and their cycle size in samples. */
const ToneCycleType Tones[] = {
	{Tone0, 2},
	{Tone1, 3},
	{Tone2, 4},
	{Tone3, 5},
	{Tone4, 6},
	{Tone5, 7},
	{Tone6, 8},
	{Tone7, 9},
	{Tone8, 10},
	{Tone9, 11},
	{Tone10, 12},
	{Tone11, 13},
	{Tone12, 14},
	{Tone13, 15},
	{Tone14, 16},
	{Tone15, 17},
}; 


/******************************************************************************  
    Function: TonePlay

	Description
	-----------
    This function plays a given tone for a specified time duration.
  
	Arguments
	---------
	ToneIdx:
	A number between 0 and 15 which represents a tone. The tones are
	ordered from highest to lowest frequency. Tone 0 has the highest frequency.

	Duration:
	A time duration in system tick units.

	ToneCallBackP:
	A pointer to callback function that will be called when the time is over.
	If this pointer is NULL this function will wait until the time is over.
	If the pointer is non-NULL this function will be non-blocking i.e., it will
	exit after initializing the tone. 
*******************************************************************************/
void TonePlay(unsigned int ToneIdx, unsigned int Duration,
			  void (*ToneCallBackP)(void))
{
	/* Set up the audio transmit data link. The ToneFillBuf function is set as
	   consumer action. Every time samples are consumed from the data link
	   it will re-fill it immediately.
	*/
	mem2audio_dl = &mem2audio_dl_s;
	dl_init(mem2audio_dl, mem2audio_buf, sizeof(mem2audio_buf));
	dl_init_ptr(mem2audio_dl, mem2audio_ptr);
	dl_c_set_action(mem2audio_dl, ToneFillBuf);

	/* Initialize codec/audio. */
	codec_init(mem2audio_dl, NULL, NULL, NOCVSDCONV);

	/* Assign a pointer to the desired tone. */
	ToneP = &Tones[ToneIdx];

	/* Save the pointer to the callback function. */
	ToneCBP = ToneCallBackP;

	/* Fill data link buffer with samples initially. */
	ToneFillBuf();

	/* Kick off the codec/audio */
	codec_start();

	if (ToneCBP == NULL) {
		/* No callback function. Wait until time expires. */
		timer_wait(Duration);
		/* Then stop codec. */
		codec_stop();
	}
	else
		/* There is a real callback function. Set a wakeup call for ToneEnd. */

		timer_set_wakeup(ToneEnd, Duration, FALSE);
}

/******************************************************************************  
    Function: ToneFillBuf

	Description
	-----------
    This function fill the audio transmit data link buffer with tone samples.
  
	Arguments
	---------
	None. 
*******************************************************************************/
static void ToneFillBuf(void)
{
	int j;

	/* Loop while there's enough room in data link buffer for another
	   tone cycle.
	*/
	while (dl_p_space_available(mem2audio_dl) >= ToneP->CycleSize * 2) {

		/* Data link buffer has room. Add tone cycle samples.*/
		for (j = 0; j < ToneP->CycleSize; j++) {
			*(unsigned short *)mem2audio_ptr = (ToneP->CycleData)[j];
			dl_advance_multiple(mem2audio_dl, mem2audio_ptr, 2);
		}
		dl_p_commit(mem2audio_dl, ToneP->CycleSize * 2);
	}

}

/******************************************************************************  
    Function: ToneEnd

	Description
	-----------
    This function terminates a tone. It stops the codec and calls the callback
	function.
  
	Arguments
	---------
	None. 
*******************************************************************************/
static void ToneEnd(void)
{
	codec_stop();
	(*ToneCBP)();
}
